unit DisturbFR;

interface

const maxdata = 3599;
      maxlongtable = maxdata + 3600;

type
  TDisturb = class
    private
      Index: Integer;
      D1, D2, D3, dSlow, dRange: Real;
      DTable, TTable: array[0..MaxData] of Real;
      LongTable: array[0..MaxLongTable] of Real;
    public
      procedure NewTable; Overload;
      procedure NewTable(Slow, Range: Real); Overload;
      function NextValue: Real;
      function Range: Real;
      function Slow: Real;
    end;

implementation

procedure TDisturb.NewTable;
var
  max, min: real;
    i, j: integer;
begin
  Index := 0;
  repeat
    // make long disturbance table
    d1 := 0.0; d2 := 0.0; d3 := 0.0;
    for i := 0 to maxlongtable do
    begin
      d1 := d1 + dSlow*(20000*(random - 0.5));
      d2 := d2 + dSlow*(d1 - d2);
      d3 := d3 + dSlow*(d2 - d3);
      longtable[i] := d3;
    end;
            // find first place where data[i] = data[i+3599] within 50
    i := 0;
    if (longtable[i] >= longtable[i+maxdata]) then
      while not (longtable[i] <= longtable[i+maxdata]) or
         (i + maxdata < maxlongtable)  do inc(i)
    else
      while not ((longtable[i] >= longtable[i+maxdata]) or
         (i + maxdata > maxlongtable)) do inc(i);
  until (i + maxdata < maxlongtable);

  // transfer to output table
  for j := i to i + maxdata do ttable[j-i] := longtable[j];
  //Scale output table to make max - min value = dRange;
  max := -1e6;
  min := 1e6;
  for j := 0 to maxdata do
  begin
    if ttable[j] > max then max := ttable[j];
    if ttable[j] < min then min := ttable[j];
  end;
  for j := 0 to maxdata do ttable[j] := ttable[j] - (max + min)/2;
  // center within max - min range
  for i := 0 to maxdata do ttable[i] := ttable[i]/(max - min)*dRange;
  // find first zero crossing;
  i := 0;
  While ttable[i]*ttable[i+1] > 0 do inc(i);
  // transfer to dtable with zero at start
  j := 0;
  while j <= maxdata do
  begin
    dtable[j] := ttable[i];
    i := (i + 1) mod (maxdata + 1);
    inc(j);
  end;
end;

procedure TDisturb.NewTable(Slow, Range: Real);
begin
  dSlow := Slow;
  dRange := Range;
  NewTable;
end;

function TDisturb.NextValue: Real;
begin
  NextValue := DTable[Index];
  if Index = MaxData then
    begin
      Index := 0;
      NewTable;
    end else Inc(Index);
end;

function TDisturb.Range: Real;
begin
  Range := dRange;
end;

function TDisturb.Slow: Real;
begin
  Slow := dSlow;
end;


end.
