unit GrUtils5;

interface
uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls;

const
  maxh = 100;
  maxw = 100;
  maxvert = 10;
  a360 = 4096*2;
  a180 = a360 div 2;
  col: array[0..15] of TColor = (clRed,    clBlue,   clGreen,   clAqua,
                                 clltGray, clSilver, clWhite,   clTeal,
                                 clMaroon, clYellow, clFuchsia, clPurple,
                                 clDkGray, clNavy,   clSilver,  clOlive);
type
  PointType = record
                x3d, y3d, z3d, x2d, y2d: SmallInt;
                end;

  AngleType = record
                theta, theta0, phi, phi0, rho, rho0: Integer;
              end;

  FaceType = array of PointType;

  FaceArrayType = array of record
                    f: FaceType;
                    c, cr: TColor;  // color and reflected color
                    a: AngleType;
                  end;

  TPoint2D = array of TPoint; // DYNAMIC ARRAY FOR "POLYGON" PROCEDURE

  ListType = array of SmallInt;

  Lighttype = record
                x, y, z: Single;
              end;

var
  Isin, Icos: array[0..a360] of Single;  // SINE AND COSINES AS TABLES
  SortKeys, SortLinks: ListType;
  p: Tpoint2d;
  Light: LightType;
  Scale: Double;
  NumW, NumH, Nface, Nvert: Word;
  WireFrame: Boolean;
  ScreenDistance, xdiameter, ydiameter, zdiameter: SmallInt;
  SurfaceColor: TColor;

procedure MakeIsinIcos;
procedure ClearBitMap(BackColor: Integer; Bm: TBitmap);
procedure LimitAngle(var A: Integer);
procedure Rotate3D(var source: FaceArrayType;  //rotate 3D points in place
                   Axis: Char;             // rotate about 'x', 'y', or 'z'
                   Angle: Integer;        // a360 = 2pi
                   x0, y0, z0: SmallInt);// center of rotation (offset)
procedure ProjectFaces(var Source: FaceArrayType; x, y, z: Integer);
procedure ShowFaces(var Source: FaceArrayType; bm: TBitmap);
procedure ShowArray(var Source: FaceArrayType; n: Byte; bm: TBitmap);
procedure MakeSphere(var Source: FaceArrayType);
procedure MakeCube(var Source: FaceArrayType);
procedure MakeRect(var Source: FaceArrayType);

implementation

procedure MakeIsinIcos;
//Create table of sines and cosines
//The table contains 2*4096 values of each
var
  i: SmallInt;
begin
  for i := 0 to a360 do
  begin
    Isin[i] := Sin(2*Pi*i/a360);
    Icos[i] := Cos(2*Pi*i/a360);
  end;
end;

procedure ClearBitmap(BackColor: Integer; bm: TBitmap);
begin
  with Bm, Canvas do
  begin
    Brush.Color := BackColor;
    FillRect(Rect(0, 0, Width, Height));
  end;
end;

procedure LimitAngle(var A: Integer);
//Angle is an integer value between 0 and 2*4096
//Keep angle between those bounds by wrapping around
begin
  while A < 0 do A := A + a360;
  while A > a360 do A := A - a360;
end;

procedure QuickSort(L, R: Word);
//Sorts elements in the array SortKeys with indices between
//LO and HI (both inclusive). It executes itself recursively.
var
  i, j, X, Y: Integer;

  function GetValue(i: Integer): Integer; // sort only the pointers to the keys.
  begin
    Result := SortKeys[SortLinks[i]];
  end;

begin
  i := L; j := R; X := (GetValue(L)+ GetValue(R)) DIV 2;
  repeat
    while GetValue(i) > x do i := i + 1;
    while X > GetValue(j) do j := j - 1;
    if i <= j then
    begin
      Y := SortLinks[i];
      SortLinks[i] := SortLinks[j];
      SortLinks[j] := Y;
      i := i + 1; j := j - 1;
    end;
  until i > j;
  if L < j then QuickSort(L, j);
  if i < R then QuickSort(i, R);
end;

procedure SortByDistance(var Source: FaceArrayType);
// sort indexes to the faces by the average z distance to the faces.
// this is for drawing only visible faces
var i, j, AvgZ: SmallInt;
begin
  SetLength(SortKeys, Nface);
  SetLength(SortLinks, Nface);
  for i := 0 to Nface - 1 do  // scan through all faces
  begin
    AvgZ := 0;
    for j := 0 to Nvert-1 do AvgZ := AvgZ + Source[i].f[j].z3d;
    AvgZ := AvgZ div Nvert;   // avg z-distance of vertex
    SortLinks[i] := i;        // load index array in order of indices
    SortKeys[i] := AvgZ;
  end;
  QuickSort(0, Nface-1);     // rearrange the links in place
end;

procedure ShowArray(var source: FaceArrayType;
                    n: Byte; bm: TBitmap);
// This draws the wire-frame version
var
  i, j: Word;
  x, y: SmallInt;
begin
  for i := 0 to Nface-1 do
  for j := 0 to Nvert-1 do
  with Source[i].f[j mod nvert] do   // "mod" wraps around to starting vertex
    with bm, Canvas do
    begin
      Pen.Color := clRed;
      if n = 3 then
        begin
          x := Width div 2 + x3d; y := Height div 2 - y3d;
        end
      else
        begin
          x := Width div 2 + x2d; y := Height div 2 - y2d;
        end;
      if j  = 0 then MoveTo(x, y) else LineTo(x, y);
    end;
end;

procedure ShowFaces(var Source: FaceArrayType; bm: TBitmap);
//This draws the faces with color filled in
var
  i, j, k: Word;
begin
  SortByDistance(Source);
  SetLength(p, Nvert);
  for k := 0 to Nface-1 do
  with bm, Canvas do
  begin
    i := SortLinks[k];
    for j := 0 to Nvert - 1 do
      with Source[i] do
      begin
        with f[j] do
        begin
          p[j].x := Width div 2 + x2d ; // load array of tpoints for drawing
          p[j].y := Height div 2 - y2d;
        end;
        Brush.Color := cr;
      end;
    Pen.Color := Brush.Color; // set pen color and face color the same
    Polygon(p);              // draw face to bitmap; fills in with brush color.
  end;
end;

procedure CalcLighting(var S: FaceArrayType);
// This computes angle between incoming light rays and the normal to the
// faces of the object. The color reflected is proportional to the cosine
// of the angle.
var
  l1, m1, n1, l2, m2, n2: Integer;
  lnorm, mnorm, nnorm: Double;
  Temp: Double;
  Denom: Double;
  brt: Single;
  r, g, b: Byte;
  i: Integer;
  Swap: Boolean;
begin
  for i := 0 to Nface-1 do
    with s[i] do
    begin
      l1 := f[1].x3d - f[0].x3d;  // direction numbers of first two lines
      m1 := f[1].y3d - f[0].y3d;  // in face
      n1 := f[1].z3d - f[0].z3d;
      Swap := (l1 = 0) and (m1 = 0) and (n1 = 0);
      if Swap then                // use next two lines if all d.n. are zero
      begin
        Swap := true;
        l1 := f[2].x3d - f[1].x3d;  // direction numbers of next two lines
        m1 := f[2].y3d - f[1].y3d;  // in face
        n1 := f[2].z3d - f[1].z3d;
      end;
      if not (Swap or ((l2 = 0) and (m2 = 0) and (n2 = 0))) then
      begin
        l2 := f[2].x3d - f[1].x3d;  // use last choices of lines
        m2 := f[2].y3d - f[1].y3d;
        n2 := f[2].z3d - f[1].z3d;
      end
      else
      begin
        l2 := f[3].x3d - f[2].x3d;
        m2 := f[3].y3d - f[2].y3d;
        n2 := f[3].z3d - f[2].z3d;
      end;
    lnorm := m1*n2 - m2*n1;        // direction cosines of normal to plane
    mnorm := n1*l2 - n2*l1;
    nnorm := l1*m2 - l2*m1;
    Temp := (light.x*lnorm + light.y*mnorm + light.z*nnorm);
    Denom := Sqrt(Sqr(lnorm) + Sqr(mnorm) + Sqr(nnorm));
    if denom = 0 then brt := 1.0 else
      brt := 0.4 - 0.6*Temp/Denom; // algorithm for calculating angle
    if brt < 0 then brt := 0 else if brt > 1.0 then brt := 1.0;
    b := Round( ((c and $00ff0000) shr 16) * brt) and $ff; // blue brightness
    g := Round(((c and $0000ff00) shr 8) * brt) and $ff;   // green brighhtness
    r := Round((c and $000000ff) * brt) and $ff;           // red brightness
    cr := RGB(r, g, b);
  end;
end;

procedure Rotate3D(var Source: FaceArrayType;  //rotate 3D points in place
                   Axis: Char;             // rotate about 'x', 'y', or 'z'
                   Angle: Integer;        // a360 = 2pi
                   x0,y0,z0: SmallInt);  // center of rotation
// ROTATE ALL POINTS ABOUT ONE AXIS, X, Y, OR Z. REPLACE PREVIOUS COORDINATES
var
  i, j: SmallInt;
  ca, sa: Single;
  x, y, z, xp, yp, zp: SmallInt;
begin
  LimitAngle(Angle);
  for i := 0 to Nface-1 do
  begin
    for j := 0 to Nvert-1 do
      with Source[i] do
      begin
        with f[j] do
        begin
          x := x3d - x0;     // REMOVE CENTER OF ROTATION
          y := y3d - y0;
          z := z3d - z0;
        end;
        with a do           // ADD ROTATIONS TO ANGLES
          case Axis of
            'x': Theta := Theta0 + Angle;
            'y': Phi := Phi0 + Angle;
            'z': Rho := Rho0 + Angle;
          end;
        ca := icos[Angle]; sa := isin[Angle];   // look up sine, cosine
        case axis of
          'x': begin
                 xp := x;
                 yp := Round(ca*y + sa*z);     // rotation matrix
                 zp := Round(-sa*y + ca*z);
               end;
          'y': begin
                 xp := Round(ca*x + sa*z);
                 yp := y;
                 zp := Round(-sa*x + ca*z);
               end;
          'z': begin
                 xp := Round(ca*x + sa*y);
                 yp := Round(-sa*x + ca*y);
                 zp := z;
               end;
          else begin
                 xp := x;  // if illegal axis, don't rotate
                 yp := y;
                 zp := z;
               end;
        end;
        with f[j] do
        begin
          x3d := Round(Scale*xp) + x0;   // RESTORE OFFSET FROM AXIS OF ROTATION
          y3d := Round(Scale*yp) + y0;
          z3d := Round(Scale*zp) + z0;
        end;
      end; // of <with source[i]>
      if not WireFrame then CalcLighting(Source);
  end;
end;

procedure ProjectFaces(var Source: FaceArrayType; x, y, z: Integer);
// PROJECT ALL POINTS FOR DRAWING; TRANSLATE CENTER TO X,Y,Z
// X, Y, Z change the 2-D projection of the image but do not affect
// the internal 3-D values of the points.
var
  i, j, d: SmallInt;
begin
  for i := 0 to Nface-1 do
    for j := 0 to Nvert-1 do
      with Source[i].f[j] do
      begin
        d := z3d + z; // avoid division by zero, but don't change z3d
        if d = -ScreenDistance then Inc(d);
        x2d := Round((x3d + x)*ScreenDistance/(ScreenDistance + d));
        y2d := Round((y3d + y)*ScreenDistance/(ScreenDistance + d));
       end;
end;

procedure MakeSphere(var Source: FaceArrayType);
var
  i, j, k, t, p, dtheta, dphi: Integer;
  N: Word;
begin
  Nface := NumH * NumW;
  SetLength(Source, Nface);
  for i := 0 to Nface - 1 do SetLength(Source[i].f, Nvert);
  dtheta := Round(a360/numw);
  dphi := Round(a180/numh);
  for i := 0 to NumW-1 do
    for j := 0 to NumH-1 do
    begin
      N := i*NumH + j;
      for k :=  0 to Nvert-1 do
        with Source[N], f[k] do
        begin
          a.theta0 := Round(i*dtheta); // coordinates of lower left corner
          a.phi0 := Round(j*dphi);
          a.rho0 := 0;
          LimitAngle(a.theta0);
          LimitAngle(a.phi0);
          case k of
            0: begin t := a.theta0; p := a.phi0; end;
            1: begin t := a.theta0 + dtheta; p := a.phi0; end;
            2: begin t := a.theta0 + dtheta; p := a.phi0 + dphi; end;
            3: begin t := a.theta0; p := a.phi0 + dphi; end;
          end;
          LimitAngle(t);
          with Source[n] do
          begin
            with f[k] do
            begin
              x3d := Round((xdiameter)*isin[p]*icos[t]);
              y3d := Round((ydiameter)*isin[p]*isin[t]);
              z3d := Round(zdiameter*icos[p]);
            end;
            c := SurfaceColor;
          end;
        end;
    end;
end;

procedure MakeRect(var Source: FaceArrayType);
// make standard size rectangle in x, y, or z plane
var
  i, j: Word;
begin
  Nface := 1; Nvert := 4;
  SetLength(Source,Nface);
  for i := 0 to Nface-1 do SetLength(Source[i].f,Nvert);
  for i := 0 to Nface-1 do
    for j := 0 to Nvert-1 do
    begin
      with Source[i] do   // source[i] is the ith face in the object
      begin
        with f[j] do   // f[j] is the vertex number (counterclockwise) in the face
        begin
          case j of
            3: begin x3d := 70; y3d := 70; end;
            2: begin x3d := -70; y3d := 70; end;
            1: begin x3d := -70; y3d := -70; end;
            0: begin x3d := 70; y3d := -70; end;
          end;
          z3d := 100;
        end;
        c := Col[i mod 16]; // make faces different colors
      end;
    end;
end;

procedure MakeCube(var Source: FaceArrayType);
var
  i, j: Word;
begin
  Nface := 6; Nvert := 4;
  SetLength(Source, Nface);
  for i := 0 to Nface-1 do // construct each of 6 faces.
  begin
    SetLength(Source[i].f, Nvert);
    with Source[i] do
    begin
      for j := 0 to Nvert-1 do
        with f[j] do
          case i of
            0: begin
                 case j of
                   3: begin x3d := 100; y3d := 100; end;
                   2: begin x3d := -100; y3d := 100; end;
                   1: begin x3d := -100; y3d := -100; end;
                   0: begin x3d := 100; y3d := -100; end;
                 end;
                 z3d := 100;
               end;
            1: begin
                 case j of
                   0: begin x3d := 100; y3d := 100; end;
                   1: begin x3d := -100; y3d := 100; end;
                   2: begin x3d := -100; y3d := -100; end;
                   3: begin x3d := 100; y3d := -100; end;
                 end;
                 z3d := -100;
               end;
            2: begin
                 case j of
                   0: begin x3d := 100; z3d := 100; end;
                   1: begin x3d := -100; z3d := 100; end;
                   2: begin x3d := -100; z3d := -100; end;
                   3: begin x3d := 100; z3d := -100; end;
                 end;
                 y3d := 100;
               end;
            3: begin
                 case j of
                   3: begin x3d := 100; z3d := 100; end;
                   2: begin x3d := -100; z3d := 100; end;
                   1: begin x3d := -100; z3d := -100; end;
                   0: begin x3d := 100; z3d := -100; end;
                 end;
                 y3d := -100;
               end;
            4: begin
                 case j of
                   3: begin y3d := 100; z3d := 100; end;
                   2: begin y3d := -100; z3d := 100; end;
                   1: begin y3d := -100; z3d := -100; end;
                   0: begin y3d := 100; z3d := -100; end;
                 end;
                 x3d := 100;
               end;
            5: begin
                 case j of
                   0: begin y3d := 100; z3d := 100; end;
                   1: begin y3d := -100; z3d := 100; end;
                   2: begin y3d := -100; z3d := -100; end;
                   3: begin y3d := 100; z3d := -100; end;
                 end;
                 x3d := -100;
               end;
            end; //case i of
      if i = 1 then c := clLime else // make one face green
           c := clWhite; //col[i mod 16]; // and the rest white
    end; // with source[i]
  end; // for i
end;

end.
