unit Main;

//Program: ChooseControl.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ExtCtrls, StdCtrls;

const
  LastData = 3600;

type
  DataArray = array[1..LastData] of Real;

  TMainForm = class(TForm)
    ControlPanel: TGroupBox;
    MainPanel: TPanel;
    UpperPanel: TGroupBox;
    ObjectPanel: TPanel;
    ObjectPaint: TPaintBox;
    GroupBox3: TGroupBox;
    Button1: TButton;
    AboutButton: TButton;
    ShapeGraph: TPaintBox;
    ShapeBox: TGroupBox;
    ShapePanel: TPanel;
    LowerPanel: TPanel;
    AxisBox: TGroupBox;
    AxisPanel: TPanel;
    AxisGraph: TPaintBox;
    PositionBox: TGroupBox;
    PositionPanel: TPanel;
    PositionGraph: TPaintBox;
    InstructBtrn: TButton;
    KeyBox: TGroupBox;
    Label10: TLabel;
    Label11: TLabel;
    Label12: TLabel;
    Label13: TLabel;
    Label14: TLabel;
    Label15: TLabel;
    Label16: TLabel;
    Shape1: TShape;
    Shape2: TShape;
    Shape3: TShape;
    Shape4: TShape;
    Panel1: TPanel;
    SDCorrM: TEdit;
    SCorrM: TEdit;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    SCorrSD: TEdit;
    Panel2: TPanel;
    ACorrAD: TEdit;
    ADCorrM: TEdit;
    ACorrM: TEdit;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Panel3: TPanel;
    HCorrHD: TEdit;
    HDCorrM: TEdit;
    HCorrM: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Panel4: TPanel;
    ControlPadPanel: TGroupBox;
    Pad: TPanel;
    MouseTrack: TTrackBar;
    RunControlPanel: TGroupBox;
    StartBtn: TButton;
    ResetBtn: TButton;
    ETcounter: TEdit;
    GroupBox6: TGroupBox;
    Status: TEdit;
    ModelBtn: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure MouseTrackChange(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure StartBtnClick(Sender: TObject);
    procedure ResetBtnClick(Sender: TObject);
    procedure InstructBtrnClick(Sender: TObject);
    procedure AboutButtonClick(Sender: TObject);
    procedure PositionGraphPaint(Sender: TObject);
    procedure AxisGraphPaint(Sender: TObject);
    procedure ShapeGraphPaint(Sender: TObject);
    procedure ModelBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure Initialize;
    procedure OperateShape;
    procedure Iterate;
    procedure RecordData;
    procedure IdleLoop(Sender: TObject; var Done: boolean);
    procedure PlotGraph(Color: Integer; var Data: DataArray;
                        YRange: Real; Bitmap: TBitmap);
    procedure PlotGraphs;
    procedure CalcStats;
    procedure ShowStats;
  end;

var
  MainForm: TMainForm;
  CV, Stage: Integer;
  HorizPos, Symmet, Direct, MousePos,
    HDisturb, DDisturb, SDisturb: DataArray;
  Plotted: Boolean;

implementation

uses GrUtils5, DisturbFR, Instructions, About, Model;

{$R *.dfm}

var
  T1, T2, Freq, Elapsed, StartDelay: Int64;
  Shape: FaceArrayType;
  OffScreen, HP, SY, DI: TBitmap;
  HPosition, Symmetry, Angle, DX, DY, DZ, MouseY: Integer;
  HrHD, HrM, HDrM, ArAD, ArM, ADrM, SrSD, SrM, SDrM: Real;
  D1, D2, D3: TDisturb;
  VRange: Real;

procedure TMainForm.Initialize;
begin
  Stage := 0; //Practice
  StartDelay := 180; //3 seconds
  Elapsed := -StartDelay;
  Plotted := False;
  SurfaceColor := clRed; // Sphere color
  Randomize;
  CV := -1;

  HP.Canvas.Pen.Width := 2;
  SY.Canvas.Pen.Width := 2;
  DI.Canvas.Pen.Width := 2;

//Initialize sphere
  MakeIsinIcos; //Make sine and cosine table
//  SurfaceColor := clWhite;
  ScreenDistance := 1860;
  Scale := 0.7;        //0.7
  Nvert := 4;         // vertices per face
  NumH := 8;
  NumW := 8;
  Light.X := Sqrt(2)/2;  // direction cosines
  Light.Y := 0;
  Light.Z := -Sqrt(2)/2;

//Initialize disturbance objects
  VRange := 1200.0;
  randomize;
  D1.NewTable(0.05, VRange); // Disturbance ranges match range of mouse trackbar
  D2.NewTable(0.05, VRange);
  D3.NewTable(0.05, VRange);
end;

procedure TMainForm.Iterate;
begin
  OperateShape;
  Case Stage of
    1: begin
         Plotted := False;
         RecordData;
       end;
    2: if NOT Plotted then
         begin
           PlotGraphs;
           Plotted := True;
         end;
    3: if ModelingForm.Visible then ModelingForm.RunModel;
  end;
end;

procedure TMainForm.OperateShape;
begin
  //Position
  DX := Round(D1.NextValue);
  HPosition:= round(1.5*(MouseY + DX));
  if HPosition > 1000 then HPosition := 1000
  else
  if HPosition < -1000 then HPosition := -1000;
  //Axis Angle
  DY := Round(D2.NextValue);
  Angle := MouseY + DY;
  if angle > 1000 then angle := 1000
  else
  if angle < -1000 then angle := -1000;
  //Shape
  DZ := Round(D3.NextValue);
  Symmetry := (MouseY + DZ);
  if symmetry > 800 then symmetry := 800
  else
  if symmetry < -800 then symmetry := -800;

  with ObjectPaint do //Draw the sphere
  // Scaling is now done here.
  begin
    YDiameter := ClientHeight div 3 - Symmetry div 12;
    XDiameter := ClientHeight div 3 +  Symmetry div 12;
    ZDiameter := ClientHeight div 3 - Symmetry div 12;
    MakeSphere(Shape);
    Rotate3D(Shape, 'x', 2*Angle ,0, 0, 0);
    ProjectFaces(Shape, HPosition*Width div 2600, 0, 0); //Mouse + Dist range
    ClearBitMap(clBlack, OffScreen);
    ShowFaces(Shape, OffScreen);
    Canvas.Draw(0, 0, OffScreen);
  end;
end;

procedure TMainForm.RecordData;
var
  ET: Integer;
begin
  Inc(Elapsed);
  ET := Elapsed div 60;
  ETcounter.Text := Format('%3d',[ET]);
  if Elapsed = 0 then
  begin
     Surfacecolor := clLime;
     Status.Color := clLime;
     Status.Text := 'Recording';
  end;
  if Elapsed > 0 then // Elapsed starts at < 0 to allow time to grab slider.
  begin
    HorizPos[Elapsed] := HPosition;
    HDisturb[Elapsed] := DX;
    Direct[Elapsed] := Angle;
    DDisturb[Elapsed] := DY;
    Symmet[Elapsed] := Symmetry;
    SDisturb[Elapsed] := DZ;
    MousePos[Elapsed] := MouseY;
  end;
  if Elapsed >= LastData then //End run at 60 seconds
  begin
    Elapsed := -StartDelay;
    Status.Text := 'Plotting';
    Stage := 2;
  end;
end;

procedure TMainForm.PlotGraph(Color: Integer;
  var Data: DataArray; YRange: Real; Bitmap: TBitmap);
var
  I, Xpos, Ypos, Middle: Integer;
begin
  Middle := Bitmap.Height div 2;
  with Bitmap.Canvas do
  begin
    Pen.Color := Color;
    Ypos := Middle - round(Data[1]*Bitmap.Height/(YRange));
    MoveTo(0, Ypos);
    for I := 1 to LastData do
    begin
      XPos := I*Bitmap.Width div LastData;
      YPos := Middle - round(Data[I]*Bitmap.Height/(YRange));
      LineTo(XPos, YPos);
    end;
  end;
end;

procedure TMainForm.PlotGraphs;
var
  PlotRange: Real;
begin
  PlotRange := 1.4*VRange;
  //Plot position
  HP.Canvas.FillRect(Rect(0, 0, HP.Width, HP.Height));
  PlotGraph(clBlack, HorizPos, PlotRange, HP);
  PlotGraph(clGreen, HDisturb, PlotRange, HP);
  PlotGraph(clRed, MousePos, PlotRange, HP);
  //Plot angle
  DI.Canvas.FillRect(Rect(0, 0, DI.Width, DI.Height));
  PlotGraph(clBlack, Direct, PlotRange, DI);
  PlotGraph(clGreen, DDisturb, PlotRange, DI);
  PlotGraph(clRed, MousePos, PlotRange, DI);
  //Plot shape
  SY.Canvas.FillRect(Rect(0, 0, SY.Width, SY.Height));
  PlotGraph(clBlack, Symmet, PlotRange, SY);
  PlotGraph(clGreen, SDisturb, PlotRange, SY);
  PlotGraph(clRed, MousePos, PlotRange, SY);
  //Reset indicators to 'practice' state
  SurfaceColor := clRed;
  Status.Color := clRed;
  Status.Text := 'Practice';
  //Show the graphs
  PositionGraph.Canvas.Draw(0, 0, HP);
  AxisGraph.Canvas.Draw(0, 0, DI);
  ShapeGraph.Canvas.Draw(0, 0, SY);
  //Compute and show the stats
  CalcStats;
  ShowStats;
  Stage := 3;
end;

function SumSq(Data: DataArray): Real;
var
  I: Integer;
  SumX, SumXSq, SS: Real;
begin
  SumX := 0.0;
  SumXSq := 0.0;
  for I := 1 to LastData do
  begin
    SumX := SumX + Data[I];
    SumXSq := SumXSq + Data[I]*Data[I];
  end;
  SS := SumXSq - SumX*SumX/LastData;
  SumSq := SS;
end;

function SumProducts(Data1, Data2: DataArray): Real;
var
  I: Integer;
  SumX, SumY, SumXY, SP: Real;
begin
  SumX := 0.0;
  SumY := 0.0;
  SumXY := 0.0;
  for I := 1 to LastData do
  begin
    SumX := SumX + Data1[I];
    SumY := SumY + Data2[I];
    SumXY := SumXY + Data1[I]*Data2[I];
  end;
  SP := SumXY - SumX*SumY/LastData;
  SumProducts := SP;
end;

function Corr(Data1, Data2: DataArray): Real;
var
  SSx, SSy, SP: Real;
begin
   SSx := SumSq(Data1);
   SSy := SumSq(Data2);
   SP := SumProducts(Data1, Data2);
   if (SSx = 0.0) or (SSy = 0.0) then Corr := 0.0
   else Corr := SP/Sqrt(SSx*SSy);
end;

procedure TMainForm.CalcStats;
begin
  HrHD := Corr(HorizPos, HDisturb);
  HrM  := Corr(HorizPos, MousePos);
  HDrM := Corr(HDisturb, MousePos);

  ArAD := Corr(Direct, DDisturb);
  ArM  := Corr(Direct, MousePos);
  ADrM := Corr(DDisturb, MousePos);

  SrSD := Corr(Symmet, SDisturb);
  SrM  := Corr(Symmet, MousePos);
  SDrM := Corr(SDisturb, MousePos);
end;

procedure TMainForm.ShowStats;
var
  Min: Real;
begin
  HCorrHD.Text := Format('%4.2f', [HrHD]);
  HCorrM.Text := Format('%4.2f', [HrM]);
  HDCorrM.Text := Format('%4.2f', [HDrM]);

  ACorrAD.Text := Format('%4.2f', [ArAD]);
  ACorrM.Text := Format('%4.2f', [ArM]);
  ADCorrM.Text := Format('%4.2f', [ADrM]);

  SCorrSD.Text := Format('%4.2f', [SrSD]);
  SCorrM.Text := Format('%4.2f', [SrM]);
  SDCorrM.Text := Format('%4.2f', [SDrM]);

  Min := HDrM;
  if ADrM < Min then Min := ADrM;
  if SDrM < Min then Min := SDrM;
  if (HDrM = Min) and (HDrM < 0.0 ) then
    begin
      PositionBox.Color := clYellow;
      CV := 0;
    end
  else if (ADrM = Min) and (ADrM < 0.0) then
    begin
      AxisBox.Color := clYellow;
      CV := 1;
    end
  else if (SDrM = Min) and (SDrM < 0.0) then
    begin
      ShapeBox.Color := clYellow;
      CV := 2;
    end;
    ModelingForm.InitializeModel;
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: boolean);
begin
  Done := False;
  Iterate;
  repeat
    QueryPerformanceCounter(T2);
  until (T2 - T1) >= Freq/60;
  T1 := T2;
  Application.ProcessMessages;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
//  ChangeScale(Screen.Width, 1024); // 768
  WindowState := wsMaximized;
//Create offscreen bitmap for sphere
  OffScreen := TBitmap.Create;
//Create ffscreen bitmaps for graphs
  HP := TBitmap.Create;
  SY := TBitmap.Create;
  DI := TBitmap.Create;
  FormResize(Sender); // Size bitmaps to their onscreen windows
//Create three disturbance objects
  D1 := TDisturb.Create;
  D2 := TDisturb.Create;
  D3 := TDisturb.Create;
  Initialize;
//Get timer info and establish idle loop;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(T1);
  Application.OnIdle := IdleLoop;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
//Free memory allocated to bitmap and to disturbance objects
 OffScreen.Free;
 HP.Free;
 SY.Free;
 DI.Free;
 D1.Free;
 D2.Free;
 D3.Free;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  UpperPanel.Height := MainForm.ClientHeight div 2;
  ObjectPanel.Height := UpperPanel.Height - 2* ObjectPanel.Top;
  ControlPadPanel.Height := Panel4.Height - 2*ControlPadPanel.Top;
  Pad.Height := ControlPadPanel.ClientHeight - 30;
  MouseTrack.Height := Pad.Height - 2*MouseTrack.Top;
  PositionBox.Height := LowerPanel.ClientHeight div 3;
  PositionPanel.Height := PositionBox.ClientHeight - 30;
  AxisBox.Height := PositionBox.Height;
  AxisPanel.Height := AxisBox.ClientHeight - 30;
  ShapeBox.Height := PositionBox.Height;
  ShapePanel.Height := ShapeBox.ClientHeight - 30;
  KeyBox.Top := UpperPanel.Height;

//Resize offscreen bitmap to fit onscreen window
  with OffScreen do
  begin
    Height := ObjectPaint.ClientHeight;
    Width := ObjectPaint.ClientWidth;
  end;
  with HP do
  begin
    Width := PositionGraph.ClientWidth;
    Height := PositionGraph.ClientHeight;
  end;
  with SY do
  begin
    Width := ShapeGraph.ClientWidth;
    Height := ShapeGraph.ClientHeight;
  end;
  with DI do
  begin
    Width := AxisGraph.ClientWidth;
    Height := AxisGraph.ClientHeight;
  end;
end;

procedure TMainForm.MouseTrackChange(Sender: TObject);
begin
  MouseY := MouseTrack.Position;
end;

procedure TMainForm.Button1Click(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.StartBtnClick(Sender: TObject);
begin
  If Stage <> 1 then
  begin
    Stage := 1;
    Status.Text := 'Started';
    PositionBox.Color := clBtnFace;
    AxisBox.Color := clBtnFace;
    ShapeBox.Color := clBtnFace;
    HP.Canvas.FillRect(Rect(0, 0, HP.Width, HP.Height));
    DI.Canvas.FillRect(Rect(0, 0, DI.Width, DI.Height));
    SY.Canvas.FillRect(Rect(0, 0, SY.Width, SY.Height));
    PositionGraph.Canvas.Draw(0, 0, HP);
    AxisGraph.Canvas.Draw(0, 0, DI);
    ShapeGraph.Canvas.Draw(0, 0, SY);
  end;
end;

procedure TMainForm.ResetBtnClick(Sender: TObject);
begin
  if Stage = 1 then
  begin
    Stage := 0;
    Elapsed := -StartDelay;
    Status.Color := clRed;
    Status.Text := 'Practice';
    Surfacecolor := clRed;  //Sphere color
  end;
end;

procedure TMainForm.InstructBtrnClick(Sender: TObject);
begin
  InstructForm.Show;
end;

procedure TMainForm.AboutButtonClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

procedure TMainForm.PositionGraphPaint(Sender: TObject);
begin
  PositionGraph.Canvas.Draw(0, 0, HP);
end;

procedure TMainForm.AxisGraphPaint(Sender: TObject);
begin
  AxisGraph.Canvas.Draw(0, 0, DI);
end;

procedure TMainForm.ShapeGraphPaint(Sender: TObject);
begin
  ShapeGraph.Canvas.Draw(0, 0, SY);
end;

procedure TMainForm.ModelBtnClick(Sender: TObject);
begin
  If Stage > 1 then
    begin
      ModelingForm.Characteristic.ItemIndex := CV;
      ModelingForm.ShowModal;
    end;
end;

end.

