unit Model;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, ComCtrls, Main;

type
  TModelingForm = class(TForm)
    ModelBox: TGroupBox;
    ParamsBox: TGroupBox;
    ModelPaint: TPaintBox;
    GainUpDown: TUpDown;
    GainEdit: TEdit;
    Label1: TLabel;
    RefEdit: TEdit;
    RefUpDown: TUpDown;
    Label2: TLabel;
    Characteristic: TRadioGroup;
    RMSedit: TEdit;
    Label3: TLabel;
    Panel1: TPanel;
    Shape1: TShape;
    Label4: TLabel;
    Shape2: TShape;
    Label5: TLabel;
    Shape3: TShape;
    Label6: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure RefUpDownClick(Sender: TObject; Button: TUDBtnType);
    procedure GainUpDownClick(Sender: TObject; Button: TUDBtnType);
    procedure CharacteristicClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure InitializeModel;
    procedure CalcModel(var Disturbance: DataArray);
    procedure PlotModel(PlotRange: Real);
    procedure RunModel;
  end;

var
  ModelingForm: TModelingForm;
  ModelBitmap: TBitmap;

implementation

{$R *.dfm}



var
  ModelMouse: DataArray;
  Middle: Integer;
  Output, Gain, Ref, dtime, AvgMouse: Real;


procedure TModelingForm.InitializeModel;
begin
  Middle := ModelBitmap.Height div 2;
  Characteristic.ItemIndex := CV;
  Ref := 0.0;
  Gain := 1.0;
  dtime := 1/60.0;
end;

function RMSerror: Real;
var
  I: Integer;
  Error, SumSqError, min, max: Double;
begin
  SumSqError := 0.0;
  avgmouse := 0.0;
  min := 1e6;
  max := -1e6;
  for I := 1 to LastData do
  begin
   avgmouse := avgmouse + MousePos[i];
   if mousepos[i] > max then max := mousepos[i];
   if mousepos[i] < min then min := mousepos[i];
  end;
  avgmouse := (max + min)/2;
  for I := 1 to LastData do
  begin
    Error := MousePos[I] - avgmouse - ModelMouse[I];
    SumSqError := SumSqError + Error*Error;
  end;
  if Max = Min then Max := Max + 1;
  RMSerror := 100.0*Sqrt(Sumsqerror/Lastdata)/(max - min);
end;

procedure TModelingForm.CalcModel(var Disturbance: DataArray);
var
  I: Integer;
  p, e: Real;
begin
  Output := 0.0;
  for I := 1 to LastData do
  begin
    p := Output + Disturbance[I];
    e := 2400/ModelBitmap.Height*Ref - p;
    Output := Output + Gain*e*dtime;
    ModelMouse[I] := Output;
  end;
end;

procedure TModelingForm.PlotModel(PlotRange: Real);
var
  I, Xpos, Mpos, Midscale: Integer;
  Error: Real;
begin
  with ModelBitmap.Canvas do
  begin
    FillRect(Rect(0, 0, ModelBitmap.Width, ModelBitmap.Height));
    Midscale := ModelBitmap.Height div 2;
    //Draw zero-line
    Pen.Color := clGray;
    MoveTo(0, Midscale);
    LineTo(ModelBitmap.Width, Midscale);
    //Plot actual mouse position
    Pen.Color := clRed;
    Mpos := Midscale - round(MousePos[1]
      * ModelBitmap.Height/(PlotRange));
    MoveTo(0, Mpos);
    for I := 1 to LastData do
    begin
      XPos := I*ModelBitmap.Width div LastData;
      Mpos := (ModelBitmap.Height div 2) - round(MousePos[I]
        * ModelBitmap.Height/PlotRange);
      LineTo(XPos, Mpos);
    end;
    //Plot model mouse position
    Pen.Color := clLime;
    Mpos := Midscale - round(ModelMouse[1]
      * ModelBitmap.Height/(PlotRange));
    MoveTo(0, Mpos);
    for I := 1 to LastData do  //Plot model data
    begin
      XPos := I*ModelBitmap.Width div LastData;
      Mpos := ModelBitmap.Height div 2 - round(ModelMouse[I]
        * ModelBitmap.Height/PlotRange);
      LineTo(XPos, Mpos);
    end;
    // Plot error
    Pen.Color := clBlack;
    Error := MousePos[1] - ModelMouse[1];
    Mpos := Midscale - round(Error
      * ModelBitmap.Height/(PlotRange));
    MoveTo(0, Mpos);
    for I := 1 to LastData do  //Plot model data
    begin
      XPos := I*ModelBitmap.Width div LastData;
      Error := MousePos[I] - ModelMouse[I];
      Mpos := ModelBitmap.Height div 2 - round(Error
        * ModelBitmap.Height/PlotRange);
      LineTo(XPos, Mpos);
    end;
  end;
  ModelPaint.Canvas.Draw(0, 0, ModelBitmap);
  RMSedit.Text := Format('%10.2f', [RMSerror]);
end;

procedure TModelingForm.RunModel;
begin
  Case CV of
    0: CalcModel(HDisturb);
    1: CalcModel(DDisturb);
    2: CalcModel(SDisturb);
    end;
  PlotModel(2400.0);
end;

procedure TModelingForm.FormCreate(Sender: TObject);
begin
  with ModelingForm do
  begin
    ChangeScale(Screen.Height, 768);
    ModelBitmap := TBitmap.Create;
    Left := Screen.WorkAreaWidth div 2 - Width div 2;
    Top := Screen.WorkAreaTop;
  end;
  with ModelBitmap do
  begin
    Height := ModelPaint.Height;
    Width := ModelPaint.Width;
    Canvas.Brush.Color := clWhite;
    Canvas.FillRect(Rect(0, 0, Width, Height));
    Canvas.Pen.Width := 2;
    InitializeModel;
  end;
end;

procedure TModelingForm.FormDestroy(Sender: TObject);
begin
  ModelBitmap.Free;
end;

procedure TModelingForm.FormPaint(Sender: TObject);
begin
  ModelPaint.Canvas.Draw(0, 0, ModelBitmap);
end;

procedure TModelingForm.RefUpDownClick(Sender: TObject;
  Button: TUDBtnType);
begin
  Ref := RefUpDown.Position;
  RunModel;
end;

procedure TModelingForm.GainUpDownClick(Sender: TObject;
  Button: TUDBtnType);
begin
  Gain := GainUpDown.Position/10.0;
  RunModel;
end;

procedure TModelingForm.CharacteristicClick(Sender: TObject);
begin
  CV := Characteristic.ItemIndex;
  RunModel;
end;

procedure TModelingForm.FormShow(Sender: TObject);
begin
  Stage := 3;
end;

end.
