unit Main;

//Program: LiveBlock.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, CheckLst;

type
  TMainForm = class(TForm)
    GraphPanel: TGroupBox;
    DiagramPanel: TGroupBox;
    ControlPanel: TGroupBox;
    GraphSpace: TPaintBox;
    ErrorPanel: TPanel;
    FeedbackPanel: TPanel;
    FeedbackLabel: TLabel;
    SetSpeed: TRadioGroup;
    ShowSliders: TRadioGroup;
    FeedbackCounter: TEdit;
    Panel1: TPanel;
    GraphControlsPanel: TGroupBox;
    GraphSelectPanel: TPanel;
    SetVariablesToGraph: TCheckListBox;
    VariablesToGraphLabel: TLabel;
    ColorKeyPanel: TPanel;
    Shape1: TShape;
    Shape2: TShape;
    Shape3: TShape;
    Shape4: TShape;
    Shape5: TShape;
    Shape6: TShape;
    Shape7: TShape;
    SetScalePanel: TPanel;
    SetScale: TTrackBar;
    ScaleLabel: TLabel;
    ProgramBox: TGroupBox;
    AboutButton: TButton;
    ExitButton: TButton;
    ActionsBox: TGroupBox;
    ResetButton: TButton;
    PauseButton: TButton;
    SetGraphs: TButton;
    Panel2: TPanel;
    InputAdjustPanel: TPanel;
    InputGainLabel: TLabel;
    InOutDelayLabel: TLabel;
    SetInOutDelay: TTrackBar;
    SetInputGain: TTrackBar;
    InputGainCounter: TEdit;
    InOutDelayCounter: TEdit;
    OutputAdjustPanel: TPanel;
    OutputGainLabel: TLabel;
    TimeConstLabel: TLabel;
    SetOutputGain: TTrackBar;
    SetOutputTC: TTrackBar;
    OutputGainCounter: TEdit;
    OutputTCCounter: TEdit;
    DisturbanceAdjustPanel: TPanel;
    DisturbanceLabel: TLabel;
    SetDisturbance: TTrackBar;
    AutoDisturb: TRadioGroup;
    DisturbanceCounter: TEdit;
    FeedbackAdjustPanel: TPanel;
    SetFeedbackGain: TTrackBar;
    FeedbackGainCounter: TEdit;
    InputPanel: TPanel;
    InputLabel: TLabel;
    InputCounter: TEdit;
    OutputPanel: TPanel;
    OutputLabel: TLabel;
    OutputCounter: TEdit;
    PerceptualPanel: TPanel;
    PerceptualSignalLable: TLabel;
    PerceptualCounter: TEdit;
    ReferenceAdjustPanel: TPanel;
    ReferenceLabel: TLabel;
    SetReference: TTrackBar;
    ReferenceCounter: TEdit;
    ComparatorBox: TPanel;
    OutputBox: TPanel;
    FeedbackBox: TPanel;
    InputBox: TPanel;
    ErrorCounter: TEdit;
    ErrorLabel: TLabel;
    Diagram: TPaintBox;
    BoundaryLabel: TLabel;
    CVCircle: TShape;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SetReferenceChange(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure SetDisturbanceChange(Sender: TObject);
    procedure SetInputGainChange(Sender: TObject);
    procedure SetInOutDelayChange(Sender: TObject);
    procedure SetOutputGainChange(Sender: TObject);
    procedure SetOutputTCChange(Sender: TObject);
    procedure ShowSlidersClick(Sender: TObject);
    procedure AutoDisturbClick(Sender: TObject);
    procedure SetFeedbackGainChange(Sender: TObject);
    procedure SetSpeedClick(Sender: TObject);
    procedure SetScaleChange(Sender: TObject);
    procedure InitializeParameters;
    procedure ResetButtonClick(Sender: TObject);
    procedure PauseButtonClick(Sender: TObject);
    procedure AboutButtonClick(Sender: TObject);
    procedure ExitButtonClick(Sender: TObject);
    procedure SetGraphsClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure ZeroDelayBuffers;
    procedure ClearBitmap(BkColor: TColor; Bmp: TBitmap);
    procedure DrawDiagram;
    procedure UpdatePerceptualCounter;
    procedure UpdateErrorCounter;
    procedure UpdateOutputCounter;
    procedure UpdateFeedbackCounter;
    procedure UpdateInputCounter;
    procedure UpdateDisturbanceCounter;
    procedure Initialize;
    function MakeDisturbance: Double;
    function InTransportLag(P: Double; TimeLag: Integer): Double;
    function OutTransportLag(P: double; TimeLag: Integer): Double;

    procedure StepModel;
    procedure UpdateCounters;
    procedure UpdateGraphs;
    procedure IdleLoop(Sender: TObject; var Done: Boolean);
  end;

var
  MainForm: TMainForm;
  ReferenceSignal, Disturbance: Double;
  InputQuantity, PerceptualSignal, ErrorSignal,
    OutputQuantity, FeedbackEffect: Double;
  InputGain, OutputGain, OutputTC, FeedbackGain: Double;
  AutoDisturbance, Fast, Pause, CockPause, GraphsOn: Boolean;
  InDelayBuffer, OutDelayBuffer: array[0..511] of Double;
  Slow, dt, YScale: Double;
  t1, t2, Freq, TimeConstant: Int64;
  InOutDelay, InptRin, OutptRin, InptRout, OutptRout: Integer;
  Count, OldP, OldR, OldE, OldO, OldI, OldD, OldF, Y0: Integer;
  DestRect, SourceRect: TRect;
  Stage: Integer;

implementation

uses About;

{$R *.dfm}

var
  BlockBitmap, GraphBitmap: TBitmap;

procedure TMainForm.ZeroDelayBuffers;
var
  I: Integer;
begin
  for I := 0 to 511 do //Zero the delay buffers
  begin
    InDelayBuffer[I] := 0.0;
    OutDelayBuffer[I] := 0.0;
  end;
  InptRin := 0;
  InptRout := 0;
end;

procedure TMainForm.Initialize;
begin
  TimeConstant     := 60;

  ReferenceSignal  :=  0.0;
  Disturbance      :=  0.0;

  InputQuantity    :=  0.0;
  SetInputGain.Position := 100;
  InputGain := Exp(Ln(SetInputGain.Position) - Ln(100));
  InOutDelay       :=  8;

  PerceptualSignal :=  0.0;
  ErrorSignal      :=  0.0;

  OutputQuantity   :=  0.0;
  OutputGain       := 100.0;
  OutputTC         :=  33.0;

  FeedbackEffect   :=  0.0;
  FeedbackGain     :=  1.0;

  ZeroDelayBuffers;

  //Initialize graph elements
  OldR := 0; OldI := 0; OldE := 0; OldO := 0;
  OldF := 0; OldD := 0;
  YScale := (11 - SetScale.Position)/3.0;

  with GraphBitmap do //Shift graph to left 1 pixel row
  begin
    SourceRect := Rect(1, 0, Width, Height-1);
    DestRect := Rect(0, 0, Width-1, Height-1);
  end;

  UpdatePerceptualCounter;
  UpdateOutputCounter;
  UpdateFeedbackCounter;
  UpdateInputCounter;

  Fast := True;
  Pause := False;

  with SetVariablesToGraph do
  begin
    Checked[0] := True;
    Checked[1] := True;
    Checked[2] := True;
    Checked[4] := True;
    Checked[6] := True;
  end;

  dt := 1/60.0;
  Count := 0;

end;

{$j+}
function TMainForm.MakeDisturbance: Double;
const
  d1: Double = 0;
  d2: Double = 0;
  d3: Double = 0;
begin
  Slow := 0.8;
  d1 := d1 + (1000*(Random - 0.5) - d1)/Slow*dt;
  d2 := d2 + (d1 - d2)/Slow*dt;
  d3 := d3 + (d2 - d3)*dt/Slow;
  Result := d3;
end;
{%J-}

function TMainForm.InTransportLag(P: Double; TimeLag: Integer): Double;
var
  T: Integer;
begin
  if Fast then T := TimeLag else T := 10 * TimeLag;
  InptRin := (InptRin + 1) and 511;
  InDelayBuffer[InptRin] := P;
  OutptRin := (InptRin - T + 512) and 511;
  Result := InDelayBuffer[OutptRin];
end;

function TMainForm.OutTransportLag(P: double; TimeLag: Integer): Double;
var
  T: Integer;
begin
  if Fast then T := TimeLag else T := 10 * TimeLag;
  InptRout := (InptRout + 1) and 511;
  OutDelayBuffer[InptRout] := P;
  OutptRout := (InptRout - t + 512) and 511;
  Result := OutDelayBuffer[OutptRout];
end;

procedure TMainForm.ClearBitmap(BkColor: TColor; Bmp: TBitmap);
var
  BitMapRect: TRect;
begin
  with BitMapRect do
  begin
    Left := 0; Right := Width - 1;
    Top := 0; Bottom := Height - 1;
  end;
  with Bmp.Canvas do
  begin
    Brush.Color := BkColor;
    FillRect(BitMapRect);
  end;
end;

procedure TMainForm.DrawDiagram;
const
  RightArrow = 0;
  LeftArrow = 1;
  UpArrow = 2;
  DownArrow = 3;
var
  X1, X2, Y1, Y2, X0, Y0: Integer;

  procedure DrawArrow(X, Y, Arrow: Integer);
  begin
    with BlockBitmap.Canvas do
    begin
      Brush.Color := clBlack;
      Case Arrow of
        0: //Left
           Polygon([Point(X,Y), Point(X-11, Y-5), Point(X-7, Y),
             Point(X-7, Y), Point(X-11, Y+5)]);
        1: //Right
           Polygon([Point(X,Y), Point(X+11, Y-5), Point(X+7, Y),
             Point(X+7, Y), Point(X+11, Y+5)]);
        2: //Up
           Polygon([Point(X,Y), Point(X-5, Y+11), Point(X, Y+7),
             Point(X, Y+7), Point(X+5, Y+11)]);
        3: //Down
           Polygon([Point(X,Y), Point(X-5, Y-11), Point(X, Y-7),
             Point(X, Y-7), Point(X+5, Y-11)]);
      end;
      Brush.Color := clBtnFace;
    end;
  end;

begin
  ClearBitmap(clBtnFace, BlockBitmap);
  with BlockBitmap.Canvas do
  begin
    X0 := 0;  //Offsets between GroupBox panel and its client area  2
    Y0 := 0; //16;

    //Reference Adjust Panel to Comparator Box
    X1 := ComparatorBox.Left - X0 + ComparatorBox.Width div 2;
    Y1 := ReferenceAdjustPanel.Top - Y0 + ReferenceAdjustPanel.Height;
    MoveTo(X1, Y1);
    X2 := X1;
    Y2 := ComparatorBox.Top - Y0;
    LineTo(X2, Y2);
    DrawArrow(X2, Y2, DownArrow);
    TextOut(X2+7, Y2-25, '+');

    //Comparator Box to Output Box
    X1 := ComparatorBox.Left - X0 + ComparatorBox.Width;
    Y1 := ComparatorBox.Top - Y0 + ComparatorBox.Height div 2;
    MoveTo(X1, Y1);
    X2 := OutputBox.Left - X0 + OutputBox.Width div 2;
    Y2 := Y1;
    LineTo(X2, Y2);
    Y2 := OutputBox.Top - Y0;
    LineTo(X2, Y2);
    DrawArrow(X2, Y2, DownArrow);

    //Output Box to Feedback Box
    X1 := FeedbackBox.Left - X0 + FeedbackBox.Width;
    Y1 := FeedbackBox.Top - Y0 + FeedbackBox.Height div 2;
    MoveTo(X1, Y1);
    X2 := OutputBox.Left - X0 + OutputBox.Width div 2;
    Y2 := Y1;
    LineTo(X2, Y2);
    Y2 := OutputBox.Top - Y0 + OutputBox.Height;
    LineTo(X2, Y2);
    DrawArrow(X1, Y1, LeftArrow);

    //Feedback Box to CV Circle
    X1 := FeedbackBox.Left - X0;
    Y1 := CVCircle.Top - Y0 + CVCircle.Height div 2;
    MoveTo(X1, Y1);
    X2 := CVCircle.Left - X0 + CVCircle.Width;
    Y2 := Y1;
    LineTo(X2, Y2);
    DrawArrow(X2, Y2, LeftArrow);
    TextOut(X2+15, Y2+2, '+');
    Ellipse(216, 336, 233, 353);

    //CV Circle to Input Box
    X1 := CVCircle.Left - X0 + CVCircle.Width div 2;
    Y1 := CVCircle.Top - Y0;
    MoveTo(X1, Y1);
    X2 := X1;
    Y2 := InputBox.Top - Y0 + InputBox.Height;
    LineTo(X2, Y2);
    DrawArrow(X2, Y2, UpArrow);

    //Input Box to Comparator Box
    X1 := InputBox.Left - X0 + InputBox.Width div 2;
    Y1 := InputBox.Top - Y0;
    MoveTo(X1, Y1);
    X2 := X1;
    Y2 := ComparatorBox.Top - Y0 + ComparatorBox.Height div 2;
    LineTo(X2, Y2);
    X2 := ComparatorBox.Left - X0;
    LineTo(X2, Y2);
    DrawArrow(X2, Y2, RightArrow);
    TextOut(X2-20, Y2+1, '_');

    //Disturbance Adjust Panel to CV Circle
    X1 := CVCircle.Left - X0 + CVCircle.Width div 2;
    Y1 := DisturbanceAdjustPanel.Top - Y0;
    MoveTo(X1, Y1);
    X2 := X1;
    Y2 := CVCircle.Top - Y0 + CVCircle.Height;
    LineTo(X2, Y2);
    DrawArrow(X2, Y2, UpArrow);
    TextOut(X2-12, Y2+10, '+');

    //System-environment Boundary
    Pen.Color := clRed;
    Pen.Style := psDash;
    X1 := Left;
    Y1 := BoundaryLabel.Top - Y0 + BoundaryLabel.Height div 2;
    MoveTo(X1, Y1);
    X2 := BoundaryLabel.Left - X0 - 5;
    Y2 := Y1;
    LineTo(X2, Y2);
    Font.Name := 'Arial';
    Font.Color := clRed;
    Font.Size := 10;
    TextOut(X2+3, Y2-7, 'System-Environment Boundary');
    X1 := BoundaryLabel.Left - X0 + BoundaryLabel.Width + 5;
    MoveTo(X1, Y1);
    X2 := Width - X0;
    LineTo(X2, Y2);

    Pen.Color := clBlack;
    Pen.Style := psSolid;
  end;
  with Diagram.Canvas do Draw(0, 0, BlockBitmap);
end;


procedure TMainForm.UpdateCounters;
begin
  UpdatePerceptualCounter;
  UpdateErrorCounter;
  UpdateOutputCounter;
  UpdateFeedbackCounter;
  UpdateInputCounter;
  UpdateDisturbanceCounter;
end;

procedure TMainForm.StepModel;
begin
  FeedbackEffect := FeedbackGain * OutputQuantity;
  if AutoDisturbance then Disturbance := MakeDisturbance
  else Disturbance := SetDisturbance.Position / 10;
  InputQuantity :=  FeedbackEffect + Disturbance;
  PerceptualSignal := InputGain * InTransportLag(InputQuantity, InOutDelay);
  ErrorSignal := ReferenceSignal - PerceptualSignal;
  if ErrorSignal > 1000 then ErrorSignal := 1000
  else
    if ErrorSignal < -1000 then ErrorSignal := -1000;
  OutputQuantity := OutputQuantity
    + (OutputGain*ErrorSignal
    - OutputQuantity) * dt/OutputTC;
  if OutputQuantity > 10000 then OutputQuantity := 10000
  else
    if OutputQuantity < -10000 then OutputQuantity := -10000;
  UpdateCounters;
  UpdateGraphs;
end;

procedure TMainForm.UpdateGraphs;
var
  Y0, Y: integer;

  procedure DrawLine(var Y, OldY: Integer; LineColor: TColor);
  begin
    with GraphBitmap do
    begin
      Canvas.Pen.Color := LineColor;
      Canvas.Pen.width := 4;
      Canvas.MoveTo(Width-2, Y0 - OldY);
      Canvas.LineTo(Width-1, Y0 - Y);
    end;
  end;

begin
    with GraphBitmap do
    begin
      Y0 := Height div 2;
      Canvas.Pen.Width := 4;
      Canvas.CopyRect(DestRect, Canvas, SourceRect);
      Canvas.Pen.Color := clWhite;
      Canvas.MoveTo(Width-1, 0);
      Canvas.LineTo(Width-1, Height-1);
      with SetVariablesToGraph do
      begin
        //Reference signal
        Y := Round(YScale*ReferenceSignal);
        if Checked[0] then Drawline(Y, OldR, clGreen);
        OldR := Y;
        //Perceptual signal
        Y := Round(YScale*PerceptualSignal);
        if Checked[1] then Drawline(Y, OldP, clBlue);
        OldP := Y;
        //Error signal
        Y := Round(YScale*ErrorSignal);
        if Checked[2] then Drawline(Y, OldE, clBlack);
        OldE := Y;
        //Input quantity
        Y := Round(YScale*InputQuantity);
        if Checked[3] then Drawline(Y, OldI, clPurple);
        OldI := Y;
        //Output quantity
        Y := Round(YScale*OutputQuantity);
        if Checked[4] then Drawline(Y, OldO, clOlive);
        OldO := Y;
        //Feedback effect
        Y := Round(YScale*FeedbackEffect);
        If Checked[5] then Drawline(Y, OldF, clAqua);
        OldF := Y;
        //Disturbance
        Y := Round(YScale*Disturbance);
        if Checked[6] then Drawline(Y, OldD, clRed);
        OldD := Y;
      end;
      Y := 0;
      Inc(Count);
      If Count = 60 then
        begin
          Canvas.Pen.Color := clBlack;
          Canvas.MoveTo(Width-1, Y0 + 5);
          Canvas.LineTo(Width-1, Y0 - 5);
          Count := 0;
          if CockPause then
           begin
            Pause := true;
            CockPause := false;
           end;
        end;
      Drawline(Y, Y, clBlack);
      Canvas.Pen.Color := clBlack;
    end;
  GraphSpace.Canvas.Draw(0, 0, GraphBitmap);
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: Boolean);
begin
  Done := False;
  Case Stage of
    0: begin  // Initialization
         with BlockBitmap do
         begin
           Height := Diagram.Height;
           Width := Diagram.Width;
           Canvas.Brush.Color := clBtnFace;
           Canvas.Pen.Color := clBlack;
         end;
         with GraphBitmap do
         begin
           Height := GraphSpace.Height;
           Width := GraphSpace.Width;
           ClearBitmap(clWhite, GraphBitmap);
         end;
         // Center LiveBlock diagram panel on DiagramPanel
         Panel2.Left := DiagramPanel.Width div 2 - Panel2.Width div 2;
         Panel2.Top := DiagramPanel.ClientHeight div 2 - Panel2.Height div 2;
         DrawDiagram;
         Initialize;
         OutputGain := SetOutputGain.Position;
         OutputGainCounter.Text := Format('%6.2f', [OutputGain]);
         OutputTC := SetOutputTC.Position / 10.0;
         OutputTCCounter.Text := Format('%7.1f', [OutputTC]);
         InOutDelay := SetInOutDelay.Position;
         InOutDelayCounter.Text := Format('%7.3f', [inoutdelay*dt]);

         GraphSelectPanel.ShowHint := True;
         Panel1.visible := false;
         GraphsOn := false;
         ShowSlidersClick(sender);
         Stage := 1;
       end;
    1: if not Pause then StepModel;
  end;
  repeat
    QueryPerformanceCounter(t2);
    Application.ProcessMessages;
  until (t2 - t1) >= Freq/60;
  t1 := t2;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  ChangeScale(Screen.Width, 1280);
  WindowState := wsMaximized;
  BlockBitmap := TBitmap.Create;
  GraphBitmap := TBitmap.Create;
  Stage := 0;
  Application.OnIdle := IdleLoop;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(t1);
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  BlockBitmap.Free;
  GraphBitmap.Free;
end;

procedure TMainForm.SetReferenceChange(Sender: TObject);
var
  RefString: String;
begin
  ReferenceSignal := SetReference.Position / 10.0;
  RefString := Format('%7.2f', [ReferenceSignal]);
  ReferenceCounter.Text := RefString;
end;

procedure TMainForm.FormPaint(Sender: TObject);
begin
  DrawDiagram;
  GraphSpace.Canvas.Draw(0, 0, GraphBitmap);
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  with BlockBitmap do
  begin
    Height := Diagram.Height;
    Width := Diagram.Width;
  end;
  with GraphBitmap do
  begin
    Height := GraphSpace.Height;
    Width := GraphSpace.Width;
  end;
  Panel2.Left := DiagramPanel.Width div 2 - Panel2.Width div 2;
end;

procedure TMainForm.SetDisturbanceChange(Sender: TObject);
begin
  Disturbance := SetDisturbance.Position/10.0;
  DisturbanceCounter.Text := Format('%7.2f', [Disturbance]);
end;

procedure TMainForm.SetInputGainChange(Sender: TObject);
begin
  InputGain := exp(ln(SetInputGain.Position) - ln(100));
  InputGainCounter.Text := Format('%7.2f', [InputGain]);
end;

procedure TMainForm.SetInOutDelayChange(Sender: TObject);
begin
  InOutDelay := SetInOutDelay.Position;
  InOutDelayCounter.Text := Format('%7.3f', [InOutDelay*dt]);
end;

procedure TMainForm.SetOutputGainChange(Sender: TObject);
begin
  OutputGain := SetOutputGain.Position;
  OutputGainCounter.Text := Format('%7.1f', [OutputGain]);
end;

procedure TMainForm.SetOutputTCChange(Sender: TObject);
begin
  OutputTC := SetOutputTC.Position / 10.0;
  OutputTCCounter.Text := Format('%7.1f', [OutputTC]);
end;

procedure TMainForm.ShowSlidersClick(Sender: TObject);
begin
  Case ShowSliders.ItemIndex of
    0: begin
         SetInputGain.Show;
         SetInOutDelay.Show;
         SetOutputGain.Show;
         SetOutputTC.Show;
         SetFeedbackGain.Show;
       end;
    1: begin
         SetInputGain.Hide;
         SetInOutDelay.Hide;
         SetOutputGain.Hide;
         SetOutputTC.Hide;
         SetFeedbackGain.Hide;
       end;
    end;
end;

procedure TMainForm.AutoDisturbClick(Sender: TObject);
begin
  Case AutoDisturb.ItemIndex of
    0:  begin
          SetDisturbance.Show;
          AutoDisturbance := False;
          SetDisturbance.Position := Round(Disturbance) * 10;
        end;
    1:  begin
          SetDisturbance.Hide;
          AutoDisturbance := True;
        end;
    end;
end;

procedure TMainForm.SetFeedbackGainChange(Sender: TObject);
begin
  FeedbackGain := SetFeedbackGain.Position / 4.0;
  FeedbackGainCounter.Text := Format('%6.2f', [FeedbackGain]);
end;

procedure TMainForm.UpdatePerceptualCounter;
begin
  PerceptualCounter.Text := Format('%7.2f', [PerceptualSignal]);
end;

procedure TMainForm.UpdateErrorCounter;
begin
  ErrorCounter.Text := Format('%7.2f', [ErrorSignal]);
end;

procedure TMainForm.UpdateOutputCounter;
begin
  OutputCounter.Text := Format('%7.2f', [OutputQuantity]);
end;

procedure TMainForm.UpdateFeedbackCounter;
begin
  FeedbackCounter.Text := Format('%7.2f', [FeedbackEffect]);
end;

procedure TMainForm.UpdateInputCounter;
begin
  InputCounter.Text := Format('%7.2f', [InputQuantity]);
end;

procedure TMainForm.UpdateDisturbanceCounter;
begin
  if AutoDisturbance then
    DisturbanceCounter.Text := Format('%7.2f', [Disturbance]);
end;

procedure TMainForm.InitializeParameters;
begin
  SetinputGain.Position := 100;
  SetInOutDelay.Position := 4;
  SetOutputGain.Position := 100;
  SetOutputTC.Position := 330;
  SetFeedbackGain.Position := 4;
end;

procedure TMainForm.SetSpeedClick(Sender: TObject);
begin
  Case SetSpeed.ItemIndex of
    0: begin
         Fast := True;
         dt := 1/60;
       end;
    1: begin
          Fast := False;
          dt := 1/600;
       end;
  end;
end;

procedure TMainForm.SetScaleChange(Sender: TObject);
begin
  YScale := (11 - SetScale.Position)/3.0;
end;

procedure TMainForm.ResetButtonClick(Sender: TObject);
begin
  Initialize;
  InitializeParameters;
end;

procedure TMainForm.PauseButtonClick(Sender: TObject);
begin
  if Pause then
  begin
    Pause := False;
    PauseButton.Caption := 'Pause';
  end
  else
  begin
    CockPause := True;
    Pause := false;
    PauseButton.Caption := 'Resume';
  end;
end;

procedure TMainForm.AboutButtonClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

procedure TMainForm.ExitButtonClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.SetGraphsClick(Sender: TObject);
begin
 GraphsOn := not GraphsOn;
 if GraphsOn then
  begin
   Panel1.Visible := True;
   SetGraphs.Caption := 'Graphs Off';
  end
 else
  begin
   Panel1.Visible := False;
   SetGraphs.Caption := 'Graphs On';
  end;
end;

end.
