unit AngleFont;

interface

uses Windows, Graphics;

  function CreateAngledFont(Font: HFont; Angle: Longint;
    Quality: byte = PROOF_QUALITY): HFont;

  procedure TextOutA(Canvas: TCanvas; X, Y, Angle: Integer;
    Text: string);


implementation

  function CreateAngledFont(Font: HFont; Angle: Longint;
    Quality: byte): HFont;
  var
    FontInfo: TLogFontA;    // Font information structure
  begin
    // Get the information of the font passed as parameter
    if GetObject(Font, SizeOf(FontInfo), @FontInfo) = 0 then begin
      Result := 0;
      exit;
    end;
    // Set the angle
    FontInfo.lfEscapement := Angle;
    FontInfo.lfOrientation := Angle;
    // Set the quality
    FontInfo.lfQuality := Quality;
    // Create a new font with the modified information
    // The new font must be released calling DeleteObject
    Result := CreateFontIndirect(FontInfo);
  end;

procedure TextOutA(Canvas: TCanvas; X, Y, Angle: Integer;
    Text: string);
  var
    OriginalFont, AngledFont: HFont;
  begin
    // Create an angled font from the current font
    AngledFont := CreateAngledFont(Canvas.Font.Handle, Angle);
    if AngledFont <> 0 then begin
      // Set it temporarily as the current font
      OriginalFont := SelectObject(Canvas.Handle, AngledFont);
      if OriginalFont <> 0 then begin
        // Write the text
        Canvas.TextOut(X, Y, Text);
        // Restore the original font
        if SelectObject(Canvas.Handle, OriginalFont) = 0 then begin
          Canvas.Font.Handle := AngledFont;
          // raise Exception.Create('Couldn''t restore font');
          exit;
        end;
      end;
      // Release the angled font
      DeleteObject(AngledFont)
    end;
  end;


end.
 