unit Main;

//Program: NonAdaptive.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, ComCtrls, Math, jpeg;

type
  TMainForm = class(TForm)
    ControlBox: TGroupBox;
    ProgramBox: TGroupBox;
    ExitBtn: TButton;
    AboutBtn: TButton;
    InstructionBtn: TButton;
    MainPanel: TPanel;
    TimePlotBox: TGroupBox;
    BodePlotsBox: TGroupBox;
    TimePlotPanel: TPanel;
    TimePlotCtrlPanel: TPanel;
    GainBox: TGroupBox;
    LoadParamsBox: TGroupBox;
    PosGainEdit: TEdit;
    Label1: TLabel;
    VelGainEdit: TEdit;
    Label2: TLabel;
    MassEdit: TEdit;
    SpringEdit: TEdit;
    Label3: TLabel;
    DampingEdit: TEdit;
    Label4: TLabel;
    Label5: TLabel;
    Shape1: TShape;
    Label6: TLabel;
    DiagramPanel: TPanel;
    TimePlotPaint: TPaintBox;
    Label11: TLabel;
    Bode1Panel: TPanel;
    Bode1Paint: TPaintBox;
    Bode2Panel: TPanel;
    Bode3Panel: TPanel;
    BodeUpDateBtn: TButton;
    PropLabel: TLabel;
    IntLabel: TLabel;
    DoubleIntLabel: TLabel;
    Bode2Paint: TPaintBox;
    Bode3Paint: TPaintBox;
    MassTrack: TTrackBar;
    SpringTrack: TTrackBar;
    DampingTrack: TTrackBar;
    PosGainTrack: TTrackBar;
    VelGainTrack: TTrackBar;
    TimePlotKey: TGroupBox;
    Shape2: TShape;
    Label7: TLabel;
    Label8: TLabel;
    Shape3: TShape;
    Label9: TLabel;
    Shape4: TShape;
    RefAdjustBox: TGroupBox;
    TargetPosEdit: TEdit;
    TargetTrack: TTrackBar;
    DiagramImage: TImage;
    procedure ExitBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BodeUpDateBtnClick(Sender: TObject);
    procedure MassTrackChange(Sender: TObject);
    procedure SpringTrackChange(Sender: TObject);
    procedure DampingTrackChange(Sender: TObject);
    procedure PosGainTrackChange(Sender: TObject);
    procedure VelGainTrackChange(Sender: TObject);
    procedure TargetTrackChange(Sender: TObject);
    procedure TimePlotPaintPaint(Sender: TObject);
    procedure Bode1PaintPaint(Sender: TObject);
    procedure Bode2PaintPaint(Sender: TObject);
    procedure Bode3PaintPaint(Sender: TObject);
    procedure InstructionBtnClick(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure InitModel;
    procedure PositionPlots;
    procedure RunModel;
    procedure PlotVar(x,y: integer; var oldx,oldy: integer; col: integer);
    procedure RunFreqResp(var InputVar, OutputVar: Real);
    procedure RunSim;
    procedure Iterate;
    procedure IdleLoop(Sender: TObject; var Done: boolean);
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses AngleFont, Instructions, About;

const margin = 250;

var Pos, LastPos, Vel, IntFB, VelFB, Accel: Real;
    SepRef, Target, StepRef, Force, ForceDist: Real;
    OldX, OldPos, OldVel, Phase, OldAccel: Integer;
    OldTarget, OldForceDist: Integer;
    PosGain, VelGain, e3, e2: Real;     // gains
    M, Ma, Ml, S, D: Real;              //mass, spring and damping
    F, T, dt: Real;
    Stage: Integer;
    Iterating: Boolean;
    t1, t2, Freq: Int64;
    TimePlot, Bode1, Bode2, Bode3: TBitmap;
    // Below for Bode Plots
    LeftMargin, RightMargin, TopMargin, BottomMargin, PlotWidth,
      PlotHeight, PlotBottom, PlotRight, FontSize: Integer;

// -------------------- Initialization -------------------

procedure ResetModel;
begin
  Vel := 0.0;
  Pos := 0.0;
  IntFB := 0.0;
  Accel := 0.0;
  SepRef := 0.0;
  Target := 0.0;
  ForceDist := 0.0;
  Ma := 0.2;
  M := Ml + Ma;
end;

procedure TMainForm.InitModel;
begin

  PosGainTrack.Position := 15;
  PosGain := 0.10*PosGainTrack.Position;
  PosGainEdit.Text := Format('%5.1f', [PosGain]);
//  PosGain := 1.5;

  VelGainTrack.Position := 12;
  VelGain := 5.0*VelGainTrack.Position;
  VelGainEdit.Text := Format('%5.1f', [VelGain]);
//  VelGain := 60;

  MassTrack.Position := 0;
  Ma := 0.02;
  Ml := 0.1*MassTrack.Position;
  M := Ma + Ml;
  MassEdit.Text := Format('%5.1f', [Ml]);

//  S := 0.0;
  SpringTrack.Position := 0;
  S := SpringTrack.Position;
  SpringEdit.Text := Format('%5.1f', [S]);

//  D := 0.0;
  DampingTrack.Position := 0;
  D := 0.1*DampingTrack.Position;
  DampingEdit.Text := Format('%5.1f', [D]);

  TargetTrack.Position := -20;
  StepRef := -5*TargetTrack.Position;
  TargetPosEdit.Text := Format('%5.0f', [StepRef]);

  SepRef := 0.0;
  LastPos := 0.0;
  dt := 0.0005;

  OldTarget := 0;
  OldForceDist := 0;
  OldPos := 0;
  OldX := 0;
end;

// -------------- Run Model and Do Time Plot -------------------

procedure TMainForm.RunModel;
begin
  // acceleration = applied force minus velocity and spring feedback, over mass
  Accel := (Force + ForceDist - D*Vel - S*Pos)/M;
  LastPos := Pos;    // save position for computing velocity
  // New position = old position + velocity*dt + 1/2 acceleration*dt-squared
  Pos := Pos + (Vel + 0.5*Accel*dt)*dt;
  Vel := Vel + Accel*dt;   // update velocity; dt is time step per iteration
  VelFB := (Pos - LastPos)/dt;  // compute velocity from delta-pos
  e3 := SepRef - (Pos - Target);  // pos - target is actual separation
  e2 := PosGain * e3 - velfb;  // posgain*e3 is output of position control system
  Force := VelGain * e2;       // which is the velocity reference signal
end;


procedure TMainForm.PlotVar(X, Y: Integer; var OldX, OldY: Integer;
                            Col: Integer);
begin
 with TimePlot.Canvas do
  begin
   Pen.Color := Col;
   Pen.Width := 2;
   MoveTo(OldX, TimePlot.Height div 2 - OldY);
   LineTo(X, TimePlot.Height div 2 - Y);
   OldY := Y;
  end;
end;

// --------- Initialize and Conduct Simulation Run -------------

procedure TMainForm.RunSim;
var VCenter, X, Y: Integer;
begin
  ResetModel;
  LastPos := 0.0;

  OldX := 0;
  OldForceDist := 0;
  OldTarget := 0;
  OldPos := 0;
  OldVel := 0;
  OldAccel := 0;
  with TimePlot.Canvas do
  begin
    Brush.Color := clCream;
    FillRect(Rect(0, 0, TimePlot.Width, TimePlot.Height));
    VCenter := TimePlot.Height div 2;
    Pen.Width := 2;
    Pen.Color := clGray;
    MoveTo(0, VCenter);
    LineTo(TimePlot.Width-1, VCenter);
    for X := 0 to 10*(TimePlot.Width-1) do
    begin
      RunModel;
      if X mod 10 = 0 then
      begin
         if (X div 10) >= 30 then Target := StepRef // StepRef)
          else Target := 0.0;

        Y := Round(Target);
        PlotVar(X div 10, Y, OldX, OldTarget, clBlack);

        Y := Round(Pos);
        PlotVar(X div 10, Y, OldX, OldPos, clRed);

        Y := Round(0.1*Vel);
        PlotVar(X div 10, Y, OldX, OldVel, clLime);

        Y := Round(0.001*Force);
        PlotVar(X div 10, Y, OldX, OldAccel, clBlue);
        OldX := X div 10;
      end;
    end;
  end;
  TimePlotPaint.Canvas.Draw(0, 0, TimePlot);
end;

procedure TMainForm.PositionPlots;
var
  XMargin: Integer;
begin
  DiagramPanel.Height := DiagramPanel.Width;
  XMargin := (BodePlotsBox.ClientWidth - 3*Bode1Paint.Width) div 4;
  Bode1Panel.Left := XMargin;
  PropLabel.Left := Bode1Panel.Left;
  Bode2Panel.Left := XMargin + Bode1Panel.Width + XMargin;
  IntLabel.Left := Bode2Panel.Left;
  Bode3Panel.Left := Bode2Panel.Left + Bode2Panel.Width + XMargin;
  DoubleIntLabel.Left := Bode3Panel.Left;
  TimePlotPanel.Width := TimePlotBox.Width - TimePlotPanel.Left - XMargin;
  TimePlot.Width := TimePlotPaint.Width;
  TimePlot.Height := TimePlotPaint.Height;
  Bode1.Height := Bode1Paint.Height;
  Bode2.Width := Bode2Paint.Width;
  Bode2.Height := Bode2Paint.Height;
  Bode3.Width := Bode3Paint.Width;
  Bode3.Height := Bode3Paint.Height;
end;

// ------------- Make the Bode Plots --------------

procedure MakeBodeFrame(Bitmap: TBitmap);
var
  I, J: Integer;
  XInc, YInc,
  XPos, YPos, HalfTextHeight: integer;
  S: string;
begin
  LeftMargin := 40;
  RightMargin := 40;
  TopMargin := 20;
  BottomMargin := 40;
  PlotRight := Bitmap.Width-1 - RightMargin;
  PlotBottom := Bitmap.Height-1 - BottomMargin;
  PlotWidth := PlotRight - LeftMargin;
  PlotHeight := PlotBottom - TopMargin;
  YInc := PlotHeight div 8;
  with Bitmap.Canvas do
  begin
    Brush.Color := clCream;
    FillRect(Rect(0,0,Bitmap.Width-1,Bitmap.Height-1));
    Pen.Color := clGray;
    Font.Name := 'Arial';
    Font.Size := FontSize;
    Font.Color := clBlack;
    Pen.Color := clGray;
    HalfTextHeight := TextHeight('0') div 2;
    for I := 0 to 8 do // Plot horizontal lines
    begin
      YPos := PlotBottom - I*YInc;
      MoveTo(LeftMargin, YPos);
      LineTo(PlotRight, YPos);
      // Write left Y-axis numerals
      S := Format('%3.0f', [10.0 * I]);
      TextOut(LeftMargin - TextWidth(S) - 5, YPos - HalfTextHeight, S);
      // Write right Y-axis numerals
      case I of
        0:  begin
              S := '-180';
              TextOut(PlotRight+5, YPos - HalfTextHeight, S);
            end;
        2:  begin
              S := ' -90';
              TextOut(PlotRight+5, YPos - HalfTextHeight, S);
           end;
        4:  begin
              S := ' 0';
              TextOut(PlotRight+5, YPos - HalfTextHeight, S);
            end;
        6:  begin
              S := ' 90';
              TextOut(PlotRight+5, YPos - HalfTextHeight, S);
            end;
        8:  begin
              S := '+180';
              TextOut(PlotRight+5, YPos - HalfTextHeight, S);
            end;
      end;
    end;
    Font.Color := clRed;
    S := 'Decibels';
    TextOutA(Bitmap.Canvas, LeftMargin-35, PlotBottom - PlotHeight div 2
      + TextWidth(S) div 2, 900, S);
    Font.Color := clBlue;
    S := 'Degrees';
    TextOutA(Bitmap.Canvas, PlotRight+20, TopMargin
      + PlotHeight div 2 + TextWidth(S) div 2, 900, S);
    // Write X-axis Units
    Font.Color := clBlack;
    S := 'Radians/Second';
    TextOut(LeftMargin + PlotWidth div 2 - TextWidth(S) div 2,
      PlotBottom+20, S);
    // Vertical lines and labels
    XInc := PlotWidth div 3;
    for I := 0 to 3 do  // Plot vertical lines
        for J := 1 to 9 do
        begin
          XPos := LeftMargin + Round(XInc*(Log10(J) + I));
          MoveTo(XPos, YPos);
          LineTo(XPos, PlotBottom);
          if J = 1 then  // Write X-axis numerals
          begin
            S := Format('%3.1f', [Power(10, I-1)]);
            Font.Color := clBlack;
            TextOut(XPos - TextWidth(S) div 2, PlotBottom+5, S);
            if I = 3 then Exit; // Stop after 100.0 line
          end;
        end;
  end;
end;

// Compute and construct Bode plots

procedure TMainForm.RunFreqResp(var InputVar, OutputVar: Real);
var FCount, Kind, X, Y, NCycles: Integer;
    RMSinput, RMSoutput, RMS: Double;
    InputZeroCross, OutputZeroCross: Double;
    OldInputVar, OldOutputVar: Double;
    OldGainX, OldGain, OldPhaseX, OldPhase: Integer;
    BodeBitmap: TBitmap;

  procedure PlotGain;
  begin
    with BodeBitmap.Canvas do
    begin
      Pen.Color := clRed;
      Font.Size := FontSize;
      X := LeftMargin + Round(Log10(10*F*2*Pi)*PlotWidth/3.0);// log frequency
      Y := PlotBottom - Round(20*Log10(RMS)*(PlotHeight)/80.0)+2;// log amplitude
      Font.Color := clRed;
      if (FCount = 1) or (FCount = 13) then
        TextOut(X+ 10, Y-15, FloatToStrF(RMS,Fffixed,4,1));
      MoveTo(OldGainX, OldGain);
      if FCount > 1 then LineTo(X, Y);
      OldGainX := X;
      OldGain := Y;
    end;
  end;

  procedure PlotPhase;
  begin
    with BodeBitmap.Canvas do
    begin
      Pen.Color := clBlue;
      // FREQUENCY * DELTA-T = PHASE SHIFT, RADIANS
      Phase := Round(F*(OutputZeroCross - InputZeroCross)*360.0);
      if Phase > 180 then Phase := Phase - 360;
      // X is computed in PlotGain before calling PlotPhase
      Y := PlotBottom - PlotHeight div 2 + Round(Phase * PlotHeight/360)+2;
        //minus changed to plus in above line, 5/8/08
      Font.Color := clBlue;
      if (Fcount = 1) or (Fcount = 13) then
        TextOut(X-5, Y - 15,IntToStr(Phase));
      MoveTo(OldPhaseX, OldPhase);
      if FCount > 1 then LineTo(X, Y);
      OldPhaseX := X;
      OldPhase := Y;
    end;
  end;

begin
  Iterating := False;  // stop idle time interrupts
  ResetModel;
  for Kind := 0 to 2 do
  begin
    case Kind of
      0: begin  // Proportional (Spring only)
           D := 0; Ml := 0.0; S := 250;
           BodeBitmap := Bode1;
         end;
      1: begin // Integral (Damper only)
           D := 100; Ml := 0.0; S := 0;
           BodeBitmap := Bode2;
         end;
      2: begin // Double-Integral (Mass Only)
           D := 0; Ml := 20.0; S := 0;
           BodeBitmap := Bode3;
         end;
      end;
    MakeBodeFrame(BodeBitmap);
    with BodeBitmap.Canvas do
    begin
      Pen.Width := 2;
      Font.Size := 9;
    end;
    OldGainX := 0;
    OldGain := 0;
    OldPhaseX := 0;
    OldPhase := 0;

    M := Ma + Ml; // arm mass plus load mass
    F := 0.1/2.0/Pi;  // initial frequency = 0.1 Hz
    OutputZeroCross := 0.0;
    InputZeroCross := 0.0;
    for Fcount := 1 to 13 do
    begin
      T := 0.0;
      RMSinput := 0.0;
      RMSoutput := 0.0;
      Vel := 0.0;
      ForceDist := 0.0;
      Target := 0.0;
      Pos := 0.0;
      OldInputVar := 0.0;
      OldOutputVar := 0.0;

      NCycles := 0;

      //BEGIN FREQUENCY RESPONSE MEASUREMENT
      while Ncycles < 8 do
      begin
        Target := 200*Sin(2*Pi*F*T);
        RunModel;
        RMSinput := RMSinput + Sqr(InputVar);
        RMSoutput := RMSoutput + Sqr(OutputVar);
        if Ncycles = 7 then // find zero crossings (independent of amplitude)
        begin
          if (InputVar < 0) and (OldInputVar >= 0) then InputZeroCross := T;
          OldInputVar := InputVar;
          if (OutputVar < 0) and (OldOutputVar >= 0) then OutputZeroCross := T;
          OldOutputVar := OutputVar;
        end;
        T := T + dt; // advance time by dt
        if F*T >= 1.0 then T := 0.0;  // wrap around at 2 pi radians
        if T = 0 then Inc(Ncycles);
      end;   // END FREQUENCY RESPONSE MEASUREMENT

      //PLOT AMPLITUDE RATIO VS FREQUENCY
      RMS := Sqrt(RMSoutput/RMSinput); // amplitude ratio
      PlotGain;
      // PLOT PHASE SHIFT VERSUS FREQUENCY
      PlotPhase;

      F := F * Exp(Ln(130)/12); // Step by 12th root of 130
    end; // END OF FREQUENCY RESPONSE ITERATIONS.
    with BodeBitmap.Canvas do
    begin
      Pen.Width := 1;
//      Font.Size := 10;
      Pen.Color := clBlack;
    end;
  end; // kind
  Ml := 0.1*MassTrack.Position;
  M := Ma + Ml;
  S := SpringTrack.Position;
  D := 0.1*DampingTrack.Position;
  Iterating := True;
end;

// ------------------ Operate Main Program Loop ---------------

procedure TMainForm.Iterate;
begin
  case Stage of
    0: begin
         InitModel;
         PositionPlots;
         MakeBodeFrame(Bode1);
         MakeBodeFrame(Bode2);
         MakeBodeFrame(Bode3);
         Bode1Paint.Canvas.Draw(0, 0, Bode1);
         Bode2Paint.Canvas.Draw(0, 0, Bode2);
         Bode3Paint.Canvas.Draw(0, 0, Bode3);
         Stage := 1;
       end;
    1: RunSim;
    end;
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: boolean);
begin
  Done := False;
  if not Iterating then Exit;
  Iterate;
  repeat
    QueryPerformanceCounter(t2);
  until t2 - t1 > Freq/4;
  t1 := t2;
end;

// -- Create and Resize Main Form and Bitmaps; Destroy Main Form ---

procedure TMainForm.FormCreate(Sender: TObject);
begin
  ChangeScale(Screen.Height, 768);
  WindowState := wsMaximized;
  TimePlot := TBitmap.Create;
  Bode1 := TBitmap.Create;
  Bode2 := TBitmap.Create;
  Bode3 := TBitmap.Create;
  TimePlot.Width := TimePlotPaint.Width;
  TimePlot.Height := TimePlotPaint.Height;
  Bode1.Width := Bode1Paint.Width;
  Bode1.Height := Bode1.Height;
  Bode2.Width := Bode2Paint.Width;
  Bode2.Height := Bode2.Height;
  Bode3.Width := Bode3Paint.Width;
  Bode3.Height := Bode3.Height;
  if Screen.Height < 700 then FontSize := 8 else FontSize := 10;
  Stage := 0;
  Iterating := True;
  T := 0;
  Pos := 0.0;
  Vel := 0.0;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(t1);
  Application.OnIdle := IdleLoop;
 end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Bode3.Free;
  Bode2.Free;
  Bode1.Free;
  TimePlot.Free;
end;

// -------------- Program Buttons ------------------

procedure TMainForm.InstructionBtnClick(Sender: TObject);
begin
  InstructForm.Show;
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Application.Terminate;
end;

// ----------- Bode Plot Update Button -----------------

procedure TMainForm.BodeUpDateBtnClick(Sender: TObject);
begin
  RunFreqResp(e3, force);
  Bode1Paint.Canvas.Draw(0, 0, Bode1);
  Bode2Paint.Canvas.Draw(0, 0, Bode2);
  Bode3Paint.Canvas.Draw(0, 0, Bode3);
end;

// --------------- Trackbar Controls -------------------

procedure TMainForm.PosGainTrackChange(Sender: TObject);
begin
  PosGain := 0.10*PosGainTrack.Position;
  PosGainEdit.Text := Format('%5.1f', [PosGain]);
end;

procedure TMainForm.VelGainTrackChange(Sender: TObject);
begin
  VelGain := 5.0*VelGainTrack.Position;
  VelGainEdit.Text := Format('%5.1f', [VelGain]);
end;

procedure TMainForm.MassTrackChange(Sender: TObject);
begin
  Ml := 0.1*MassTrack.Position;
  M := Ma + Ml;
  MassEdit.Text := Format('%5.1f', [Ml]);
end;

procedure TMainForm.SpringTrackChange(Sender: TObject);
begin
  S := SpringTrack.Position;
  SpringEdit.Text := Format('%5.1f', [S]);
end;

procedure TMainForm.DampingTrackChange(Sender: TObject);
begin
  D := 0.1*DampingTrack.Position;
  DampingEdit.Text := Format('%5.1f', [D]);
end;

procedure TMainForm.TargetTrackChange(Sender: TObject);
begin
  StepRef := -5*TargetTrack.Position;
  TargetPosEdit.Text := Format('%5.0f', [StepRef]);
end;

// ---------------- Plot Repainting -----------------

procedure TMainForm.TimePlotPaintPaint(Sender: TObject);
begin
    TimePlotPaint.Canvas.Draw(0, 0, TimePlot);
end;

procedure TMainForm.Bode1PaintPaint(Sender: TObject);
begin
  Bode1Paint.Canvas.Draw(0, 0, Bode1);
end;

procedure TMainForm.Bode2PaintPaint(Sender: TObject);
begin
  Bode2Paint.Canvas.Draw(0, 0, Bode2);
end;

procedure TMainForm.Bode3PaintPaint(Sender: TObject);
begin
  Bode3Paint.Canvas.Draw(0, 0, Bode3);
end;

end.

