unit Main;

//Program: LiveThree.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ExtCtrls, Instructions, About;

type
  TMainForm = class(TForm)
    ControlBox: TGroupBox;
    ProgramBox: TGroupBox;
    InstructionBtn: TButton;
    AboutBtn: TButton;
    ExitBtn: TButton;
    LiveBlockBox: TGroupBox;
    YRefTrackBar: TTrackBar;
    RRefTrackBar: TTrackBar;
    GRefTrackBar: TTrackBar;
    D3TrackBar: TTrackBar;
    D2TrackBar: TTrackBar;
    D1TrackBar: TTrackBar;
    YOut1Edit: TEdit;
    YOut2Edit: TEdit;
    YOut3Edit: TEdit;
    RIn2Edit: TEdit;
    RIn1Edit: TEdit;
    RIn3Edit: TEdit;
    ROut2Edit: TEdit;
    ROut1Edit: TEdit;
    ROut3Edit: TEdit;
    GIn2Edit: TEdit;
    GIn1Edit: TEdit;
    GIn3Edit: TEdit;
    GOut2Edit: TEdit;
    GOut1Edit: TEdit;
    GOut3Edit: TEdit;
    RInPanel: TPanel;
    GInPanel: TPanel;
    YOutPanel: TPanel;
    ROutPanel: TPanel;
    GOutPanel: TPanel;
    YCompPanel: TPanel;
    RCompPanel: TPanel;
    GCompPanel: TPanel;
    YInPanel: TPanel;
    YIn1Edit: TEdit;
    YIn2Edit: TEdit;
    YIn3Edit: TEdit;
    YPEdit: TEdit;
    YEEdit: TEdit;
    YREdit: TEdit;
    RPEdit: TEdit;
    RREdit: TEdit;
    REEdit: TEdit;
    GPEdit: TEdit;
    GREdit: TEdit;
    GEEdit: TEdit;
    EV1Panel: TPanel;
    EV2Panel: TPanel;
    EV3Panel: TPanel;
    EV1Edit: TEdit;
    EV2Edit: TEdit;
    EV3Edit: TEdit;
    D3Edit: TEdit;
    D2Edit: TEdit;
    D1Edit: TEdit;
    YQiEdit: TEdit;
    RQiEdit: TEdit;
    GQiEdit: TEdit;
    YQoEdit: TEdit;
    RQoEdit: TEdit;
    GQoEdit: TEdit;
    WeightsBox: TGroupBox;
    NewWeightsBtn: TButton;
    Shape21: TShape;
    LiveBlockPaint: TPaintBox;
    YRefLabel: TStaticText;
    RRefLabel: TStaticText;
    GRefLabel: TStaticText;
    YRLabel: TStaticText;
    RRLabel: TStaticText;
    YPLabel: TStaticText;
    YELabel: TStaticText;
    RPLabel: TStaticText;
    YQiLabel: TStaticText;
    YQoLabel: TStaticText;
    RQiLabel: TStaticText;
    RQoLabel: TStaticText;
    RELabel: TStaticText;
    GPLabel: TStaticText;
    GQiLabel: TStaticText;
    GQoLabel: TStaticText;
    GELabel: TStaticText;
    GRLabel: TStaticText;
    D3Label: TStaticText;
    Dist3Label: TStaticText;
    D2Label: TStaticText;
    Dist2Label: TStaticText;
    D1Label: TStaticText;
    Dist1Label: TStaticText;
    procedure ExitBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure NewWeightsBtnClick(Sender: TObject);
    procedure YRefTrackBarChange(Sender: TObject);
    procedure RRefTrackBarChange(Sender: TObject);
    procedure GRefTrackBarChange(Sender: TObject);
    procedure D1TrackBarChange(Sender: TObject);
    procedure D2TrackBarChange(Sender: TObject);
    procedure D3TrackBarChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure LiveBlockPaintPaint(Sender: TObject);
    procedure InstructionBtnClick(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure IdleLoop(Sender: TObject; var Done: Boolean);
    procedure InitPointers;
    procedure PositionControls;
    procedure DrawLines;
    procedure DisplayWeights;
    procedure UpdateDisplays;
    procedure Iterate;
  end;

var
  MainForm: TMainForm;
  T1, T2, Freq: int64;
  Step: Integer;
  dt: Double;
  CSystem: array[0..2] of
            record
             P, R, E, Qo, D, Gain: Double;
             Wi, Wo: array[0..2] of Double;
            end;
  EV: array[0..2] of Double;
  RefTrackBar, DTrackBar: array[0..2] of ^TTrackbar;
  RefLabel, RLabel, PLabel, ELabel, QiLabel, QoLabel, DLabel, DistLabel:
    array[0..2] of ^TStaticText;
  REdit, PEdit, EEdit, QiEdit, QoEdit, EVEdit, DEdit: array[0..2] of ^TEdit;
  CompPanel, InPanel, OutPanel, EVPanel: array[0..2] of ^TPanel;
  WiEdit, WoEdit: array[0..2] of array[0..2] of ^TEdit;
  Bitmap1: TBitmap;

implementation

{$R *.dfm}

const
  Fmt = '%5.1f'; // Format for non-weight edit windows

procedure TMainForm.InitPointers;
begin
  RefLabel[0] := @YRefLabel; RefLabel[1] := @RRefLabel;
  RefLabel[2] := @GRefLabel;
  RLabel[0] := @YRLabel; RLabel[1] := @RRLabel; RLabel[2] := @GRLabel;
  PLabel[0] := @YPLabel; PLabel[1] := @RPLabel; PLabel[2] := @GPLabel;
  ELabel[0] := @YELabel; ELabel[1] := @RELabel; ELabel[2] := @GELabel;
  RefTrackBar[0] := @YRefTrackBar; RefTrackBar[1] := @RRefTrackBar;
  RefTrackBar[2] := @GRefTrackBar;
  REdit[0] := @YREdit; REdit[1] := @RREdit; REdit[2] := @GREdit;
  CompPanel[0] := @YCompPanel; CompPanel[1] := @RCompPanel;
  CompPanel[2] := @GCompPanel;
  InPanel[0] := @YInPanel; InPanel[1] := @RInPanel; InPanel[2] := @GInPanel;
  OutPanel[0] := @YOutPanel; OutPanel[1] := @ROutPanel;
  OutPanel[2] := @GOutPanel;
  QiEdit[0] := @YQiEdit; QiEdit[1] := @RQiEdit; QiEdit[2] := @GQiEdit;
  QoEdit[0] := @YQoEdit; QoEdit[1] := @RQoEdit; QoEdit[2] := @GQoEdit;
  PEdit[0] := @YPEdit; PEdit[1] := @RPEdit; PEdit[2] := @GPEdit;
  EEdit[0] := @YEEdit; EEdit[1] := @REEdit; EEdit[2] := @GEEdit;
  PLabel[0] := @YPLabel; PLabel[1] := @RPLabel; PLabel[2] := @GPLabel;
  ELabel[0] := @YELabel; ELabel[1] := @RELabel; ELabel[2] := @GELabel;
  QiLabel[0] := @YQiLabel; QiLabel[1] := @RQiLabel; QiLabel[2] := @GQiLabel;
  QoLabel[0] := @YQoLabel; QoLabel[1] := @RQoLabel; QoLabel[2] := @GQoLabel;
  WiEdit[0,0] := @YIn1Edit; WiEdit[0,1] := @YIn2Edit; WiEdit[0,2] := @YIn3Edit;
  WiEdit[1,0] := @RIn1Edit; WiEdit[1,1] := @RIn2Edit; WiEdit[1,2] := @RIn3Edit;
  WiEdit[2,0] := @GIn1Edit; WiEdit[2,1] := @GIn2Edit; WiEdit[2,2] := @GIn3Edit;
  WoEdit[0,0] := @YOut1Edit; WoEdit[0,1] := @YOut2Edit;
  WoEdit[0,2] := @YOut3Edit;
  WoEdit[1,0] := @ROut1Edit; WoEdit[1,1] := @ROut2Edit;
  WoEdit[1,2] := @ROut3Edit;
  WoEdit[2,0] := @GOut1Edit; WoEdit[2,1] := @GOut2Edit;
  WoEdit[2,2] := @GOut3Edit;
  EVPanel[0] := @EV1Panel; EVPanel[1] := @EV2Panel; EVPanel[2] := @EV3Panel;
  EVEdit[0] := @EV1Edit; EVEdit[1] := @EV2Edit; EVEdit[2] := @EV3Edit;
  DTrackBar[0] := @D1TrackBar; DTrackBar[1] := @D2TrackBar;
  DTrackBar[2] := @D3TrackBar;
  DEdit[0] := @D1Edit; DEdit[1] := @D2Edit; DEdit[2] := @D3Edit;
  DLabel[0] := @D1Label; DLabel[1] := @D2Label; DLabel[2] := @D3Label;
  DistLabel[0] := @Dist1Label; DistLabel[1] := @Dist2Label;
  DistLabel[2] := @Dist3Label;
end;

procedure TMainForm.PositionControls;
var
  OneThird, YSpace: Integer;
  XCenter: array[0..2] of Integer;
  I, X, Y: Integer;
begin
  with LiveBlockBox do
  begin
    YSpace := LiveBlockBox.ClientHeight div 32;
    OneThird := LiveBlockBox.ClientWidth div 3;
    XCenter[0] := OneThird div 2;
    XCenter[1] := XCenter[0] + OneThird;
    XCenter[2] := XCenter[1] + OneThird;
    for I := 0 to 2 do
    begin
      RefTrackBar[I].Left := XCenter[I] - RefTrackBar[I].Width div 2;
      RefTrackBar[I].Top := YSpace * 2;

      RefLabel[I].Left := XCenter[I] - RefLabel[I].Width div 2;
      RefLabel[I].Top := RefTrackBar[I].Top - RefLabel[I].Height;

      REdit[I].Left := XCenter[I] - REdit[I].Width div 2;
      REdit[I].Top := RefTrackBar[I].Top + RefTrackBar[I].Height;

      RLabel[I].Left := REdit[I].Left + REdit[I].Width + 5;
      RLabel[I].Top := REdit[I].Top + 5;

      CompPanel[I].Left := XCenter[I] - CompPanel[I].Width div 2;
      CompPanel[I].Top := YSpace*7;

      InPanel[I].Left := CompPanel[I].Left - InPanel[I].Width;
      InPanel[I].Top := YSpace*9;

      OutPanel[I].Left := CompPanel[I].Left + CompPanel[I].Width;
      OutPanel[I].Top := YSpace*9;

      QiEdit[I].Left := InPanel[I].Left + InPanel[I].Width div 2
        - QiEdit[I].Width div 2;
      QiEdit[I].Top := InPanel[I].Top + InPanel[I].Height;

      QoEdit[I].Left := OutPanel[I].Left + OutPanel[I].Width div 2
        - QoEdit[I].Width div 2;
      QoEdit[I].Top := OutPanel[I].Top + OutPanel[I].Height;

      PEdit[I].Left := QiEdit[I].Left;
      PEdit[I].Top := CompPanel[I].Top - PEdit[I].Height;

      PLabel[I].Left := PEdit[I].Left - PLabel[I].Width;
      PLabel[I].Top := PEdit[I].Top + 5;

      EEdit[I].Left := QoEdit[I].Left;
      EEdit[I].Top := PEdit[I].Top;

      ELabel[I].Left := EEdit[I].Left + EEdit[I].Width + 5;
      ELabel[I].Top := PLabel[I].Top;

      QiLabel[I].Left := PLabel[I].Left;
      QiLabel[I].Top := QiEdit[I].Top + 5;

      QoLabel[I].Left := ELabel[I].Left;
      QoLabel[I].Top := QiLabel[I].Top;

      WiEdit[I,1].Left := QiEdit[I].Left;
      Y := YSpace * 14;
      WiEdit[I,1].Top := Y;
      WiEdit[I,0].Left := WiEdit[I,1].Left - WiEdit[I,0].Width;
      WiEdit[I,0].Top := Y;
      WiEdit[I,2].Left := WiEdit[I,1].Left + WiEdit[I,1].Width;
      WiEdit[I,2].Top := Y;

      WoEdit[I,1].Left := QoEdit[I].Left;
      WoEdit[I,1].Top := Y;
      WoEdit[I,0].Left := WoEdit[I,1].Left - WoEdit[I,0].Width;
      WoEdit[I,0].Top := Y;
      WoEdit[I,2].Left := WoEdit[I,1].Left + WoEdit[I,1].Width;
      WoEdit[I,2].Top := Y;

      EVPanel[I].Left := WiEdit[0,I].Left;
      EVEdit[I].Left := EVPanel[I].Left;
    end;
    EVPanel[0].Top := YSpace * 17;
    EVPanel[1].Top := YSpace * 19;
    EVPanel[2].Top := YSpace * 21;
    // Locate disturbance trackbars
    DTrackBar[0].Left := WiEdit[0,2].Left + WiEdit[0,2].Width div 2
      - DTrackBar[0].Width div 2;
    DTrackBar[0].Top := YSpace * 30;
    DTrackBar[1].Left := WoEdit[0,2].Left + WoEdit[0,2].Width div 2
      - DTrackBar[1].Width div 2;
    DTrackBar[1].Top := YSpace * 28;
    DTrackBar[2].Left := WiEdit[1,2].Left + WiEdit[1,2].Width div 2
      - DTrackBar[2].Width div 2;
    DTrackBar[2].Top := YSpace * 26;
    for I := 0 to 2 do
    begin
      EVEdit[I].Top := EVPanel[I].Top + EVPanel[I].Height;
      DEdit[I].Left := DTrackBar[I].Left + DTrackBar[I].Width div 2
        - DEdit[I].Width div 2;
      DEdit[I].Top := DTrackBar[I].Top - DEdit[I].Height;
      DLabel[I].Left := DEdit[I].Left + DEdit[I].Width + 5;
      DLabel[I].Top := DEdit[I].Top + 5;
      DistLabel[I].Left := DTrackBar[I].Left + DTrackBar[I].Width div 2
        - DistLabel[I].Width div 2;
      DistLabel[I].Top := DTrackBar[I].Top + DTrackBar[I].Height;
    end;
  end;
end;

procedure TMainForm.DrawLines;
const
  Big = 15;
var
  I, J, X0, Y0, X1, Y1, Xo, Yo, Y: Integer;
  EdBoxH, EdBoxW, EdBoxMidH, EdBoxMidW: Integer;
  Arrow: array[0..2] of TPoint;

  procedure VertArrow(XX, YY, Size: Integer);  // use -size to flip
  begin
    with Arrow[0] do begin X := XX; Y := YY; end; // tip of arrow
    with Arrow[1] do begin X := XX - Round(0.35*Size); Y := YY + Size; end;
    with Arrow[2] do begin X := XX + Round(0.35*Size); Y := YY + Size; end;
    with Bitmap1.Canvas do Polygon(Arrow);
  end;

  procedure HorizArrow(XX, YY, Size: Integer);  // use -size to flip
  begin
    with Arrow[0] do begin X := XX; Y := YY; end;   // tip of arrow
    with Arrow[1] do begin X := XX - Size; Y := YY + Round(0.35*Size); end;
    with Arrow[2] do begin X := XX - Size; Y := YY - Round(0.35*Size); end;
    with Bitmap1.Canvas do Polygon(Arrow);
    end;

begin
  EdBoxW := YREdit.Width;
  EdBoxMidW := EdBoxW div 2;
  EdBoxH := YREdit.Height;
  EdBoxMidH := EdBoxH div 2;
  Bitmap1.Canvas.FillRect(Rect(0, 0, Width-1, Height-1));
  with Bitmap1.Canvas do
  begin
    Pen.Width := 2;
    Xo := 2;
    Yo := 18;
    // ------- System 1 --------
    for I := 0 to 2 do
    begin
      Case I of
        0: Pen.Color := clYellow;
        1: Pen.Color := clRed;
        2: Pen.Color := clGreen;
        end;
      Brush.Color := Pen.Color;
      // Ref to Comp
      X0 := CompPanel[I].Left + CompPanel[I].Width div 2 - Xo;
      Y0 := REdit[I].Top + EdBoxH - Yo;
      Y1 := CompPanel[I].Top - Yo;
      MoveTo(X0, Y0); LineTo(X0, Y1);
      VertArrow(X0, Y0, Big);
      // Input to Comp
      X0 := InPanel[I].Left+InPanel[I].Width div 2 - Xo;
      Y0 := InPanel[I].Top - Yo;
      X1 := CompPanel[I].Left - Xo;
      Y1 := CompPanel[I].Top+CompPanel[I].Height div 2 - Yo;
      MoveTo(X0, Y0); LineTo(X0, Y1); LineTo(X1, Y1);
      HorizArrow(X1, Y1, Big);
      // Comp to Output
      X1 := X1 + CompPanel[I].Width;
      X0 := OutPanel[I].Left + OutPanel[I].Width div 2 - Xo;
      Y0 := YOutPanel.Top -Yo;
      MoveTo(X1, Y1); LineTo(X0, Y1); LineTo(X0, Y0);
      VertArrow(X0, Y0, -Big);
      X0 := QiEdit[I].Left + EdBoxMidW - Xo;
      Y0 := QiEdit[I].Top + EdBoxH - Yo;
      VertArrow(X0, Y0, Big);
      for J := 0 to 2 do //Input and Output Weights
      begin
        // Weights to Input
        if  J = 1 then Y := Y0 else Y := Y0 + Big + 5;
        X1 := WiEdit[I,J].Left + EdBoxMidW - Xo;
        Y1 := WiEdit[I,J].Top - Yo;
        MoveTo(X0, Y); LineTo(X1, Y1);
      end;
      // Output to Weights
      X0 := QoEdit[I].Left + EdBoxMidW - Xo;
      Y0 := QoEdit[I].Top + EdBoxH - Yo;
      VertArrow(X0, Y0+Big+5, -Big);
      for J := 0 to 2 do
      begin
        if J = 1 then Y := Y0 else Y := Y0 + Big + 5;
        X1 := WoEdit[I,J].Left + EdBoxMidW - Xo;
        Y1 := WoEdit[I,J].Top - Yo;
        MoveTo(X0, Y); LineTo(X1, Y1);
      end;
    end;
    Y := 3;
    for I := 0 to 2 do
    begin
      Case I of
        0: Pen.Color := clYellow;
        1: Pen.Color := clRed;
        2: Pen.Color := clGreen;
        end;
      Brush.Color := Pen.Color;
      for J := 0 to 2 do // EVs to Qi
      begin
        X0 := WiEdit[I,J].Left + EdBoxMidW - Xo;
        Y0 := WiEdit[I,J].Top + EdBoxH - Yo;
        if I = 0 then // System 1 input weights connect from top of EV box
          begin
            Y1 := EVPanel[J].Top - Yo;
            MoveTo(X0, Y0); LineTo(X0, Y1);
            VertArrow(X0, Y0, Big);
          end
        else
          begin // Remainder connect to side of EV box
            X1 := EVPanel[J].Left + EV1Panel.Width - Xo;
            Y1 := EVPanel[J].Top + Y - Yo;
            MoveTo(X0, Y0); LineTo(X0, Y1); LineTo(X1, Y1);
            VertArrow(X0, Y0, Big);
          end;
      end;
      for J := 0 to 2 do // Qo to EVs
      begin
        X0 := WoEdit[I, J].Left + EdBoxMidW - Xo;
        Y0 := WoEdit[I, J].Top + EdBoxH - Yo;
        Y1 := EVPanel[J].Top + Y+5 - Yo;
        MoveTo(X0, Y0); LineTo(X0, Y1);
        LineTo(EVPanel[J].Left + EVPanel[J].Width - Xo, Y1);
        VertArrow(X0, Y0+Big+5, -Big);
      end;
      Y := Y + 13;
      // DEdit to EVEdit
      Pen.Color := clBlack;
      Brush.Color := clBlack;
      X0 := DEdit[I].Left - Xo;
      Y0 := DEdit[I].Top + EdBoxMidH - Yo;
      X1 := EVEdit[I].Left + EdBoxMidW - Xo;
      Y1 := EVEdit[I].Top + EDBoxH - Yo;
      MoveTo(X0, Y0); LineTo(X1, Y0); LineTo(X1, Y1);
      VertArrow(X1, Y1, Big);
    end;
    Brush.Color := clMoneyGreen;
  end;
end;

procedure MakeWeights;
var
  I, J: Integer;
  SumSq: Double;
begin
  for I := 0 to 2 do
  begin
    for J := 0 to 2 do CSystem[I].Wi[J] := 2*(Random - 0.5);
    sumsq := 0.0;
    for J := 0 to 2 do SumSq := SumSq + Sqr(CSystem[I].Wi[J]);
    sumsq := sqrt(sumsq);
    for J := 0 to 2 do
      CSystem[I].Wi[J] := CSystem[I].Wi[J]/SumSq; // Normalize vector length to 1
  end;
  for i := 0 to 2 do
    for j := 0 to 2 do
      with CSystem[I] do Wo[J] := Wi[J];
end;

procedure TMainForm.DisplayWeights;
const
  WtFmt = '%5.3f';
var
  I, J: Integer;
begin
  for I := 0 to 2 do
    with CSystem[I] do
      for J := 0 to 2 do
        begin
          WiEdit[I, J].Text := Format(WtFmt, [Wi[J]]);
          WoEdit[I, J].Text := Format(WtFmt, [Wo[J]]);
        end;
end;

procedure RunControlSys;
var
  Sys, Wt: Integer;
begin
  // COMPUTE PERCEPTUAL SIGNALS, ERROR SIGNALS, OUTPUT SIGNALS
  for Sys := 0 to 2 do
    with CSystem[Sys] do
    begin
      P := 0.0;
      for Wt := 0 to 2 do P := P + Wi[Wt]*EV[Wt];
      P := P + D;
      E := R - P;
      Qo := Qo + 40.0*E*dt;
      if Qo > 2000 then Qo := 2000
      else
        if Qo < -2000 then Qo := -2000;
    end;
    // COMPUTE EFFECTS ON ENVIRONMENTAL VARIABLES
    for Wt := 0 to 2 do
    begin
      EV[Wt] := 0.0;
      for Sys := 0 to 2 do
        with CSystem[Sys] do EV[Wt] := EV[Wt] + Qo*Wo[Wt];
    end;
end;

procedure TMainForm.UpdateDisplays;
var
  I: Integer;
begin
  for I := 0 to 2 do
  begin
    with CSystem[I] do
    begin
      PEdit[I].Text := Format(Fmt, [P]);
      REdit[I].Text := Format(Fmt, [R]);
      EEdit[I].Text := Format(Fmt, [E]);
      QoEdit[I].Text := Format(Fmt, [Qo]);
      QiEdit[I].Text := Format(Fmt, [P]);
      DEdit[I].Text := Format(Fmt, [D]);
    end;
    EvEdit[I].Text := Format(Fmt, [Ev[I]]);
  end;
end;

procedure TMainForm.Iterate;
begin
  case Step of
    0: begin
         //Initialize
         with Bitmap1 do
         begin
           Height := LiveBlockPaint.Height;
           Width := LiveBlockPaint.Width;
           Canvas.Brush.Color := clMoneyGreen;
         end;
         Randomize;
         MakeWeights;
         DisplayWeights;
         PositionControls;
         DrawLines;
         LiveBlockPaint.Canvas.Draw(0, 0, Bitmap1);
         Step := 1;
       end;
    1: begin
         RunControlSys;
         UpdateDisplays;
       end;
     end;
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: Boolean);
begin
  Done := False;
  Iterate;
  repeat
    QueryPerformanceCounter(T2);
  until (T2 - T1) >= Freq/60;
  T1 := T2;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  ChangeScale(Screen.Width, 1024);
  WindowState := wsMaximized;
  Bitmap1 := TBitmap.Create;
  InitPointers;
  Application.OnIdle := IdleLoop;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(T1);
  dt := 1/60;
  Step := 0;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Bitmap1.Free;
end;

procedure TMainForm.LiveBlockPaintPaint(Sender: TObject);
begin
  LiveBlockPaint.Canvas.Draw(0, 0, Bitmap1);
end;

procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.NewWeightsBtnClick(Sender: TObject);
begin
  MakeWeights;
  DisplayWeights;
end;

procedure TMainForm.YRefTrackBarChange(Sender: TObject);
begin
  CSystem[0].R := YRefTrackBar.Position/100.0;
  YREdit.Text := Format(Fmt, [CSystem[0].R]);
end;

procedure TMainForm.RRefTrackBarChange(Sender: TObject);
begin
  CSystem[1].R := RRefTrackBar.Position/100.0;
  RREdit.Text := Format(Fmt, [CSystem[1].R]);
end;

procedure TMainForm.GRefTrackBarChange(Sender: TObject);
begin
  CSystem[2].R := GRefTrackBar.Position/100.0;
  GREdit.Text := Format(Fmt, [CSystem[2].R]);
end;

procedure TMainForm.D1TrackBarChange(Sender: TObject);
begin
  CSystem[0].D := D1TrackBar.Position/100.0;
  D1Edit.Text := Format(Fmt, [CSystem[0].D]);
end;

procedure TMainForm.D2TrackBarChange(Sender: TObject);
begin
  CSystem[1].D := D2TrackBar.Position/100.0;
  D2Edit.Text := Format(Fmt, [CSystem[1].D]);
end;

procedure TMainForm.D3TrackBarChange(Sender: TObject);
begin
  CSystem[2].D := D3TrackBar.Position/100.0;
  D3Edit.Text := Format(Fmt, [CSystem[2].D]);
end;

procedure TMainForm.InstructionBtnClick(Sender: TObject);
begin
  InstructForm.Show;
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

end.
