unit Main;

//Program: Ecoli
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, ComCtrls, StdCtrls;

type
  TMainForm = class(TForm)
    ControlPanel: TGroupBox;
    Panel1: TPanel;
    StatusBar1: TStatusBar;
    GroupBox2: TGroupBox;
    ActionsPanel: TGroupBox;
    Options: TRadioGroup;
    OptionTabs: TTabControl;
    ProgramPanel: TGroupBox;
    AboutButton: TButton;
    ExitButton: TButton;
    EcoliDisplay: TPaintBox;
    TumbleBtn: TButton;
    GroupBox1: TGroupBox;
    SpeedBar: TTrackBar;
    Instructions: TMemo;
    procedure ExitButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure TumbleBtnClick(Sender: TObject);
    procedure OptionsClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure OptionTabsChange(Sender: TObject);
    procedure SpeedBarChange(Sender: TObject);
    procedure AboutButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure IdleLoop(Sender: TObject; var Done: boolean);
    procedure Initialize;
    procedure DrawGoal;
    procedure NewSetUp;
    procedure Iterate;
    procedure ClearBitMap(BkColor: TColor; Bmp: TBitmap);
    procedure Tumble;
  end;

var
  MainForm: TMainForm;
  DisplayBitmap: TBitmap;
  LineBitmap: TBitmap;
  Sprite: TBitmap;

implementation

uses About;

{$R *.dfm}

var
  GoalX, GoalY, EcolX, EcolY, Theta, Ds: Double;
  Distance, LastDistance, Angle: Double;
  OldX, OldY, Stage, Scale, Degree, Count: Integer;
  DisplayWidth, DisplayHeight, GoalRadius: Integer;
  TotalTumbles, NumTumbles, TotalHits, TimeCount: Longint;
  T1, T2, Freq, Speed: Int64;
  Running, Tumbling: Boolean;

procedure TMainForm.Initialize;
begin
  Stage := 0;
  Running := false;
  Tumbling := false;
  Degree := 0;
  Scale := 10;
  Count := 0;
  Randomize;
  Speed := 200;
  Distance := 0;
  LastDistance := 0;
  Options.ItemIndex := 0;
  Ds := 1.0;
end;

procedure TMainForm.ClearBitmap(BkColor: TColor; Bmp: TBitmap);
var
  BitMapRect: TRect;
begin
  with BitMapRect do
  begin
    Left := 0; Right := Width - 1;
    Top := 0; Bottom := Height - 1;
  end;
  with Bmp.Canvas do
  begin
    Brush.Color := BkColor;
    FillRect(BitMapRect);
  end;
end;

procedure TMainForm.DrawGoal;
begin
  with DisplayBitmap.Canvas do
  begin
    Pen.Color := clBlack;
    Ellipse(Rect(Round(GoalX - GoalRadius), Round(GoalY - GoalRadius),
            Round(GoalX + GoalRadius), Round(GoalY + GoalRadius)));
    Pen.Color := clRed;
  end;
end;

procedure TMainForm.NewSetUp;
begin
  with DisplayBitmap.Canvas do
  begin
    Brush.Color := clWhite; //clBtnFace;
    Pen.Width := 2;
    DisplayWidth := EcoliDisplay.Width;
    DisplayHeight := EcoliDisplay.Height;
    TumbleBtn.Left := DisplayWidth div 2 - (TumbleBtn.Width div 2);
    TumbleBtn.Top := DisplayHeight - TumbleBtn.Height;
    GoalX := DisplayWidth div 2;
    GoalY := DisplayHeight div 2;
    EcolX := Random(DisplayWidth);
    EcolY := Random(DisplayHeight);
    Theta := Random*2*pi;
    ClearBitmap(clWhite, DisplayBitmap);
    GoalRadius := DisplayHeight div 40;
    DrawGoal;
  end;
  EcoliDisplay.Canvas.Draw(0, 0, DisplayBitmap);
  TimeCount := 0;
end;

procedure TMainForm.Tumble;
  procedure DrawBug(Color, Xcenter, Ycenter: Integer; Angle: Real);
  var
    Xdev, Ydev: Integer;
   begin
    Scale := 10;
    Xdev := Round(Scale*Cos(Angle));
    Ydev := Round(Scale*Sin(Angle));
    with Sprite.Canvas do
    begin
      Pen.Color := Color;
      Polyline([Point(5-Xdev, 5-Ydev), Point(5+Xdev, 5+Ydev)]);
    end;
    EcoliDisplay.Canvas.Draw(XCenter-5, YCenter-5, Sprite);
  end;

begin
  DrawBug(clWhite, Round(EcolX), Round(EcolY), Angle);
  Degree := Degree + 5;
  if Degree > 359 then Degree := 0;
  Angle := Degree*(2*Pi)/360; //convert angle to radians
  DrawBug(clRed, Round(EcolX), Round(EcolY), Angle);
  Inc(Count);
  if (Count > 200) then
  begin
    Tumbling := False;
    Count := 0;
    DrawBug(clWhite, Round(EcolX), Round(EcolY), Angle);
  end;
end;

procedure TMainForm.Iterate;
begin
  OldX := Round(EcolX);
  OldY := Round(EcolY);
  Distance := Sqrt(Sqr(EcolX - GoalX) + Sqr(EcolY - GoalY));
  case Options.ItemIndex of
    0: begin  //manual reorganization
       end;
    1: begin  //automatic reorganization
         dec(TimeCount);
         if TimeCount <= 0 then
         begin
           TimeCount := 5;
           if Distance > LastDistance then //tumble
           begin
             Theta := Random * 2 * Pi; // Select new direction at random
             Inc(NumTumbles);
             Tumbling := True;
           end;
           LastDistance := Distance;
         end;
       end;
     2: begin  //evolution
          EcolX := Random(DisplayWidth);
          EcolY := Random(DisplayHeight);
          Inc(NumTumbles);
        end;
     end; {case Options of}
  if Options.ItemIndex < 2 then
  begin
    EcolX := EcolX + Ds*Cos(Theta);
    if EcolX >= DisplayWidth then EcolX := 0
    else if EcolX < 0 then EcolX := DisplayWidth;
    EcolY := EcolY + Ds*Sin(Theta);
    if EcolY >= DisplayHeight then EcolY := 0
    else if EcolY < 0 then EcolY := DisplayHeight;
  end;
  if Distance <= GoalRadius then //start new run
  begin
    Inc(TotalHits);
    TotalTumbles := TotalTumbles + NumTumbles;
    NumTumbles := 0;
    NewSetup;
  end;
  with DisplayBitmap.Canvas do
  begin
    //Draw ecoli point
    Ellipse(Rect(Round(EcolX-2), Round(EcolY-2),
            Round(EcolX+1), Round(EcolY+1)));
  end;
  if TotalHits > 0 then
  StatusBar1.Panels.Items[1].Text :=
  FloatToStrF(TotalTumbles/TotalHits, fffixed, 6, 2);
  with EcoliDisplay.Canvas do Draw(0, 0, DisplayBitmap);
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: Boolean);
begin
  Done := False;
  Case Stage of
    0: begin
         NewSetUp;
         Stage := 1;
         with Sprite do
         begin
           Width := 10;
           Height := 10;
           Canvas.Pen.Width := 3;
         end;
       end;
    1: begin
         if Running then
           if Tumbling then Tumble else Iterate;
       end;
    end;
  repeat
   QueryPerformanceCounter(t2);
   Application.ProcessMessages;
  until (T2 - T1) >= Freq/Speed;
  T1 := T2;   
end;

procedure TMainForm.ExitButtonClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  ChangeScale(Screen.Width, 1280); //Scale to screen res
  QueryPerformanceCounter(T1);
  QueryPerformanceFrequency(Freq);
  Application.OnIdle := IdleLoop;
  WindowState := wsMaximized;
  DisplayBitmap := TBitmap.Create;
  Sprite := TBitmap.Create;
  Initialize;
end;

procedure TMainForm.TumbleBtnClick(Sender: TObject);
begin
 Inc(NumTumbles);
 Theta := Random*2*Pi;
 Tumbling := True;
end;

procedure TMainForm.OptionsClick(Sender: TObject);
begin
  TotalTumbles := 0; NumTumbles := 0; TotalHits := 0;
  Case Options.ItemIndex of
    0: begin
         TumbleBtn.Show;
         Instructions.Show;
         Speed := 200;
       end;
    1: begin
         TumbleBtn.Hide;
         Instructions.Hide;
         Speed := 200;
       end;
    2: begin
         TumbleBtn.Hide;
         Instructions.Hide;
         Speed := 100;
       end;
    end;
  NewSetUp;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  NewSetup;
end;

procedure TMainForm.OptionTabsChange(Sender: TObject);
begin
  Case OptionTabs.TabIndex of
    0:  begin
          Running := True;
        end;
    1:  begin
          Running := False;
        end;
    2:  begin
          Running := False;
          NewSetUp;
          TotalHits := 0;
          TotalTumbles := 0;
          StatusBar1.Panels.Items[1].Text :=
          FloatToStrF(0, fffixed, 6, 2);
        end;
    end;
end;

procedure TMainForm.SpeedBarChange(Sender: TObject);
begin
  Speed := SpeedBar.Position;
end;

procedure TMainForm.AboutButtonClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Sprite.Free;
  DisplayBitmap.Free;
end;

procedure TMainForm.FormPaint(Sender: TObject);
begin
  with DisplayBitmap do
  begin
    Width := EcoliDisplay.Width;
    Height := EcoliDisplay.Width;
  end;
  TumbleBtn.Left := EcoliDisplay.Width div 2 - TumbleBtn.Width div 2;
end;

end.
