unit Main;

//Program: ThreeSys.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, ComCtrls;

type
  TMainForm = class(TForm)
    ControlBox: TGroupBox;
    ProgramBox: TGroupBox;
    ExitBtn: TButton;
    AboutBtn: TButton;
    InstructBtn: TButton;
    MainPanel: TPanel;
    BoxDisplay: TGroupBox;
    ThreeDPaintBox: TPaintBox;
    BoxControlPanel: TPanel;
    PatternRadio: TRadioGroup;
    RefControlPanel: TPanel;
    SetRefX: TTrackBar;
    SetRefY: TTrackBar;
    Label6: TLabel;
    Label4: TLabel;
    SetRefZ: TTrackBar;
    Label5: TLabel;
    ViewAdjustPanel: TPanel;
    Panel3: TPanel;
    GroupBox1: TGroupBox;
    GroupBox2: TGroupBox;
    Panel4: TPanel;
    ReorgPaintBox: TPaintBox;
    RMSErrEdit: TEdit;
    IterEdit: TEdit;
    ReorgEdit: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Panel2: TPanel;
    RotateX: TTrackBar;
    RotateY: TTrackBar;
    RotateZ: TTrackBar;
    ChangeSize: TTrackBar;
    Label7: TLabel;
    CubeView: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    KeyBox: TGroupBox;
    ViewResetBtn: TButton;
    Panel1: TPanel;
    Label12: TLabel;
    Label13: TLabel;
    Label14: TLabel;
    Panel5: TPanel;
    WeightsPaintBox: TPaintBox;
    NewWtsPanel: TPanel;
    WeightsBtn: TButton;
    DetTrackBar: TTrackBar;
    Label15: TLabel;
    DetRangeLabel: TLabel;
    SpeedRadio: TRadioGroup;
    PauseBtn: TButton;
    ReorgRadio: TRadioGroup;
    Label16: TLabel;
    WeightRadio: TRadioGroup;
    procedure FormCreate(Sender: TObject);
    procedure ExitBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RotateXChange(Sender: TObject);
    procedure RotateYChange(Sender: TObject);
    procedure RotateZChange(Sender: TObject);
    procedure ViewResetBtnClick(Sender: TObject);
    procedure ChangeSizeChange(Sender: TObject);
    procedure WeightsBtnClick(Sender: TObject);
    procedure DetTrackBarChange(Sender: TObject);
    procedure ReorgRadioClick(Sender: TObject);
    procedure PauseBtnClick(Sender: TObject);
    procedure SpeedRadioClick(Sender: TObject);
    procedure WeightRadioClick(Sender: TObject);
    procedure WeightsPaintBoxPaint(Sender: TObject);
    procedure ReorgPaintBoxPaint(Sender: TObject);
    procedure ThreeDPaintBoxPaint(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
    procedure InstructBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure InitBitmaps;
    procedure InitProgram;
    procedure SetNewWeights;
    procedure Plot3D(X, Y, Z, Distance: Double; C: Tcolor);
    procedure ShowWeights;
    procedure ShowFrame;
    procedure PlotLogScale;
    procedure ClearLogPlot;
    procedure ShowReorg;
    procedure UpDateReorgCounters;
    procedure PlotVariables;
    procedure Lissajous;
    procedure Axes;
    procedure PosControl;
    procedure Idleloop(Sender: TObject; var Done: Boolean);
    procedure Iterate;
    procedure SetCtheta;
    procedure SetCphi;
    procedure SetCrho;
    procedure ResetView;
    procedure ShowWeightLabels;
  end;

  procedure InitWi;
  procedure InitWo;
  function DetWi: Double;
  function DetWo: Double;
  procedure Reorganize;
  procedure Control;

var
  MainForm: TMainForm;

implementation

uses About, Instructions;

{$R *.dfm}

const
  LastCS = 2;
  LastEV = 2;
  CubeS = 225;
type
  SystemType = record
                 p, r, e, Qo, Qi, g, Ks, Fb: double;
                 Wi, Wo: array[0..LastEV] of double;
               end;

var
  Ctheta, Stheta, Cphi, Sphi, Crho, Srho: Double;
  Angle, Distance, Z0, MinDet, dt: Double;
  ReorgGain, GlobalErSq, TotalErSq: Double;
  HCenter, VCenter, HCent, Stage, X, CycleNumber: Integer;
  ThreeDBitmap, WtBitmap, ReorgBitmap: TBitmap;
  Iters, Reorgs: LongWord;
  CS: array[0..LastCS] of SystemType;
  EV: array[0..LastEV] of Double;
  Correction: array[0..2, 0..2] of Double;
  LocalErSq, LastErr: array[0..LastCS] of Double;
  GlobalReorg, ReorgOK, WasReorg, Fast, Pause,
    CycleDone, ShowInputWeights: Boolean;
  Freq, t1, t2: Int64;

procedure InitWi;
var
  I, J: Integer;
  Sum: Double;
begin
  for I := 0 to LastCS do
  begin
    Sum := 0.0;
    for J := 0 to LastEV do
      with CS[I] do
      begin
        Wi[J] := 2.0*(Random - 0.5);
        Sum := Sum + Sqr(Wi[J]);  // get sum of squares
      end;
    Sum := Sqrt(Sum);
    for J := 0 to LastEV do
      with CS[I] do
        Wi[J] := Wi[J]/Sum;   // normalize to unit vector length
  end;
end;

procedure InitWo;
var
  I, J: Integer;
begin
  for I := 0 to LastCS do
    for J := 0 to LastEv do
      CS[I].Wo[J] := 0.0;
end;

procedure InitVars;
var
  I, J: Integer;
begin
  MinDet := 0.5;
  WasReorg := False;
  GlobalReorg := False;
  Fast := False;
  CycleDone := False;
  Pause := False;
  X := 0;
  Z0 := 0;
  for I := 0 to LastCS do
    for J := 0 to LastEV do
      Correction[I,J] := 2.0*(Random - 0.5);
  for I := 0 to LastCS do
    with CS[I] do
    begin
      Ks := 0.1;
      G := 600.0;
      dt := 1/g;
      Qo := 0.0;
    end;
  ReorgGain := 2e-6;
  InitWi;
  InitWo;
end;

procedure TMainForm.InitBitmaps;
begin
  with ThreeDBitmap do
  begin
    Width := ThreeDPaintBox.Width;
    Height := ThreeDPaintBox.Height;
    Vcenter := Height div 2;
    Hcenter := Width div 2;
    Canvas.Brush.Color := clGray;
  end;
  with WtBitmap do
  begin
    Width := WeightsPaintBox.Width;
    Height := WeightsPaintBox.Height;
    HCent := Width div 2;
    with Canvas do
    begin
      Font.Name := 'Arial';
      Font.Color := clBlack;
      Font.Size := 10;
      Brush.Color := clMoneyGreen;
      FillRect(Rect(0, 0, WtBitmap.Width, WtBitmap.Height));
    end;
  end;
  with ReorgBitmap do
  begin
    Width := ReorgPaintBox.Width;
    Height := ReorgPaintBox.Height;
  end;
end;

procedure TMainForm.InitProgram;
begin
  Randomize;
  ShowFrame;
  ShowWeights;
  Angle := 0;
  Iters := 0;
  Reorgs := 0;
  Distance := 1000;
end;

  // Plot point in 3D space, rotated by theta, phi, and rho
  // Ctheta, Stheta, Cphi, Sphi, Crho, Srho
  // are global sines, cosines of angles.

procedure Calc3D(X, Y, Z, Distance: Double; var X2d, Y2d: Integer);
var
  Xp, Yp, Zp, Temp: Double;
begin
 // X AXIS
 Yp := Ctheta*Y + Stheta*(Z);     // sines and cosines of rotation angles
 Zp := -Stheta*Y + Ctheta*(Z);    // are set globally.
 // Z AXIS
 Xp := Crho*x + Srho*Yp;
 Yp := -Srho*x + Crho*Yp;    // temp is saved starting value of xp
 // Y AXIS
 Temp := Xp;
 Xp := Cphi*Xp + Sphi*Zp;
 Zp := -Sphi*Temp + Cphi*Zp;
 if Zp <> -Distance then         // project into 2-D perspective
 begin                           // " distance" is viewing distance
  X2d := Hcenter + (Round(Xp*Distance/(Distance + Zp + Z0)))div 2;
  Y2d := Vcenter - (Round(Yp*Distance/(Distance + Zp + Z0))) div 2;
 end
 else begin X2d := 0; Y2d := 0; end;
end;

procedure TMainForm.Plot3D(X, Y, Z, Distance: Double; C: Tcolor);
var
  Xp, Yp, Zp, Temp: Double;
  X2d, Y2d, D: Integer;
begin
  // X AXIS
  Yp := Ctheta*Y + Stheta*(Z);     // sines and cosines of rotation angles
  Zp := -Stheta*Y + Ctheta*(Z);    // are set globally.
  // Z AXIS
  Xp := Crho*X + Srho*Yp;
  Yp := -Srho*X + Crho*Yp;    // temp is saved starting value of xp
  // Y AXIS
  Temp := Xp;
  Xp := Cphi*Xp + Sphi*Zp;
  Zp := -Sphi*Temp + Cphi*Zp;
  if Zp <> -Distance then         // project into 2-D perspective
  begin                           // " distance" is viewing distance
    X2d := Hcenter + (Round(Xp*Distance/(Distance + Zp + Z0)))div 2;
    Y2d := Vcenter - (Round(Yp*Distance/(Distance + Zp + Z0))) div 2;
  end
  else begin X2d := 0; Y2d := 0; end;
  with ThreeDBitmap.Canvas do
  begin
    Pen.Color := C;
    D := 1;
    if (X2d >= 0) and (X2d < 2*Hcenter) and
     (Y2d >= 0) and (Y2d < 2*Vcenter) then
     Ellipse(X2d-D, y2d-D, X2d+D, Y2d+D);  // nice fat point
  end;
end;

procedure TMainForm.ShowFrame;
var
  X2d, Y2d: Integer;
begin
  with ThreeDBitmap.Canvas do
  begin
    Pen.Color := clBlack;
    Pen.Width := 2;
    // Top Front
    Calc3d(-CubeS, CubeS, -CubeS, Distance, X2d, Y2d);
    MoveTo(X2d, Y2d);
    Calc3d( CubeS, CubeS, -CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    //Right Front;
    Calc3d(CubeS, -CubeS, -CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Bottom Front
    Calc3d(-CubeS, -CubeS, -CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Left Front
    Calc3d( -CubeS, CubeS, -CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);

    // Back Top
    Calc3d(-CubeS, CubeS, CubeS, Distance, X2d, Y2d);
    MoveTo(X2d, Y2d);
    Calc3d( CubeS, CubeS, CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Back Right
    Calc3d(CubeS, -CubeS, CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Back Bottom
    Calc3d(-CubeS, -CubeS, CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Back Left
    Calc3d(-CubeS,  CubeS, CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Bottom Left
    Calc3d(-CubeS, -CubeS,  CubeS, Distance, X2d, Y2d);
    MoveTo(X2d, Y2d);
    Calc3d(-CubeS, -CubeS, -CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Bottom Right
    Calc3d(CubeS, -CubeS, -CubeS, Distance, X2d, Y2d);
    MoveTo(X2d, Y2d);
    Calc3d(CubeS, -CubeS,  CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
    // Bottom Left
    Calc3d(-CubeS, CubeS, -CubeS, Distance, X2d, Y2d);
    MoveTo(X2d, Y2d);
    Calc3d(-CubeS, CubeS,  CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);

    Calc3d(CubeS, CubeS, -CubeS, Distance, X2d, Y2d);
    MoveTo(X2d, Y2d);
    Calc3d(CubeS, CubeS,  CubeS, Distance, X2d, Y2d);
    LineTo(X2d, Y2d);
  end;
end;

procedure TMainForm.ShowWeights;
var
  I, J: Integer;
  S, Temp: string;
begin
  with WtBitmap.Canvas do
  begin
    if Iters mod 20 = 0 then
    begin
      HCent := WtBitmap.Width div 2;
      Temp :=  'Determinant = '
        + FloatToStrF(DetWi, Fffixed, 5, 2);
      TextOut(40, 0, Temp);
    end;
    for I := 0 to LastCS do
      for J := 0 to LastEv do
      begin
        if ShowInputWeights then  S := Format('%8.3f     ', [CS[I].WI[J]])
        else S := Format('%8.3f     ', [Correction[I, J]]);
        TextOut(130+70*J, 40+15*I, S);
        S := Format('%8.3f   ', [CS[I].WO[J]]);
        TextOut(130+70*J, 110+15*I, S);
      end;
  end;
end;

procedure TMainForm.PlotLogScale;
var
  I, Y: Integer;
  K, ScaleLabel: Double;
  S: String;
begin
  ScaleLabel := 1.0;
  with ReorgBitmap, Canvas do
  begin
    Pen.Color := clSilver;
    S := 'Total RMS Error';
    TextOut(Width div 2 - TextWidth(S) div 2, 3, S);
    Font.Color := clRed;
    S := 'Red ticks indicate reorganizations';
    TextOut(Width div 2 - TextWidth(S) div 2 , Height-TextHeight(S)-3, S);
    Font.Color := clBlack;
    K := (Height-40)/(20.0*Ln(1000));
    for I := 1 to 30 do
    begin
      case I of
         1:     Y := 0;
         2..9:  Y := Round(K * 20*(Ln(I)));
        10..19: Y := Round(K * 20*Ln(  10 * (I -  9)));
        20..29: Y := Round(K * 20*Ln( 100 * (I - 19)));
        30:     Y := Round(K * 20*Ln(1000 * (I - 29)));
      end;
      MoveTo(0, Height - 20 - Y);
      LineTo(Width-40, Height - 20 - Y);
      if (I = 1) or (I mod 10 = 0) then
      begin
        TextOut(Width-45 + 10, Height - 20 - Y - 8,
          FloatToStrF(ScaleLabel, Fffixed, 5, 1));
        ScaleLabel := ScaleLabel * 10.0;
      end;
    end;
  end;
end;

procedure TMainForm.ClearLogPlot;
begin
  X := 0;
  with ReorgBitmap, Canvas do FillRect(Rect(0, 0, Width, Height));
  PlotLogScale;
end;

procedure TMainForm.ShowReorg;
var
  XX, YY: Integer;
  K: Real;
begin
  with ReorgBitmap, Canvas do
  begin                     // plot log of global error
    K := (Height-40)/(20*Ln(1000));
    Pen.Color := clBlack;
    if TotalErSq < 0.001 then TotalErSq := 0.001;
    YY := Height - 20 - Round(K*20*Ln(Sqrt(TotalErSq)));
    XX := X div 3;
    Ellipse(XX-1,YY-1,XX+1,YY+1);
    Inc(X);
    if XX > Width-40 then ClearLogPlot;
    if WasReorg then    // show red tick for at least one reorganization
    begin
      Pen.Color := clRed;
      MoveTo(XX, YY); LineTo(XX, YY-10);
      Inc(Reorgs);
      WasReorg := False;
    end;
  end;
end;

procedure TMainForm.UpDateReorgCounters;
begin
  RMSErrEdit.Text := Format('%8.2f', [Sqrt(TotalErSq)]);
  IterEdit.Text := Format('%8d', [Iters]);
  ReorgEdit.Text := Format('%8d', [Reorgs]);
end;

procedure Reorganize;
var
  Sys, Evar: Integer;
begin
  TotalErSq := GlobalErSq*dt; // save total error for use in adding correction
  if not GlobalReorg then
  begin
    GlobalErSq := 0;
    for Sys := 0 to LastCS do GlobalErSq := GlobalErSq + LocalErSq[Sys];
  end;
  GlobalErSq := Sqrt(GlobalErSq*dt); // error radius
  if GlobalReorg then // reorganize all output weights as one matrix
  begin
    if GlobalErSq >= LastErr[0] then
    begin
      for Sys := 0 to LastCS do
        for Evar := 0 to LastEV do
          Correction[Sys, Evar] := 2.0*(Random - 0.5);
      WasReorg := True;
    end;
    LastErr[0] := GlobalErSq;
  end
  else               // reorganize each system independently
  begin
    for Sys := 0 to LastCS do
    begin
      if LocalErSq[Sys] >= LastErr[Sys] then
      begin
        for Evar := 0 to LastEV do
          Correction[Sys, Evar] := 2.0*(Random - 0.5);
        WasReorg := True;
        LastErr[Sys] := LocalErSq[Sys];
      end;
      LastErr[Sys] := LocalErSq[Sys];
    end;
  end;
  GlobalErSq := 0.0;
  for Sys := 0 to LastCS do LocalErSq[Sys] := 0.0;
  ReorgOK := False;   // allow reorganizations only after complete cycle
end;

procedure Control;
var
  MaxErSq: Double;
  Sys, Evar: Integer;
begin
// FIRST, ADD CORRECTIONS TO ALL OUTPUT WEIGHTS

  // limit amount of correction. totalersq is always positive
  if TotalErSq > 20 then MaxErSq := 20 else MaxErSq := TotalErSq;
 // Add "swimming direction" matrix to output weights;
  for Sys := 0 to LastCS do
    with CS[Sys] do
      for Evar := 0 to LastEV do
        Wo[Evar] := Wo[Evar] + Correction[Sys, Evar]*MaxErSq*ReorgGain;

 // COMPUTE EACH CONTROL SYSTEM'S PERCEPTION, ERROR, AND OUTPUT QUANTITY
  for Sys := 0 to LastCS do
    with CS[Sys] do
    begin
      P := 0;
      for Evar := 0 to LastEv do p := p + Wi[Evar]*EV[Evar];
      e := r - p;
      Qo := Qo + Ks*(g*e - Qo)*dt;
      if Qo > 1000 then Qo := 1000   //Limit output quantity to prevent overflows
      else
        if Qo < -1000 then Qo := -1000;
    end;

 // COMPUTE VALUE OF EACH ENVIRONMENTAL VARIABLE.
  for Evar := 0 to LastEV do
  begin
    EV[Evar] := 0;
    for Sys := 0 to LastCS do
      with CS[Sys] do EV[Evar] := EV[Evar] + Qo*Wo[Evar];
  end;

 // Accumulate error squared until ReorgOK is true;
  for Sys := 0 to LastCS do GlobalErSq := GlobalErSq + Sqr(CS[Sys].e);
  for Sys := 0 to LastEV do LocalErSq[Sys] := LocalErSq[Sys] + Sqr(CS[Sys].e);
  if ReorgOK then Reorganize;
end;

// DETERMINANT OF ALL INPUT WEIGHTS
function DetWi: Double;
var
  S: Double;
begin
  S := 0.0;
  with CS[0] do
    S := S + CS[0].Wi[0]*
      (CS[1].Wi[1]*CS[2].Wi[2] - CS[1].Wi[2]*CS[2].Wi[1]);
  s := S - CS[0].Wi[1]*
    (CS[1].Wi[0]*CS[2].Wi[2] - CS[2].Wi[0]*CS[1].Wi[2]);
  S := S + CS[0].Wi[2]*
    (CS[1].Wi[0]*CS[2].Wi[1] - CS[2].Wi[0]*CS[1].Wi[1]);
  Result := S;
end;

// DETERMINANT OF ALL OUTPUT WEIGHTS
function DetWo: Double;
var
  S: Double;
begin
  S := 0.0;
  with CS[0] do
    S := S + CS[0].Wo[0]*
      (CS[1].Wo[1]*CS[2].Wo[2] - CS[1].Wo[2]*CS[2].Wo[1]);
  S := S - CS[0].Wo[1]*
    (CS[1].Wo[0]*CS[2].Wo[2] - CS[2].Wo[0]*CS[1].Wo[2]);
  S := S + CS[0].Wo[2]*
    (CS[1].Wo[0]*CS[2].Wo[1] - CS[2].Wo[0]*CS[1].Wo[1]);
  Result := S;
end;

procedure TMainForm.PlotVariables;
begin
 Plot3D(CS[0].r, CS[1].r, CS[2].r, Distance, clWhite);
 Plot3D(CS[0].p, CS[1].p, CS[2].p, Distance, clYellow);
 Plot3D(EV[0], EV[1], EV[2], Distance, clAqua);
end;

procedure TMainForm.Lissajous;
begin
  CS[0].r := 125.0*Sin(2*Angle);
  CS[1].r := 125.0*Sin(Angle);
  CS[2].r := 125.0*Cos(Angle);
  Angle := Angle + 2*PI*dt;
  CycleDone := Angle >= 2*PI*(1.0-dt);
end;

procedure TMainForm.Axes;
begin
  if Angle < 2*PI * (1.0-dt) then
    CS[0].r := 125.0*Sin(Angle)
  else if Angle < 4.0*PI*(1-dt) then
      CS[1].r := 125.0*Sin(Angle)
    else
      CS[2].r := 125.0*Sin(Angle);
  Angle := Angle + 2*PI*dt;
  CycleDone := Angle >= 6*PI*(1.0-dt);
 end;

procedure TMainForm.PosControl;
begin
  CS[0].r := SetRefX.Position;
  CS[1].r := SetRefY.Position;
  CS[2].r := SetRefZ.Position;
  Angle := Angle + 2*PI*dt;
  CycleDone := Angle >= 2*PI*(1.0-dt);
end;

procedure TMainForm.SetNewWeights;
var
  Temp: Double;
begin
  Iters := 0;
  repeat
    InitWi;
    Temp := DetWi;
  until (Temp >= MinDet) and (Temp < (MinDet + 0.09)) ;
  InitProgram;
  InitWo;
end;

procedure TMainForm.Iterate;
begin
  case Stage of
    1: begin
         InitBitmaps;
         Randomize;
         ShowInputWeights := True;
         ShowWeightLabels;
         SetNewWeights;
         PlotLogScale;
         Iters := 0;
         Reorgs := 0;
         TotalErSq := 1e6;
         CycleNumber := 0;
           ThreeDPaintBox.Canvas.Draw(0, 0, ThreeDBitmap);
           with ThreeDBitmap.Canvas do
             FillRect(Rect(0, 0, ThreeDBitmap.Width, ThreeDBitmap.Height));
           ShowFrame; // Draw cube;
         Stage := 2;
       end;
    2: if not Pause then
       begin
         case PatternRadio.ItemIndex of
           0: Lissajous;
           1: Axes;
           2: PosControl;
           end;
         if CycleDone then
         begin          // allow reorganizations only after complete cycle
           Angle := 0.0; // wrap angle mod 360 degrees
           ReorgOK := true;
           ShowFrame; // Draw cube;
           ThreeDPaintBox.Canvas.Draw(0, 0, ThreeDBitmap);
           with ThreeDBitmap.Canvas do
             FillRect(Rect(0, 0, ThreeDBitmap.Width, ThreeDBitmap.Height));
           ShowWeights;
           WeightsPaintBox.Canvas.Draw(0, 0, WtBitmap);
           UpDateReorgCounters;
           ShowReorg;
           ReorgPaintBox.Canvas.Draw(0, 0, ReorgBitmap);
           Inc(Iters);
         end;
         if not Fast then
         begin
           if CycleDone then ShowFrame; // Draw Cube
           ThreeDPaintBox.Canvas.Draw(0, 0, ThreeDBitmap);
         end;
         Control;
         PlotVariables;
       end;
     end; {Case Stage of}
end;

procedure TMainForm.Idleloop(Sender: TObject; var Done: Boolean);
var
  Denom: Int64;
begin
  Done := False;
  Iterate;
  if not Fast then Denom := 100 else Denom := 60000;
  repeat
    QueryPerformanceCounter(t2);
  until (t2 - t1) >= Freq/Denom;
  t1 := t2;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  ChangeScale(Screen.Height, 800);
  Stage := 1;
  Application.OnIdle := IdleLoop;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(t1);
  ThreeDBitmap := TBitmap.Create;
  WtBitmap := TBitmap.Create;
  ReorgBitmap := TBitmap.Create;
  InitVars;
  ResetView;
  MinDet := DetTrackBar.Position/10.0;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  ReorgBitmap.Free;
  WtBitmap.Free;
  ThreeDBitmap.Free;
end;

procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.SetCtheta;
begin
  Ctheta := Cos(pi*RotateX.Position/180.0);
  Stheta := Sin(pi*RotateX.Position/180.0);
end;

procedure TMainForm.SetCphi;
begin
  Cphi := Cos(pi*RotateY.Position/180.0);
  Sphi := Sin(pi*RotateY.Position/180.0);
end;

procedure TMainForm.SetCrho;
begin
  Crho := Cos(PI*RotateZ.Position/180.0);
  Srho := Sin(PI*RotateZ.Position/180.0);
end;

procedure TMainForm.RotateXChange(Sender: TObject);
begin
  SetCtheta;
end;

procedure TMainForm.RotateYChange(Sender: TObject);
begin
  SetCphi;
end;

procedure TMainForm.RotateZChange(Sender: TObject);
begin
  SetCrho;
end;

procedure TMainForm.ResetView;
begin
  RotateX.Position := 20;
  RotateY.position := 20;
  RotateZ.position := 0;
  SetCphi;
  SetCtheta;
  SetCrho;
end;

procedure TMainForm.ViewResetBtnClick(Sender: TObject);
begin
  ResetView;
end;

procedure TMainForm.ChangeSizeChange(Sender: TObject);
begin
  Z0 := ChangeSize.Position;
end;

procedure TMainForm.WeightsBtnClick(Sender: TObject);
begin
  SetNewWeights;
  ClearLogPlot;
end;

procedure TMainForm.DetTrackBarChange(Sender: TObject);
begin
  MinDet := DetTrackBar.Position/10.0;
  DetRangeLabel.Caption :=
    'Determinant = ' +
         FloatToStrF(MinDet, Fffixed, 5, 2) + ' to ' +
         FloatToStrF(MinDet + 0.09, Fffixed, 5, 2);
end;

procedure TMainForm.ReorgRadioClick(Sender: TObject);
begin
  Case ReorgRadio.ItemIndex of
    0: GlobalReorg := False;
    1: GlobalReorg := True;
    end;
end;

procedure TMainForm.PauseBtnClick(Sender: TObject);
begin
  Pause := not Pause;
  if Pause then PauseBtn.Caption := 'Continue'
  else PauseBtn.Caption := 'Pause';
end;

procedure TMainForm.SpeedRadioClick(Sender: TObject);
var
  I: Integer;
begin
  Case SpeedRadio.ItemIndex of
    0: Fast := True;
    1: Fast := False;
  end;
  if not Fast then
  begin
    Angle := 6*PI;
    for I := 0 to LastCS do CS[I].Qo := 0;
  end;
end;

procedure TMainForm.ShowWeightLabels;
var
  I: Integer;
  Temp: String;
begin
  with WtBitmap, Canvas do
  begin
    Font.Color := clRed;
    if ShowInputWeights then
      Temp :=  '            Input Weights         '
    else Temp := 'Reorganization Weights';
    TextOut(HCent+20-TextWidth(Temp) div 2, 20, Temp);
    Temp := 'Output Weights';
    TextOut(HCent+20-TextWidth(Temp) div 2, 90, Temp);
    for I := 0 to LastCS do
    begin
      Temp := Format('System %d: ',[I]);
      TextOut(70, 40+15*I, Temp);
      TextOut(70, 110+15*I, Temp);
    end;
    Font.Color := clBlack;
  end;
end;

procedure TMainForm.WeightRadioClick(Sender: TObject);
begin
  Case WeightRadio.ItemIndex of
    0:  ShowInputWeights := True;
    1:  ShowInputWeights := False;
  end;
  ShowWeightLabels;
end;

procedure TMainForm.WeightsPaintBoxPaint(Sender: TObject);
begin
  WeightsPaintBox.Canvas.Draw(0, 0, WtBitmap);
end;

procedure TMainForm.ReorgPaintBoxPaint(Sender: TObject);
begin
  ReorgPaintBox.Canvas.Draw(0, 0, ReorgBitmap);
end;

procedure TMainForm.ThreeDPaintBoxPaint(Sender: TObject);
begin
  ThreeDPaintBox.Canvas.Draw(0, 0, ThreeDBitmap);
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

procedure TMainForm.InstructBtnClick(Sender: TObject);
begin
  InstructForm.ShowModal;
end;

end.
