unit Arm;

interface

uses Windows, Messages, SysUtils, Variants, Classes, Graphics, Forms;

var
  ShldR, ShldP, ShldY,    // shoulder roll, pitch, yaw
  ElbowP, ForeR,          // elbow pitch, forearm roll
  WrP, WrY,               // wrist pitch, yaw
  Th1P, Th1R, Th2P, Th3P, // thumb base pitch, roll; thumb 2, 3 pitch
  F1P, F2P, F3P: Integer; // fingers 1, 2, 3 pitch
  SurfaceColor: TColor;

procedure InitArm;
procedure OperateArm(Bitmap: TBitmap);

implementation

Uses
  Graphics3D;

const
  ArmSegXYZ: array[0..17] of
    record
      X, Y, Z, Next: Integer;
    end = ((x:  0; y:25; z:22; Next:1), (x: 60; y:23; z:20; Next:1),
           (x:100; y:19; z:17; Next:1), (x:120; y:22; z:15; Next:1),
           (x:140; y:23; z:16; Next:1), (x:150; y:21; z:15; Next:1),
           (x:170; y:19; z:15; Next:1), (x:200; y:15; z:12; Next:1),
           (x:220; y:14; z:10; Next:1), (x:240; y:16; z:10; Next:5),
           (x:240; y:08; z:10; Next:1), (x:255; y: 8; z:10; Next:1),
           (x:270; y:09; z:08; Next:1), (x:285; y: 5; z:06; Next:0),
           (x:275; y:25; z:06; Next:1), (x:300; y:22; z:07; Next:1),
           (x:315; y:18; z:05; Next:1), (x:325; y:15; z:05; Next:0));

var
  ArmShape: FaceArrayType;

procedure InitArm;
begin
  MakeISinICos;
  Scale := 1;
//  ScreenDistance := 4000;
  NumW := 7;
  NumH := 17;
  Light.Y := -Sqrt(2) / 2;  // direction cosines
  Light.X := Sqrt(2) / 20;
  Light.Z := Sqrt(2) / 2;
  SurfaceColor := RGB(250, 200, 150);
end;

procedure MakeArm(var Source:FaceArrayType);
var
  M: Integer;
  I, J, K, T: Integer;
  N: Word;
begin
  NFace := NumH * NumW;
  NVert := 4; // Arm built of rectangular faces
  SetLength(Source, Nface);
  for I := 0 to Nface - 1 do
    SetLength(Source[I].F, NVert); // Set length of dynamic array
  for J := 0 to NumH - 1 do
  begin
    for I := 0 to NumW - 1 do
    begin
      N := J * NumW + I;
      with Source[N] do
      begin
        C := SurfaceColor;
        for K :=  0 to NVert - 1 do
          case K of
            0,3: with ArmSegXYZ[J] do
                   if Next <> 0 then
                   begin
                     T := Round(a360*(I + K div 2) / NumW);
                     F[K].x3d := X;
                     if J in [10,11,12] then M := ArmSegXYZ[9].Y
                       else M := 0;
                     F[K].y3d := M + Round(y * iSin[T]);
                     F[K].z3d := Round(Z * iCos[T]);
                   end;
            1,2: if ArmSegXYZ[J].Next <> 0 then
                   with ArmSegXYZ[J + ArmSegXYZ[J].Next] do
                   begin
                     T := Round(a360*(I + K div 2) / NumW);
                     F[k].x3d := X;
                     if J in [10,11,12,13] then M := ArmSegXYZ[9].Y
                       else M := 0;
                     F[K].y3d := M + Round(Y * iSin[T]);
                     F[K].z3d := Round(Z * iCos[T]);
                   end;
          end;
      end;
    end;
  end;
end;

procedure OperateArm(Bitmap: TBitmap);
begin
  MakeArm(ArmShape);
  Rotate3D(ArmShape,'y',  -512 +    F3P, 16, 17, ArmSegXYZ[16].x,  0, 0);
  Rotate3D(ArmShape,'y', -1024 +    F2P, 15, 17, ArmSegXYZ[15].x,  0, 0);
  Rotate3D(ArmShape,'y', -1024 +    F1P, 14, 17, ArmSegXYZ[14].x,  0, 0);
  Rotate3D(ArmShape,'y',   256 -   Th3P, 12, 13, ArmSegXYZ[12].x,  0, 0);
  Rotate3D(ArmShape,'y',   512 -   Th2P, 11, 13, ArmSegXYZ[11].x,  0, 0);
  Rotate3D(ArmShape,'y', -1324 -   Th1P, 10, 13, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'x',       -   Th1R, 10, 13, ArmSegXYZ[ 9].x, 15, 0);
  Rotate3D(ArmShape,'z',   512         , 10, 13, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'y',            WrY,  9, 17, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'z',            WrP,  9, 17, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'x', -2048 +  ForeR,  9, 17, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'y', -1536 + ElbowP,  3, 17, ArmSegXYZ[ 3].x,  0, 0);
  Rotate3D(ArmShape,'x',  1024 +  ShldR,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  Rotate3D(ArmShape,'z',          ShldP,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  Rotate3D(ArmShape,'y', 2560  +  ShldY,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  ClearBitmap(clSkyBlue, Bitmap);
  ProjectFaces(ArmShape, 0, 0, 1000);
  ShowFaces(ArmShape, Bitmap);
end;

end.
