unit Graphics3D;

interface

uses  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls;

const MaxH = 100;
      MaxW = 100;
      MaxVert = 10;
      a360 = 4096 * 2;
      a180 = a360 div 2;

      Col: array[0..15] of TColor = (clRed,     clBlue,   clGreen,   clAqua,
                                     clltGray, clSilver,  clWhite,   clTeal,
                                     clMaroon,  clYellow, clFuchsia, clPurple,
                                     clDkGray,  clNavy,   clLime,  clOlive);
type
  PointType = record
                x3d, y3d, z3d, x2d, y2d: SmallInt;
              end;
  AngleType = record
                Theta, Theta0, Phi, Phi0, Rho, Rho0: Integer;
              end;
  FaceType = array of PointType;
  FaceArrayType = array of record
                    F: FaceType;
                    C, Cr: TColor;  // color and reflected color
                    A: AngleType;
                  end;
  TPoint2D = array of TPoint; // DYNAMIC ARRAY FOR "POLYGON" PROCEDURE
  ListType = array of SmallInt;
  LightType = record
                X, Y, Z: Single;
              end;
var
  iSin, iCos: array[0..a360] of Single;  // SINE AND COSINES AS TABLES
  NumW, NumH, Nface, Nvert, Radius: Word;
  SortKeys, SortLinks: ListType;
  P: Tpoint2d;
  Light: LightType;
  Scale: Double;
  WireFrame: Boolean;
  ScreenDistance: SmallInt;

procedure ClearBitmap(BkColor: TColor; Bitmap: TBitmap);
procedure MakeSphere(var Source: FaceArrayType);
procedure MakeCube(var Source: FaceArrayType);
procedure MakeRect(var Source: FaceArrayType);
procedure LimitAngle(var A: integer);
procedure MakeiSiniCos;
procedure Rotate3D(var Source: FaceArrayType;  //rotate 3D points in place
                   Axis: Char;            // rotate about 'x', 'y', or 'z'
                   Angle: Integer;        // a360 = 2pi
                   StartAt, EndAt: Integer;
                   x0, y0, z0: SmallInt); // center of rotation
procedure ProjectFaces(var Source: FaceArrayType; X, Y, Z: Integer);
procedure ShowFaces(var Source: FaceArrayType; Bm: TBitmap);
procedure ShowArray(var Source: FaceArrayType; N: Byte; Bm: TBitmap);

implementation

procedure ClearBitmap(BkColor: TColor; Bitmap: TBitmap);
begin
  with Bitmap, Canvas do
  begin
    Brush.Color := BkColor;
    FillRect(Rect(0, 0, Width, Height));
  end;
end;


procedure LimitAngle(var A: Integer);
begin
  while A < 0 do A := A + a360;
  while A > a360 do A := A - a360;
end;

{ QUICKSORT sorts elements in the array A with indices between  }
{ LO and HI (both inclusive). It executes itself recursively.}

procedure QuickSort(L, R: word);
var
  I, J, X, Y: Integer;
  function GetValue(I: Integer): Integer; // sort only the pointers to the keys.
  begin
    Result := SortKeys[SortLinks[i]];
  end;
begin
  I := L; J := R; X := (GetValue(L)+ GetValue(R)) DIV 2;
  repeat
    while GetValue(I) > X do I := I + 1;
    while X > GetValue(J) do J := J - 1;
    if I <= J then
    begin
      Y := SortLinks[I];
      SortLinks[I] := SortLinks[J];
      SortLinks[J] := Y;
      I := I + 1; J := J - 1;
    end;
  until I > J;
  if L < J then QuickSort(L, J);
  if I < R then QuickSort(I, R);
end;

// Sort indexes to the faces by the average z distance to the faces.
procedure SortByDistance(var Source: FaceArrayType);
var
  I, J, AvgZ: SmallInt;
begin
  SetLength(SortKeys, Nface);  // allocate dynamic arrays
  SetLength(SortLinks, Nface);
  for I := 0 to Nface - 1 do  // scan through all faces
  begin
    AvgZ := 0;
    for J := 0 to Nvert - 1 do AvgZ := AvgZ + Source[I].F[J].z3d;
    AvgZ := AvgZ div Nvert;   // avg z-distance of vertex
    SortLinks[I] := I;        // load index array in order of indices
    SortKeys[I] := AvgZ;
  end;
  QuickSort(0, Nface-1);      // rearrange the links in place
end;

procedure ShowArray(var Source: FaceArrayType;
                    N: Byte; Bm: TBitmap);
var I, J: Word;
    X, Y: SmallInt;
begin
  for I := 0 to Nface-1 do
    for J := 0 to Nvert do
      with Source[I].F[J mod Nvert] do   // wrap around to starting vertex
         with Bm, Canvas do
         begin
           Pen.Color := clRed;
           if N = 3 then
           begin
             X := Width div 2 + x3d; Y := Height div 2 - y3d;
           end
           else
           begin
             X := Width div 2 + x2d; Y := Height div 2 - y2d;
           end;
           if J  = 0 then MoveTo(X, Y) else LineTo(X, Y);
         end;
end;

procedure ShowFaces(var Source: FaceArrayType; Bm: TBitmap);
var
  I, J, K: word;
begin
  SortByDistance(Source);
  SetLength(P, Nvert);  // allocate dynamic array P
  for K := 0 to Nface - 1 do
  with Bm, Canvas do
  begin
    I := SortLinks[K];
    for J := 0 to NVert - 1 do
      with Source[I] do
      begin
        with F[J] do
        begin
          P[J].X := Width div 2 + x2d ;
          P[J].Y := Height div 2 - y2d;
        end;
        Brush.Color := Cr;
      end;
      Pen.Color := Brush.Color;   // and pen color
      Polygon(P);                 // draw face to bitmap
  end;
end;

procedure CalcLighting(var S: FaceArrayType);
var
  L1, M1, N1, L2, M2, N2: Integer;
  Lnorm, Mnorm, Nnorm: Double;
  Temp: Double;
  Denom: Double;
  Brt: Single;
  R, G, B: Byte;
  I: Integer;
  Swap: Boolean;
begin
  for I := 0 to Nface - 1 do
  with S[I] do
  begin
    L1 := F[1].x3d - F[0].x3d;  // direction numbers of first two lines
    M1 := F[1].y3d - F[0].y3d;  // in face
    N1 := F[1].z3d - F[0].z3d;
    Swap := (L1 = 0) and (M1 = 0) and (N1 = 0);
    if Swap then
    begin
      Swap := True;
      L1 := F[2].x3d - F[1].x3d;  // direction numbers of first two lines
      M1 := F[2].y3d - F[1].y3d;  // in face
      N1 := F[2].z3d - F[1].z3d;
    end;
    if not (Swap or ((L2 = 0) and (M2 = 0) and (N2 = 0))) then
    begin
      L2 := F[2].x3d - F[1].x3d;
      M2 := F[2].y3d - F[1].y3d;
      N2 := F[2].z3d - F[1].z3d;
    end
    else
    begin
      L2 := F[3].x3d - F[2].x3d;
      M2 := F[3].y3d - F[2].y3d;
      N2 := F[3].z3d - F[2].z3d;
    end;
    Lnorm := (M1 * N2) - (M2 * N1); // direction cosines of normal to plane
    Mnorm := (N1 * L2) - (N2 * L1);
    Nnorm := (L1 * M2) - (L2 * M1);
    Temp := ((Light.X * Lnorm) + (Light.Y * Mnorm) + (Light.Z * Nnorm));
    Denom := Sqrt(Sqr(Lnorm) + Sqr(Mnorm) + Sqr(Nnorm));
    if Denom = 0 then Brt := 1.0 else Brt := 0.4 - (0.6 * Temp / Denom);
    if Brt < 0 then Brt := 0;
    B := Round(((C and $00ff0000) shr 16) * Brt) and $ff;
    G := Round(((C and $0000ff00) shr  8) * Brt) and $ff;
    R := Round((C and $000000ff)          * Brt) and $ff;
    Cr := RGB(R, G, B);
  end;
end;

// ROTATE ALL POINTS ABOUT ONE AXIS, X, Y, OR Z. REPLACE PREVIOUS COORDINATES
procedure Rotate3D(var Source: FaceArrayType; //rotate 3D points in place
                   Axis: Char;                // rotate about 'x', 'y', or 'z'
                   Angle: Integer;            // a360 = 2pi
                   StartAt, EndAt: Integer;
                   x0, y0, z0: SmallInt);     // center of rotation
var
  I, J: SmallInt;
  Ca, Sa: Single;
  X, Y, Z, Xp, Yp, Zp: SmallInt;
begin
  LimitAngle(Angle);
//  for I := 0 to Nface - 1 do
  for I := StartAt * NumW to EndAt * NumW - 1 do
  begin
    for J := 0 to Nvert - 1 do
      with Source[I] do
      begin
        with F[J] do
        begin
          X := x3d - x0;     // REMOVE CENTER OF ROTATION
          Y := y3d - y0;
          Z := z3d - z0;
        end;
        with A do           // ADD ROTATIONS TO ANGLES (DIRECTION NUMBERS)
          case Axis of
          'x': Theta := Theta0 + Angle;
          'y': Phi := Phi0 + Angle;
          'z': Rho := Rho0 + Angle;
          end;
        Ca := iCos[Angle]; Sa := iSin[Angle];
        case Axis of
          'x': begin
                 Xp := X;
                 Yp := Round(Ca * Y + Sa * Z);
                 Zp := Round(-Sa * Y + Ca * Z);
               end;
          'y': begin
                 Xp := Round(Ca * X + Sa * Z);
                 Yp := Y;
                 Zp := Round(-Sa * X + Ca * Z);
               end;
          'z': begin
                 Xp := Round(Ca * X + Sa * Y);
                 Yp := Round(-Sa * X + Ca * Y);
                 Zp := Z;
               end;
          else begin
                 Xp := X;  // if illegal axis, don't rotate
                 Yp := Y;
                 Zp := Z;
                end;
        end;
        with F[J] do
        begin
          x3d := Round(Xp) + x0;   // RESTORE OFFSET FROM AXIS OF ROTATION
          y3d := Round(Yp) + y0;
          z3d := Round(Zp) + z0;
        end;
      end; // of <with source[i]>
      if not WireFrame then CalcLighting(Source);
  end;
end;

// PROJECT ALL POINTS FOR DRAWING
procedure ProjectFaces(var Source: FaceArrayType; X, Y, Z: Integer);
var
  I, J, D: SmallInt;
begin
  for I := 0 to Nface - 1 do
    for J := 0 to Nvert - 1 do
      with Source[I].F[J] do
      begin
        D := Round(Scale * z3d + Z); // avoid division by zero, but don't change z3d
        if D = -ScreenDistance then Inc(D);
        x2d := Round((Scale * x3d + X) * ScreenDistance / (ScreenDistance + D));
        y2d := Round((Scale * y3d + Y) * ScreenDistance / (ScreenDistance + D));
      end;
end;

procedure MakeiSiniCos;
var
  I: SmallInt;
begin
  for I := 0 to a360 do
  begin
   iSin[i] := Sin(2 * PI * I / a360);
   iCos[i] := Cos(2 * PI * I / a360);
  end;
end;

procedure MakeRect(var Source: FaceArrayType);
var
  I, J: Word;
begin
  Nface := 1; Nvert := 4;
  SetLength(Source, Nface);
  for I := 0 to Nface - 1 do SetLength(Source[I].F, Nvert);
  // setlength(colors,nface);
  for I := 0 to Nface - 1 do
    for J := 0 to Nvert - 1 do
    begin
      with Source[I] do
      begin
        with F[J] do
        begin
          case J of
            3: begin x3d :=  70; y3d :=  70; end;
            2: begin x3d := -70; y3d :=  70; end;
            1: begin x3d := -70; y3d := -70; end;
            0: begin x3d :=  70; y3d := -70; end;
          end;
          z3d := 100;
        end;
        C := Col[I mod 16];
      end;
    end;
end;

procedure MakeCube(var Source: FaceArrayType);
var
  I, J: Word;
begin
  Nface := 6; Nvert := 4;
  SetLength(Source, Nface);
  for I := 0 to Nface - 1 do
  SetLength(Source[I].F, Nvert);
  for I := 0 to Nface - 1 do
  begin
    for J := 0 to Nvert - 1 do
    with Source[I] do
    begin
      with F[J] do
        case I of
          0: begin
               case J of
                 3: begin x3d :=  100; y3d :=  100; end;
                 2: begin x3d := -100; y3d :=  100; end;
                 1: begin x3d := -100; y3d := -100; end;
                 0: begin x3d :=  100; y3d := -100; end;
               end;
               z3d := 100;
               C := clLime;
             end;
          1: begin
               case J of
                 0: begin x3d :=  100; y3d :=  100; end;
                 1: begin x3d := -100; y3d :=  100; end;
                 2: begin x3d := -100; y3d := -100; end;
                 3: begin x3d :=  100; y3d := -100; end;
               end;
               z3d := -100;
               C := clRed;
             end;
          2: begin
               case J of
                 0: begin x3d :=  100; z3d :=  100; end;
                 1: begin x3d := -100; z3d :=  100; end;
                 2: begin x3d := -100; z3d := -100; end;
                 3: begin x3d :=  100; z3d := -100; end;
               end;
               y3d := 100;
               C := clBlue;
             end;
          3: begin
               case J of
                 3: begin x3d :=  100; z3d :=  100; end;
                 2: begin x3d := -100; z3d :=  100; end;
                 1: begin x3d := -100; z3d := -100; end;
                 0: begin x3d :=  100; z3d := -100; end;
               end;
               y3d := -100;
               C := clYellow;
             end;
          4: begin
               case J of
                 3: begin y3d :=  100; z3d :=  100; end;
                 2: begin y3d := -100; z3d :=  100; end;
                 1: begin y3d := -100; z3d := -100; end;
                 0: begin y3d :=  100; z3d := -100; end;
               end;
               x3d := 100;
               C := clSilver;
             end;
          5: begin
               case J of
                 0: begin y3d :=  100; z3d :=  100; end;
                 1: begin y3d := -100; z3d :=  100; end;
                 2: begin y3d := -100; z3d := -100; end;
                 3: begin y3d :=  100; z3d := -100; end;
               end;
               x3d := -100;
               C := clPurple;
              end;
        end; //case I of
    end;
  end;
end;

procedure MakeSphere(var Source: FaceArrayType);
var
  I, J, K, T, P, dTheta, dPhi: Integer;
  N: Word;

begin
  Radius := 100;
  Nface := NumH * NumW;
  SetLength(Source, Nface);
  for I := 0 to Nface - 1 do SetLength(Source[I].F, Nvert);
 // SetLength(Colors, Nface);
  dTheta := Round(a360 / NumW);
  dPhi := Round(a180 / NumH);
  for I := 0 to NumH - 1 do
    for J := 0 to NumW - 1 do
    begin
      N := I * NumW + J;
      for K :=  0 to Nvert - 1 do
        with Source[N], F[K] do
        begin
          A.Theta0 := Round(I * dTheta); // coordinates of lower left corner
          A.Phi0 := Round(J * dPhi);
          A.Rho0 := 0;
          LimitAngle(A.Theta0);
          LimitAngle(A.Phi0);
          case K of
            0: begin T := A.Theta0; P := A.Phi0; end;
            1: begin T := A.Theta0 + dTheta; P := A.Phi0; end;
            2: begin T := A.Theta0 + dTheta; P := A.Phi0 + dPhi; end;
            3: begin T := A.Theta0; P := A.Phi0 + dPhi; end;
          end;
          LimitAngle(T);
          // limitangle(p);
          with Source[N] do
          begin
            with F[K] do
            begin
              x3d := Round(Radius * iSin[P] * iCos[T]);
              y3d := Round(Radius * iSin[P] * iSin[T]);
              z3d := Round(Radius * iCos[P]);
            end;
            C := Col[N mod 16];
          end;
        end;
    end;
end;

end.
