unit Main;

//Program: ArmControlReorg.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ComCtrls, StdCtrls, ExtCtrls, Spin;

type
  TMainForm = class(TForm)
    StatusBar1: TStatusBar;
    ControlPanel: TGroupBox;
    ProgramBox: TGroupBox;
    BtnAbout: TButton;
    BtnExit: TButton;
    ActionsBox: TGroupBox;
    TabControl1: TTabControl;
    BtnInstruct: TButton;
    MainPanel: TPanel;
    TopPanel: TPanel;
    BottomPanel: TPanel;
    ArmDisplayPanel: TGroupBox;
    ControlSysPanel: TGroupBox;
    ErrorGraphPanel: TGroupBox;
    OutputWeightPanel: TGroupBox;
    ReferenceRadioGroup: TRadioGroup;
    ReorgRadioGroup: TRadioGroup;
    GainControlsBox: TGroupBox;
    DisturbanceRadioGroup: TRadioGroup;
    Label1: TLabel;
    Label2: TLabel;
    ArmDisplay: TPaintBox;
    ErrorGraphDisplay: TPaintBox;
    ReorgGainEdit: TSpinEdit;
    ControlGainEdit: TSpinEdit;
    ControlSysDisplay: TPaintBox;
    OutputWeightDisplay: TPaintBox;
    OrthoButton: TButton;
    procedure FormCreate(Sender: TObject);
    procedure BtnExitClick(Sender: TObject);
    procedure BtnAboutClick(Sender: TObject);
    procedure TabControl1Change(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ArmDisplayPaint(Sender: TObject);
    procedure ControlSysDisplayPaint(Sender: TObject);
    procedure ErrorGraphDisplayPaint(Sender: TObject);
    procedure OutputWeightDisplayPaint(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure ReorgGainEditChange(Sender: TObject);
    procedure ControlGainEditChange(Sender: TObject);
    procedure ReferenceRadioGroupClick(Sender: TObject);
    procedure DisturbanceRadioGroupClick(Sender: TObject);
    procedure ReorgRadioGroupClick(Sender: TObject);
    procedure BtnInstructClick(Sender: TObject);
    procedure ControlSysDisplayMouseMove(Sender: TObject;
      Shift: TShiftState; X, Y: Integer);
    procedure OrthoButtonClick(Sender: TObject);
    procedure ControlSysDisplayMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure SetBooleans;
    procedure ShowStatus;
    procedure InitSim;
    procedure Initialize;
    procedure ResetSim;
    procedure IdleLoop(Sender: TObject; var Done: Boolean);
    procedure StepModel;
  end;

const
  MaxMatrix = 14; // Number of control systems
  MaxMax = MaxMatrix + 1;
  QiLimits: array[1..MaxMatrix] of
              record
                QiMin, QiMax: Double;
              end = ((QiMin: -2000; QiMax: 1800), // ShldR
                     (QiMin: -3000; QiMax: 2700),
                     (QiMin: -1600; QiMax: 2700),
                     (QiMin: -1768; QiMax: 1600),
                     (QiMin: -2048; QiMax: 2048),
                     (QiMin: -1024; QiMax: 1024),
                     (QiMin: - 600; QiMax:  600),
                     (QiMin: -1800; QiMax: 1300),
                     (QiMin: - 500; QiMax: 1400),
                     (QiMin: - 400; QiMax:  550),
                     (QiMin: -1000; QiMax:  300),
                     (QiMin: -1024; QiMax: 1050),
                     (QiMin: -1500; QiMax: 1000),
                     (QiMin: - 900; QiMax:  600)); // F3P

type
  VectorType = array[1..MaxMax] of Single;
  CoeffType = array[1..MaxMax] of VectorType;
  TControlSysBitmap = class(TBitmap)
    private
      XScale: Double;
      MidX, YOff, YInc, YMargin, YIncL, SysNo: Integer;
      RadR, RadP, RadD: Integer; // circle sizes
      ColorR, ColorP, ColorD: Integer; // circle colors
      OldR, OldP, OldD: array[1..MaxMax] of Integer;
    public
      procedure Init;
      procedure Plot(D, P, R: VectorType);
  end;
  TErrorGraphBitmap = class(TBitmap)
    private
      YBorder, XMargin, OldError, OldY: Integer;
      KErr: Double;
      SourceRect, DestRect: TRect;
    public
      procedure Init;
      procedure DrawLogScale;
      procedure Plot(Error: Double);
    end;
  TOutputWeightBitmap = class(TBitmap)
    private
      XOff, YOff: Integer;
      ScaleX, ScaleY: Double;
      SmallMap: TBitmap;
    public
      constructor Create;
      procedure Init;
      procedure Plot(Wt: CoeffType);
      procedure Free;
    end;

var
  MainForm: TMainForm;
  t1, t2, Freq: Int64;
  dt: Double;
  Pause, FirstTime: Boolean;
  ArmBitmap: TBitmap;
  OutputWeightBitmap: TOutputWeightBitmap;
  ControlSysBitmap: TControlSysBitmap;
  ErrorGraphBitmap: TErrorGraphBitmap;
  Dist, D, D1, D2: VectorType; // Disturbance values
  Pattern: array[1..14] of array[0..99999] of Double;

  //Matrix for input and output weights
  Wi, Wo, NormWi, NormWo, ProdW, InputCorrect: CoeffType;
  //Control-system variables
  Qi, P, E, O, R, Temp: VectorType;
  //Control-system parameters
  Ko, GainReorg, Ortho: Double;
  Reorg, ReorgOK, Disturbances, Jump, Smooth: Boolean;

  TotalError, PctTotalError, SumOut, MaxErSq, MaxOutSq: Double;
  ErSq, LastErSq: VectorType;

  Pat1, Pat2, Pat3: VectorType;
  Iters, PatPtr: Integer;

procedure TestOrtho(var Ortho: Double);
procedure DoControlLoops;

implementation

uses
  Graphics3D, Arm, About, Instructions;

{$R *.dfm}

var
  Shape: FaceArrayType;
  MouseX, MouseY: SmallInt;

//-------------- Define bitmap descendant types ------------

procedure TControlSysBitmap.Init;
var
   LegText: String;
   Y, LegendSpace: Integer;
begin
  LegendSpace := Canvas.TextWidth('2   Shoulder Pitch');
  XScale := 8192 /(Width - LegendSpace);
  MidX := LegendSpace + (Width - LegendSpace) div 2;
  YInc := Height div (MaxMatrix + 3);
  YMargin := YInc * MaxMatrix;
  YIncL := (Height - YMargin) div 4;
  YOff := 3;
  RadR := 4; RadP := 5; RadD := 6;
  ColorR := clRed; ColorP := clGreen; ColorD := clNavy;
  with Canvas do
  begin
    Pen.Width := 2;
    Font.Size := 10;
    // Clear legend area to BtnFace
    Brush.Color := clBtnFace;
    FillRect(Rect(0, YMargin, Width, Height - 1));

    // Draw legend text and symbols
    Pen.Color := ColorR;
    Y := YMargin + YIncL;
    LegText := 'Reference Signals';
    TextOut(MidX - TextWidth(LegText) - 20, Y - 8, LegText);
    Ellipse(MidX - RadR, Y - RadR, MidX + RadR, Y + RadR);

    Pen.Color := ColorP;
    Y := Y + YIncL;
    LegText := 'Perceptual Signals';
    TextOut(MidX - TextWidth(LegText) - 20, Y - 8, LegText);
    Ellipse(MidX - RadP, Y - RadP, MidX + RadP, Y + RadP);

    Pen.Color := ColorD;
    Y := Y + YIncL;
    LegText := 'Disturbance Values';
    TextOut(MidX - TextWidth(LegText) - 20, Y - 8, LegText);
    Ellipse(MidX - RadD, Y - RadD, MidX + RadD, Y + RadD);

    // Clear graph area to sky blue
    Brush.Color := clMoneyGreen;
    FillRect(Rect(0, 0, Width, YMargin - 1));
  end;
end;

procedure TControlSysBitmap.Plot(D, P, R: VectorType);
// Initialize TControlSysBitmap before first use!
var
  I, Y: Integer;
begin
  with Canvas do
  begin
    // Draw Labels
    TextOut(10, 0 * YInc + YOff, ' 1   Shoulder Roll');
    TextOut(10, 1 * YInc + YOff, ' 2   Shoulder Pitch');
    TextOut(10, 2 * YInc + YOff, ' 3   Shoulder Yaw');
    TextOut(10, 3 * YInc + YOff, ' 4   Elbow Pitch');
    TextOut(10, 4 * YInc + YOff, ' 5   Forearm Roll');
    TextOut(10, 5 * YInc + YOff, ' 6   Wrist Pitch');
    TextOut(10, 6 * YInc + YOff, ' 7   Wrist Yaw');
    TextOut(10, 7 * YInc + YOff, ' 8   Thumb 1 Roll');
    TextOut(10, 8 * YInc + YOff, ' 9   Thumb 1 Pitch');
    TextOut(10, 9  * YInc + YOff, '10  Thumb 2 Pitch');
    TextOut(10, 10 * YInc + YOff, '11  Thumb 3 Pitch');
    TextOut(10, 11 * YInc + YOff, '12  Fingers 1 Pitch');
    TextOut(10, 12 * YInc + YOff, '13  Fingers 2 Pitch');
    TextOut(10, 13 * YInc + YOff, '14  Fingers 3 Pitch');
    Pen.color := clGray;   // draw zero line
    MoveTo(MidX, 0);
    LineTo(MidX, YMargin - 1);
    // Plot variables
    Y := YInc div 2;
    for I := 1 to MaxMatrix do
    begin
      // Erase circles at present positions
      Pen.Color := Brush.Color;
      Ellipse(
        MidX - OldD[I] - RadD, Y - RadD, MidX - OldD[I] + RadD, Y + RadD);
      OldD[I] := Round(D[I] / XScale);
      Ellipse(
        MidX - OldP[I] - RadP, Y - RadP, MidX - OldP[I] + RadP, Y + RadP);
      OldP[I] := Round(P[I] / XScale);
      Ellipse(
        MidX - OldR[I] - RadR, Y - RadR, MidX - OldR[I] + RadR, Y + RadR);
      OldR[I] := Round(R[I] / XScale);
      Pen.Color := clSilver;   // draw separator line
      MoveTo(0, Y - Yinc div 2);
      LineTo(Width, Y - Yinc div 2);
       // Draw circles at new positions
      Pen.Color := ColorD;
      Ellipse(
        MidX - OldD[I] - RadD, Y - RadD, MidX - OldD[I] + RadD, Y + RadD);
      Pen.Color := ColorP;
      Ellipse(
        MidX - OldP[I] - RadP, Y - RadP, MidX - OldP[I] + RadP, Y + RadP);
      Pen.Color := ColorR;
      Ellipse(
        MidX - OldR[I] - RadR, Y - RadR, MidX - OldR[I] + RadR, Y + RadR);
      Y := Y + YInc;
    end;
    Pen.Color := clSilver;   // draw last separator line        <------------
    MoveTo(0, YMargin);
    LineTo(Width, YMargin);
  end;
end;

procedure TErrorGraphBitmap.Init;
begin
  with Canvas do
  begin
    Font.Name := 'Arial';
    Font.Size := 10;
    Pen.Color := clSilver;
    Pen.Width := 1;
    XMargin := Width - TextWidth('00000000');
    YBorder := Height div 25;
    KErr := (Height - (2.0 * YBorder)) / Ln(100);
    // Used to move "paper" one pixel left
    SourceRect := Rect(1, 0, XMargin, Height - 1);
    DestRect := Rect(0, 0, XMargin - 1, Height - 1);
  end;
  DrawLogScale;
end;

procedure TErrorGraphBitmap.DrawLogScale;
var
  I, Y: Integer;
  ScaleLabel: Double;
begin
  ScaleLabel := 1.0;
  with Canvas do
  begin
    Brush.Color := clBtnFace;
    Rectangle(XMargin + 1, 0, Width - 1, Height - 1);
    Brush.Color := clWhite;
    for I := 1 to 20 do
    begin
      case I of
         1:     Y := 0;
         2..9:  Y := Round(KErr * (Ln(I)));
        10..19: Y := Round(KErr * Ln(  10 * (I -  9)));
        20:     Y := Round(KErr * Ln( 100 * (I - 19)));
      end;
      MoveTo(0, Height - YBorder - Y);
      LineTo(XMargin - 1, Height - YBorder - Y);
      Pen.Color := clSilver;
      if (I = 1) or (I mod 10 = 0) then
      begin
        Brush.Color := clBtnFace;
        TextOut(XMargin + 10, Height - YBorder - Y - 8,
          FloatToStrF(ScaleLabel, Fffixed, 5, 1));
        Brush.Color := clWhite;
        ScaleLabel := ScaleLabel * 10.0;
      end;
    end;
  end;
end;

procedure TErrorGraphBitmap.Plot(Error: Double);
var
  Y: integer;

  procedure DrawLine(var Y, OldY: Integer; LineColor: TColor);
  begin
    with Canvas do
    begin
      // Plot the new value
      Pen.Color := LineColor;
      MoveTo(XMargin-2, Height - YBorder - Y);
      LineTo(XMargin-1, Height - YBorder - Y);
    end;
  end;

  procedure PlotLogScale(LineColor: TColor);
  var
    I, OldPenWidth: Integer;
  begin
    with Canvas do
    begin
      OldPenWidth := Pen.Width;
      Pen.Width := 1;
      for I := 1 to 20 do
      begin
        case I of
           1:     Y := 0;
           2..9:  Y := Round(KErr * (Ln(I)));
          10..19: Y := Round(KErr * Ln(  10 * (I -  9)));
          20:     Y := Round(KErr * Ln( 100 * (I - 19)));
        end;
        Drawline(Y, OldY, clSilver);
      end;
      Pen.Width := OldPenWidth;
    end;
  end;

begin
  with Canvas do
  begin
    // Move "paper" left one pixel
     CopyRect(DestRect, Canvas, SourceRect);
     // Erase rightmost pixels
     Pen.Color := clWhite;
     Pen.Width := 3;
     MoveTo(XMargin, 0);
     LineTo(XMargin, Height - 1);

    // Add scale lines
    PlotLogScale(clSilver);

    // Plot variable
    Pen.Color := clBlack;

    if Error = 0.0 then Error := 0.1;
    Y := Round(Kerr * Ln(Error));
    Drawline(Y, OldY, clGreen);
    OldError := Y;
    OldY := Y;
  end;
end;

constructor TOutputWeightBitmap.Create;
begin
  inherited;
  SmallMap := TBitmap.Create;
  with SmallMap do
  begin
    Height := 15;
    Width := 15;
  end;
  Init;
end;

procedure TOutputWeightBitmap.Free;
begin
  SmallMap.Free;
  inherited;
end;

procedure TOutputWeightBitmap.Init;
begin
  XOff := 100;
  YOff := 50;
  ScaleX := (Width - XOff) / (SmallMap.Width - 1);
  ScaleY := (Height - YOff) / (SmallMap.Height - 1);
  with Canvas do
  begin
    Font.Name := 'Arial';
    Font.Size := 10;
    Brush.Color := clBtnFace;
  end;
end;

procedure TOutputWeightBitmap.Plot(Wt: CoeffType);
var
   I, J: Integer;
   Bright: Integer;
   Max: Double;
   X, Y, Adjust: Integer;
begin
   // Find Max
   Max := 4;       // for small weights lets brightness go to zero
   for I := 1 to MaxMatrix do
     for J := 1 to MaxMatrix do
       if abs(Wt[I, J]) > Max then Max := abs((Wt[I, J]));

   // Clear canvas to background color
   with Canvas do
     FillRect(rect(0, 0, Width-1, Height-1));

   // Convert the weights to brightness values and plot
     for I := 1 to MaxMatrix do
       for J := 1 to MaxMatrix do
       begin
         Bright :=  Round(255*abs(wo[i,j])/max);
         SmallMap.Canvas.Pixels[I - 1,J - 1] := RGB(Bright, Bright, Bright);
       end;

   // Draw graph labels on Bitmap
   with Canvas do
   begin
     X := XOff + Round(ScaleX * 7) - 75;
     TextOut(X, 5, 'To Environmental Variable');
     Y := YOff + Round(ScaleY * 6);
     TextOut(25, Y, 'From');
     Y := YOff + Round(ScaleY * 7);
     TextOut(20, Y, 'System');
     for I := 0 to MaxMatrix - 1 do
     begin
       X := XOff + Round(ScaleX * I);
       Y := YOff + Round(ScaleY * I);
       if I < 9 then Adjust := 0 else Adjust := -3;
       TextOut(X + 5 + Adjust, 25, IntToStr(I + 1));
       TextOut(XOff - 20 + Adjust, Y, IntToStr(I + 1));
     end;
   end;

   // Copy SmallMap to Bitmap.  Pixel squares automatically rescale to
   // size of target bitmap.
   Canvas.CopyRect(Rect(XOff, YOff, Width - 1, Height - 1),
     SmallMap.Canvas, Rect(0, 0, SmallMap.Width - 1, SmallMap.Height - 1));
//   SmallMap.Free; // Free memory taken by SmallMap;
end;

// ------------------ Matrix Operations --------------------

procedure MM(var M1, M2, M3: CoeffType);
// Multiply two matrices to get a third matrix.
var
  I, J, K: Integer;
begin
  for J := 1 to MaxMatrix do
  for I := 1 to MaxMatrix do
    begin
      M3[J, I] := 0.0;
      for K := 1 to MaxMatrix do
      M3[J, I] := M3[J, I] + M1[i, k] * M2[K, J];
    end;
end;

procedure MV(var M: CoeffType; var V1, V2: VectorType);
// Premultiply vector by square matrix: V2 = M * V1
var I, J: Integer;
begin
  for I := 1 to MaxMatrix do
  begin
    V2[I] := 0.0;
    for J := 1 to  MaxMatrix do
      V2[I] := V2[I] + M[I, J] * V1[J]
  end;
end;

procedure VS(var A, B, C: VectorType);
// Vector subtraction: C = A - B
var I: Integer;
begin
  for I := 1 to  MaxMatrix do
  begin
    C[I] := A[I] - B[I];
    if C[I] > 2000 then C[I] := 2000
    else if C[I] < -2000 then C[I] := -2000;
  end;
end;

procedure VA(var A, B, C: VectorType);
// Vector addition: C = A + B
var I: Integer;
begin
  for I := 1 to  MaxMatrix do
  C[I] := A[I] + B[I];
end;

procedure VI(var A, B: VectorType; K: Double);
// Vector integration: B = B + 0.01*(K*A - B) * dt
var I: Integer;
begin
  for I := 1 to  MaxMatrix do
    B[I] := B[I] + 0.005 * (K * A[I] - B[I]) * dt;
end;

procedure TestOrtho(var Ortho: Double);
var
  I, J: Integer;
  Sum1, Sum2: Double;
begin
  Sum1 := 0.0;
  Sum2 := 0.0;
  for J := 1 to MaxMatrix do // check diagonal concentration
    for I := 1 to MaxMatrix do
    begin
      if I = J then Sum1 := Sum1 + Sqr(Wo[J, I]); //ProdW[J, I]}; // sum of squares on diagonal
      Sum2 := Sum2 + Sqr(Wo[J, I]); //ProdW[J, I]);               // total sum of squares
    end;
  Ortho := Sqrt(Sum1 / Sum2); // Orthogonality
end;

// ------ Make Reference and Disturbance Patterns -----------

procedure MakeSmoothPattern;
// Creates 10,000 values for each axis of rotation
var
  I, J: Longint;
  Angle, Mag: Double;
begin
  for I := 1 to MaxMatrix do
  begin
    for J := 0 to 99999 do
    begin
      Angle := J / 99999 * 4 * Pi;
      case I of
        1: Mag:= 1024.0;
        2: Mag := -1024 + 1024 * Sin(Angle);
        3: Mag :=   512 + 512 * Sin(Angle + Pi / 4);
        4: Mag :=  +312 - 1524*(Sin(Angle) + 0.2 * Sin(3 * Angle));
        5: Mag := -1000 * Sin(Angle - Pi / 4);
        6: Mag :=   0.0;
        7: Mag :=   0.0;
        8: Mag :=   0.0;
        9: Mag :=   0.0;
       10: Mag :=   0.0;
       11: Mag :=   0.0;
       12: Mag := 768.0;
       13: Mag := 768.0;
       14: Mag := 256.0;
      end;
      Pattern[I, J] := Mag;
    end;
  end;
end;

procedure MakeDisturbances;
// Each call provides next smoothed disturbance value for each control system
var
  I: Integer;
begin
  for I := 1 to MaxMatrix do
  begin
    D1[I] := D1[I] + 0.001 * (4000 * (random - 0.5) - D1[I]);
    D2[I] := D2[I] + 0.001 * (D1[I] - D2[I]);
    D[I] := D[I] + 0.001 * (50 * D2[I] - D[I]);
  end;
end;

procedure ClearDisturbances;
// Each call supplies a zero disturbance value for each control system
var
  I: Integer;
begin
  for I := 1 to MaxMatrix do
  begin
    D[I] := 0.0; D1[I] := 0.0; D2[I] := 0.0;
  end;
end;

// ----------- Initialize Input and Output Matrices --------

procedure InitMatrix;
var
  I, J: Integer;
begin
  Randomize;  // Seed random number generator
  for I := 1 to  MaxMatrix do
  begin
    O[I] := 0.0;
    for J := 1 to MaxMatrix do
    begin
      if I = J then Wi[I, J] := 1.0 else Wi[I, J] := 0.0; // Input weights
      InputCorrect[I, J] := 2.0 * (0.5 - Random);
      Wo[J, I] :=  0.001;  // Initial Output Weights
    end;
  end;
end;

procedure DoControlLoops;
var
  I, J: Integer;
  Emax: Double;

  begin
  if Reorg then // add corrections to coefficients ("swim in straight line")
  for I := 1 to MaxMatrix do // for i-th control system
  begin
    Emax := Sqrt(LastErSq[I]);
    if Emax > 100 then Emax := 100;
    for J := 1 to MaxMatrix do // for j-th weight
      Wo[I, J] := Wo[I, J] + Emax * GainReorg * InputCorrect[I, J];
  end;
  // One iteration of all control systems
  MV(Wi, Qi, P);    { p = coeffs*qi --  matrix times vector}
  VS(R, P, E);                { e = r - p     --- vector subtract}
  VI(E, O, Ko);               { o = o + ko*int(e) - k2*0 -- vector integrate}
  MV(Wo ,O, Temp);             { temp = wo^*o --- output matrix x output vector}
  if Disturbances then MakeDisturbances else ClearDisturbances;
  VA(Temp, D, Qi);          { qi = temp + d   --- vector add disturbance}

  // Keep joint positions within physical limits
  for I := 1 to MaxMatrix do
  begin
    with QiLimits[I] do
    begin
      if Qi[I] < QiMin then Qi[I] := QiMin
      else if Qi[I] > QiMax then Qi[I] := QiMax;
      if R[I] < QiMin then R[I] := QiMin
      else if R[I] > QiMax then R[I] := QiMax;
    end;
  end;

  // Load values for displaying arm
  ShldR  := Round(Qi[1]);
  ShldP  := -Round(Qi[2]);
  ShldY  := Round(Qi[3]);
  ElbowP := Round(Qi[4]);
  ForeR  := Round(Qi[5]);
  WrY    := Round(Qi[6]);
  WrP    := Round(Qi[7]);
  Th1R   := Round(Qi[8]);
  Th1P   := Round(Qi[9]);
  Th2P   := Round(Qi[10]);
  Th3P   := Round(Qi[11]);
  F1P    := Round(Qi[12]);
  F2P    := Round(Qi[13]);
  F3P    := Round(Qi[14]);

  TotalError := 0.0;
  SumOut := 0.0;
  for I := 1 to MaxMatrix do // for each control system
    begin
      ErSq[I] := Sqr(E[I]);
      TotalError := TotalError + ErSq[I]; // for overall gain reorganization
      SumOut := SumOut + Sqr(O[I]);
    end;
  TotalError := Sqrt(TotalError / MaxMatrix) {/ YScale};
  SumOut := Sqrt(SumOut / MaxMatrix);

  if MaxErSq <> 0 then
    PctTotalError := 100 * TotalError / MaxErSq
  else PctTotalError := 0.0;

  if ReorgOK then
  begin
    for I := 1 to MaxMatrix do // for each control system
    begin
      if ErSq[I] >= LastErSq[I] then // "tumble"  that control system
        for J := 1 to MaxMatrix do // for each weight: change direction
          InputCorrect[I, J] := 2.0 * (0.5 - Random);
      LastErSq[I] := ErSq[I];
    end;
    ReorgOK := False;          // turn off error detection
  end;
end;

procedure TMainForm.ShowStatus;
begin
    with StatusBar1 do
    begin
      if MaxErSq <> 0 then
        Panels[1].Text :=
          'RMS Error  = ' + FloatToStrF(100 * TotalError / MaxErSq, Fffixed, 5, 1)
          + ' % of Max';
      if MaxOutSq <> 0 then
        Panels[2].Text :=
         'RMS Output = ' + FloatToStrF(100 * SumOut / MaxOutSq, Fffixed, 5, 1)
           + ' % of Max';
      Panels[3].Text := 'Orthogonality = ' + FloatToStrF(Ortho, Fffixed, 5, 2);
    end;
end;

procedure TMainForm.SetBooleans;
begin
  with ReferenceRadioGroup do
    case ItemIndex of
      0: begin Jump := True; Smooth := False; end;  // Jump
      1: begin Jump := False; Smooth := True; end;  // Smooth
      2: begin Jump := False; Smooth := False; end; // Constant
    end;
  with ReorgRadioGroup do
  begin
    case ItemIndex of
      0: Reorg := False;
      1: Reorg := True;
    end;
  with DisturbanceRadioGroup do
    case ItemIndex of
      0: Disturbances := False;
      1: Disturbances := True;
    end;
  end;
end;

procedure TMainForm.InitSim;
// Initialize or reinitialize
begin
//  Scale := 1.2;
  ScreenDistance := 1000;
  Pause := False;
  FirstTime := True;
  dt := 0.2;
  ReferenceRadioGroup.ItemIndex := 1;
  ReorgRadioGroup.ItemIndex := 1;
  DisturbanceRadioGroup.ItemIndex := 1;
  ControlGainEdit.Value := 10;
  ReorgGainEdit.Value := 25;

  GainReorg := 1.0e-7 * ReorgGainEdit.Value;
  Ko :=  ControlGainEdit.Value / MaxMatrix;
  TotalError := 1.0;
  ReorgOK := false;
  ClearDisturbances;
  InitMatrix;
  InitArm;
  Iters := 0;
  PatPtr := 0;
end;

procedure TMainForm.Initialize;
  // Done only once
begin
  Pause := False;
  Application.OnIdle := IdleLoop;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(t1);
//Create bitmaps
  ArmBitmap := TBitmap.Create;
  ControlSysBitmap := TControlSysBitmap.Create;
  ErrorGraphBitmap := TErrorGraphBitmap.Create;
  OutputWeightBitmap := TOutputWeightBitmap.Create;
  MakeSmoothPattern;
  SetBooleans;
  InitSim;
end;

procedure TMainForm.ResetSim;
  //Reinitialize simulation values
var
  I: Integer;
begin
  InitSim;
  for I := 1 to MaxMatrix do R[I] := 0.0;

end;

procedure TMainForm.StepModel;
  //Changes made with each time-step done here
var
  I: Integer;
begin
  DoControlLoops;
  if MaxErSq < TotalError then MaxErSq := TotalError;
  if MaxOutSq < SumOut then MaxOutSq := SumOut;
  Case FirstTime of
    True:
      begin
        MaxErSq := 0.0;
        MaxOutSq := 0.0;
        OperateArm(ArmBitmap);
        ArmDisplay.Canvas.Draw(0, 0, ArmBitmap);

        ShowStatus;
        ControlSysBitmap.Plot(D, P, R);
        ControlSysDisplay.Canvas.Draw(0, 0, ControlSysBitmap);

        OutputWeightBitmap.Plot(Wo);
        OutputWeightDisplay.Canvas.Draw(0, 0, OutputWeightBitmap);

        ErrorGraphDisplay.Canvas.Draw(0, 0, ErrorGraphBitmap);
        Pause := True;
        FirstTime := False;
      end;
    False:
      begin
        // Do every 60 cycles
        if Iters mod 60 = 0 then
        begin
          if Reorg then ReorgOK := true; // Reorganize every 60th iteration
          OperateArm(ArmBitmap);
          ArmDisplay.Canvas.Draw(0, 0, ArmBitmap);
          ShowStatus;
          ControlSysBitmap.Plot(D, P, R);
          ControlSysDisplay.Canvas.Draw(0, 0, ControlSysBitmap);
        end;

        // Do every 500 cycles
        if (Iters mod 500 = 0) then
        begin
          OutputWeightBitmap.Plot(Wo);
          OutputWeightDisplay.Canvas.Draw(0, 0, OutputWeightBitmap);
          TestOrtho(Ortho);
        end;

        if Smooth then  // Get next smooth-pattern reference values
        begin
          Inc(PatPtr);
          if PatPtr > 99999 then PatPtr := 0;
          for I := 1 to MaxMatrix do
            R[I] := Pattern[I, PatPtr];
        end;
        if (Iters mod 5000) = 0 then
        begin
          ErrorGraphBitmap.Plot(PctTotalError);
          ErrorGraphDisplay.Canvas.Draw(0, 0, ErrorGraphBitmap);
          if Jump then  // Get next random-pattern reference values
            for I := 1 to MaxMatrix do
            with QiLimits[I] do
              R[I] := QiMin + 10 + Random(round(QiMax - QiMin - 20));
        end;
        Inc(Iters);
      end;
    end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  ChangeScale(Screen.Width, 1280); //Scale to screen res
  WindowState := wsMaximized;
  Initialize;
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: Boolean);
begin
  Done := False;
  if not Pause then StepModel;
  repeat
    QueryPerformanceCounter(t2);
    Application.ProcessMessages;
  until (t2 - t1) >= Freq/6000;
  t1 := t2;

end;

procedure TMainForm.BtnExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.BtnAboutClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

procedure TMainForm.TabControl1Change(Sender: TObject);
begin
  Case TabControl1.TabIndex of
    0: begin
         Pause := False;
         StatusBar1.Panels[0].Text := 'Simulation Status: Running';
       end;
    1: begin
         Pause := True;
         StatusBar1.Panels[0].Text := 'Simulation Status: Paused';
       end;
    2: begin
         Pause := True;
         StatusBar1.Panels[0].Text := 'Simulation Status: Idle';
         ResetSim;
       end;
    end;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  ArmBitmap.Free;
  ControlSysBitmap.Free;
  ErrorGraphBitmap.Free;
  OutputWeightBitmap.Free;
end;


procedure TMainForm.ArmDisplayPaint(Sender: TObject);
begin
  ArmDisplay.Canvas.Draw(0, 0, ArmBitmap);
end;

procedure TMainForm.ControlSysDisplayPaint(Sender: TObject);
begin
  if TabControl1.TabIndex = 0 then ControlSysBitmap.Init;
  ControlSysBitmap.Plot(D, P, R);
  ControlSysDisplay.Canvas.Draw(0, 0, ControlSysBitmap);
end;

procedure TMainForm.ErrorGraphDisplayPaint(Sender: TObject);
begin
  ErrorGraphDisplay.Canvas.Draw(0, 0, ErrorGraphBitmap);
end;

procedure TMainForm.OutputWeightDisplayPaint(Sender: TObject);
begin
  OutputWeightDisplay.Canvas.Draw(0, 0, OutputWeightBitmap);
end;

procedure TMainForm.FormPaint(Sender: TObject);
var
  I: Integer;
begin
  with ArmBitmap do
  begin
    Height := ArmDisplay.Height;
    Width := ArmDisplay.Width;
  end;

  with ControlSysBitmap do
  begin
    Height := ControlSysDisplay.Height;
    Width := ControlSysDisplay.Width;
    Init;
  end;

  with ErrorGraphBitmap do
  begin
    Height := ErrorGraphDisplay.Height;
    Width := ErrorGraphDisplay.Width;
    Init;
  end;

  with OutputWeightBitmap do
  begin
    Height := OutputWeightDisplay.Height;
    Width := OutputWeightDisplay.Width;
    Init;
  end;

  for I := 0 to 3 do
    StatusBar1.Panels[I].Width := Width div 5;
  Scale := ArmDisplay.Height / 349.0;
end;

procedure TMainForm.ReorgGainEditChange(Sender: TObject);
begin
 GainReorg := 1.0e-7 * ReorgGainEdit.Value;
end;

procedure TMainForm.ControlGainEditChange(Sender: TObject);
begin
 Ko := ControlGainEdit.Value / MaxMatrix;
end;

procedure TMainForm.ReorgRadioGroupClick(Sender: TObject);
begin
  SetBooleans;
end;

procedure TMainForm.ReferenceRadioGroupClick(Sender: TObject);
begin
  SetBooleans;
end;

procedure TMainForm.DisturbanceRadioGroupClick(Sender: TObject);
begin
  SetBooleans;
end;

procedure TMainForm.BtnInstructClick(Sender: TObject);
begin
  InstructForm.ShowModal;
end;

procedure TMainForm.OrthoButtonClick(Sender: TObject);
var
  I, J: Integer;
begin
  for I := 1 to MaxMatrix do
    for J := 1 to MaxMatrix do
      if J = I then Wo[I, J] := 10.0 else Wo[I, J] := 0.0;
  OutputWeightBitmap.Plot(Wo);
  OutputWeightDisplay.Canvas.Draw(0, 0, OutputWeightBitmap);
end;

procedure TMainForm.ControlSysDisplayMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
  // Changes selected reference value
begin
  if (ReferenceRadioGroup.ItemIndex = 2) and (ssLeft in Shift) then
    with ControlSysBitmap do
    begin
      if (SysNo > 0) and (SysNo <= MaxMatrix) then
        begin
          R[SysNo] := (MidX - X) * XScale;
          with QiLimits[SysNo] do
          begin
            if R[SysNo] < QiMin then R[SysNo] := QiMin
            else if R[SysNo] > QiMax then R[SysNo] := QiMax;
          end;
          if TabControl1.TabIndex > 0 then
          begin
            ControlSysBitmap.Plot(D, P, R);
            ControlSysDisplay.Canvas.Draw(0, 0, ControlSysBitmap);
          end;
        end;
    end;
end;

procedure TMainForm.ControlSysDisplayMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  // Uses mouse Y to find which reference selected when
  // left mouse button pressed
var
  YPos: Integer;
begin
  if ReferenceRadioGroup.ItemIndex = 2 then
    with ControlSysBitmap do
    begin
      YPos := 0;
      SysNo := 0;
      while YPos < Y do
      begin
        YPos := YPos + YInc;
        Inc(SysNo);
      end;
    end;
    if TabControl1.TabIndex > 0 then
    begin
      ControlSysBitmap.Plot(D, P, R);
      ControlSysDisplay.Canvas.Draw(0, 0, ControlSysBitmap);
    end;

end;

end.
