unit FloatArm;

interface

uses Windows, Messages, SysUtils, Variants, Classes, Graphics, Forms;

var
  ShldR, ShldP, ShldY,    // Shoulder roll, pitch, yaw
  ElbowP, ForeR,          // elbow pitch, forearm roll
  WrP, WrY,               // wrist pitch, yaw
  Th1P, Th1R, Th2P, Th3P, // thumb base pitch, roll; thumb 2, 3 pitch
  F1P, F2P, F3P: Double; // fingers 1, 2, 3 pitch
  SurfaceColor: TColor;
  ViewY, ViewX: double;

procedure InitArm;
procedure OperateArm(Bitmap: TBitmap);

implementation

Uses
  FloatGraphics3D, Main;


Const
   ArmSegXYZ: array[0..17] of
    record
      X, Y, Z, Next: Integer;
    end = ((X:  0; Y:25; Z:22; Next:1), (X: 60; Y:23; Z:20; Next:1),
           (X:100; Y:19; Z:17; Next:1), (X:120; Y:22; Z:15; Next:1),
           (X:140; Y:23; Z:16; Next:1), (X:150; Y:21; Z:15; Next:1),
           (X:170; Y:19; Z:15; Next:1), (X:200; Y:15; Z:12; Next:1),
           (X:220; Y:14; Z:10; Next:1), (X:240; Y:16; Z:10; Next:5),
           (X:240; Y:08; Z:10; Next:1), (X:255; Y: 8; Z:10; Next:1),
           (X:270; Y:09; Z:08; Next:1), (X:285; Y: 5; Z:06; Next:0),
           (X:275; Y:25; Z:06; Next:1), (X:300; Y:22; Z:07; Next:1),
           (X:315; Y:18; Z:05; Next:1), (X:325; Y:15; Z:05; Next:0));

var
  ArmShape: FaceArrayType;

procedure InitArm;
begin
  ScreenDistance := 2400;
  NumW := 8;
  NumH := 17;
  Light.Y := -Sqrt(2) / 2;  // direction cosines
  Light.X := Sqrt(2) / 20;
  Light.Z := Sqrt(2) / 2;
  SurfaceColor := RGB(250, 200, 150);
end;

procedure GetCriticalPoints(var Source:FaceArrayType);
begin
  WristBase.X3D :=
    (Source[9*NumW].F[0].X3D + Source[9*NumW + 4].F[0].X3D) / 2;
  WristBase.Y3D :=
    (Source[9*NumW].F[0].Y3D + Source[9*NumW + 4].F[0].Y3D) / 2;
  WristBase.Z3D :=
    (Source[9*NumW].F[0].Z3D + Source[9*NumW + 4].F[0].Z3D) / 2;
  WristBase.X2D :=
    (Source[9*NumW].F[0].X2D + Source[9*NumW + 4].F[0].X2D) / 2;
  WristBase.Y2D :=
    (Source[9*NumW].F[0].Y2D + Source[9*NumW + 4].F[0].Y2D) / 2;
  KnuckleBase.X3D :=
    (Source[9*NumW].F[1].X3D + Source[9*NumW + 4].F[1].X3D) / 2;
  KnuckleBase.Y3D :=
    (Source[9*NumW].F[1].Y3D + Source[9*NumW + 4].F[1].Y3D) / 2;
  KnuckleBase.Z3D :=
    (Source[9*NumW].F[1].Z3D + Source[9*NumW + 4].F[1].Z3D) / 2;
  KnuckleBase.X2D :=
    (Source[9*NumW].F[1].X2D + Source[9*NumW + 4].F[1].X2D) / 2;
  KnuckleBase.Y2D :=
    (Source[9*NumW].F[1].Y2D + Source[9*NumW + 4].F[1].Y2D) / 2;
  PalmRight.X3D :=
    (Source[9*NumW].F[0].X3D + Source[9*NumW + 2].F[0].X3D) / 2;
  PalmRight.Y3D :=
    (Source[9*NumW].F[0].Y3D + Source[9*NumW + 2].F[0].Y3D) / 2;
  PalmRight.Z3D :=
    (Source[9*NumW].F[0].Z3D + Source[9*NumW + 2].F[0].Z3D) / 2;
  PalmRight.X2D :=
    (Source[9*NumW].F[0].X2D + Source[9*NumW + 2].F[0].X2D) / 2;
  PalmRight.Y2D :=
    (Source[9*NumW].F[0].Y2D + Source[9*NumW + 2].F[0].Y2D) / 2;
  PalmLeft.X3D :=
    (Source[9*NumW].F[0].X3D + Source[9*NumW + 6].F[0].X3D) / 2;
  PalmLeft.Y3D :=
    (Source[9*NumW].F[0].Y3D + Source[9*NumW + 6].F[0].Y3D) / 2;
  PalmLeft.Z3D :=
    (Source[9*NumW].F[0].Z3D + Source[9*NumW + 6].F[0].Z3D) / 2;
  PalmLeft.X2D :=
    (Source[9*NumW].F[0].X2D + Source[9*NumW + 6].F[0].X2D) / 2;
  PalmLeft.Y2D :=
    (Source[9*NumW].F[0].Y2D + Source[9*NumW + 6].F[0].Y2D) / 2;

  FingerTip := Source[16*NumW +4].F[0];





  Shoulder.X3D := (Source[0].F[0].X3D + Source[4].F[0].X3D) / 2;
  Shoulder.Y3D := (Source[0].F[0].Y3D + Source[4].F[0].Y3D) / 2;
  Shoulder.Z3D := (Source[0].F[0].Z3D + Source[4].F[0].Z3D) / 2;
  Shoulder.X2D := (Source[0].F[0].X2D + Source[4].F[0].X2D) / 2;
  Shoulder.Y2D := (Source[0].F[0].Y2D + Source[4].F[0].Y2D) / 2;
end;

procedure MakeArm(var Source:FaceArrayType);
var
  I, J, K, M: Integer;
  T: double;
  N: Word;
begin
  NFace := NumH * NumW;
  NVert := 4;
  SetLength(Source, Nface);
  for I := 0 to Nface - 1 do
    SetLength(Source[I].F, NVert);
  for J := 0 to NumH - 1 do
  begin
    for I := 0 to NumW - 1 do
    begin
      N := J * NumW + I;
      with Source[N] do
      begin
        C := SurfaceColor;
        for K :=  0 to NVert - 1 do
          case K of
            0,3: with ArmSegXYZ[J] do
                   if Next <> 0 then
                   begin
                     T := Round(a360*(I + K div 2) / NumW);
                     F[K].X3D := X;
                     // offset thumb in y
                     if J in [10,11,12] then M := ArmSegXYZ[9].Y
                       else M := 0;
                     F[K].Y3D := M + y * Sin(T);
                     F[K].Z3D := Z * Cos(T);
                   end;
            1,2: if ArmSegXYZ[J].Next <> 0 then
                   with ArmSegXYZ[J + ArmSegXYZ[J].Next] do
                   begin
                     T := Round(a360*(I + K div 2) / NumW);
                     F[k].X3D := X;
                     // offset thumb in y
                     if J in [10,11,12,13] then M := ArmSegXYZ[9].Y
                       else M := 0;
                     F[K].Y3D := M + Y * Sin(T);
                     F[K].Z3D := Z * Cos(T);
                   end;
          end;
      end;
    end;
  end;
end;

// position of index finger =

procedure OperateArm(Bitmap: TBitmap);
begin
  Scale := bitmap.width*1.8/1024;
  MakeArm(ArmShape);
  Rotate3D(ArmShape,'y',          F3P,   16, 17, ArmSegXYZ[16].x,  0, 0); // F3P
  Rotate3D(ArmShape,'y',          F2P,   15, 17, ArmSegXYZ[15].x,  0, 0);
  Rotate3D(ArmShape,'y',          F1P,   14, 17, ArmSegXYZ[14].x,  0, 0);
  Rotate3D(ArmShape,'y',    0.1 - Th3P,  12, 13, ArmSegXYZ[12].x,  0, 0);
  Rotate3D(ArmShape,'y',    0.2 - Th2P,  11, 13, ArmSegXYZ[11].x,  0, 0);
  Rotate3D(ArmShape,'y',   -0.5 - Th1P,  10, 13, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'x',    0.2 - Th1R,  10, 13, ArmSegXYZ[ 9].x, 15, 0);
  Rotate3D(ArmShape,'z',          0.2,   10, 13, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'z',          WrY,    9, 17, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'y',          WrP,    9, 17, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'x',   pi/2 + ForeR,  9, 17, ArmSegXYZ[ 9].x,  0, 0);
  Rotate3D(ArmShape,'y',          ElbowP, 3, 17, ArmSegXYZ[ 3].x,  0, 0);
  Rotate3D(ArmShape,'x',  -pi/2 + ShldR,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  Rotate3D(ArmShape,'z',          ShldP,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  Rotate3D(ArmShape,'y',          ShldY,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  GetCriticalPoints(ArmShape);
  Rotate3D(ArmShape,'y',          ViewX,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  Rotate3D(ArmShape,'x',          ViewY,  0, 17, ArmSegXYZ[ 0].x,  0, 0);
  ClearBitmap(clLtGray, Bitmap);
  with Bitmap, Canvas do
   begin
    Pen.Color := clRed;
    MoveTo(Width div 2 + Round(WristBase.X2D),
           Height div 2 - Round(WristBase.Y2D));
    LineTo(Width div 2 + Round(KnuckleBase.X2D),
           Height div 2 - Round(KnuckleBase.Y2D));
    MoveTo(Width div 2 + Round(PalmRight.X2D),
           Height div 2 - Round(PalmRight.Y2D));
    LineTo(Width div 2 + Round(PalmLeft.X2D),
           Height div 2 - Round(PalmLeft.Y2D));
   end;
  ProjectFaces(ArmShape, -150, 0, 2100);
  ShowFaces(ArmShape, Bitmap);
end;

end.
