unit Main;

//Program: Coordination.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ExtCtrls, Math;

type
  TMainForm = class(TForm)
    ControlPanel: TGroupBox;
    DisplayPanel: TGroupBox;
    ArmControls: TGroupBox;
    DisplayMainPanel: TPanel;
    DisplayBotPanel: TPanel;
    RotateHorizTrack: TTrackBar;
    DisplayRtPanel: TPanel;
    RotateVertTrack: TTrackBar;
    HandPanel: TGroupBox;
    FingerPanel: TGroupBox;
    ReachPanel: TGroupBox;
    PitchTrack: TTrackBar;
    YawTrack: TTrackBar;
    RollTrack: TTrackBar;
    ElevationTrack: TTrackBar;
    LateralTrack: TTrackBar;
    ReachTrack: TTrackBar;
    ShoulderTrack: TTrackBar;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    GroupBox1: TGroupBox;
    AboutBtn: TButton;
    ExitBtn: TButton;
    ArmPaintBox: TPaintBox;
    GainBox: TGroupBox;
    KoTrack: TTrackBar;
    KvTrack: TTrackBar;
    KdTrack: TTrackBar;
    KoLabel: TLabel;
    KvLabel: TLabel;
    KdLabel: TLabel;
    GainResetBtn: TButton;
    RotateViewHorizLabel: TLabel;
    RtDisplayPaint: TPaintBox;
    procedure ExitBtnClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RotateHorizTrackChange(Sender: TObject);
    procedure RotateVertTrackChange(Sender: TObject);
    procedure PitchTrackChange(Sender: TObject);
    procedure YawTrackChange(Sender: TObject);
    procedure RollTrackChange(Sender: TObject);
    procedure ElevationTrackChange(Sender: TObject);
    procedure LateralTrackChange(Sender: TObject);
    procedure ReachTrackChange(Sender: TObject);
    procedure ShoulderTrackChange(Sender: TObject);
    procedure KoTrackChange(Sender: TObject);
    procedure KvTrackChange(Sender: TObject);
    procedure KdTrackChange(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
    procedure GainResetBtnClick(Sender: TObject);
    procedure RtDisplayPaintPaint(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure IdleLoop(Sender: TObject; var Done: Boolean);
    procedure SetDefaultGains;
    procedure InitControl;
    procedure Iterate;
  end;

var
  MainForm: TMainForm;
  Stage: Integer;
  T1, T2, Freq: Int64;
  Reach, ShoulderPitch, ShoulderYaw, ShoulderRoll,
    HandPitch, HandYaw, Handelev: Double;
  ReachRef, HandPitchRef, HandYawRef, HandElevRef: Double;
  ReachError, HandPitchError, HandYawError, HandElevError: Double;
  HandRoll, HandRollRef, HandRollError: Double;
  HandLateral, HandLateralRef, HandLateralError: Double;
  Slider, ArmRoll: Double;

implementation

uses FloatArm, FloatGraphics3D, About, AngleFont;

{$R *.dfm}

const
  MaxLev1Sys = 14;
  MaxLev2Sys = 4;
  QiLimits: array[1..MaxLev1Sys] of
             record
                QiMin, QiMax: Double;
  end = ((QiMin: -pi/2; QiMax: pi/2),    // ShldY.
         (QiMin: -0.5*pi; QiMax: 0.7*pi),// ShldP.
         (QiMin: -pi/2; QiMax: pi/2),    // ShldR.
         (QiMin:     0; QiMax: 0.75*pi), // ElbowP.
         (QiMin: -pi/2; QiMax: pi/2),    // ForeR
         (QiMin: -pi/2; QiMax: pi/4),    // WrPitch
         (QiMin: -pi/8; QiMax: pi/4),    // WrYaw
         (QiMin: -pi/2; QiMax: pi/2),    // Th1Yaw
         (QiMin: -pi/8; QiMax: 0.4*pi),  // Th1Pitch
         (QiMin: -pi/8; QiMax:  pi/8),   // Th2Pitch
         (QiMin:  0;    QiMax:  pi/4),   // Th3Pitch
         (QiMin:  0;    QiMax: pi/2),    // F1Pitch
         (QiMin:  0;    QiMax: pi/2),    // F2Pitch
         (QiMin:  0;    QiMax: pi/4));   // F3Pitch

type
  CSystype = record
              v, m, p, e, ev, r, rv, qi, qo, f, dist: double;
              d, d1, d2: double;
             end;

var
  BM: TBitmap;
  CSysLev1: array[1..MaxLev1Sys] of CSysType;
  CSysLev2: array[1..MaxLev2Sys] of CSysType;
  dt: double;
  Ko, Kv, Kd: double;
  OutStr: string;

procedure MakeDist;
// Each call provides next smoothed disturbance value for each control system
var
  I: Integer;
begin
  for I := 1 to MaxLev1Sys do
  with CsysLev1[I] do
  begin
    d1 := d1 + 0.001 * (4000 * (random - 0.5) - d1);
    d2 := d2 + 0.001 * (d1 - d2);
    d := d + 0.001 * (50 * d2 - d);
  end;
end;

procedure ClearDisturbances;
// Each call supplies a zero disturbance value for each control system
var
  I: Integer;
begin
  for I := 1 to MaxLev1Sys do
  with CsysLev1[I] do
  begin
    d := 0.0; d1 := 0.0; d2 := 0.0;
  end;
end;

procedure TMainForm.SetDefaultGains;
begin
  Kd := 1.2;
  Kv := 0.8;
  Ko := 40.0;
  KdTrack.Position := Round(Kd*100);
  KvTrack.Position := Round(Kv*10);
  KoTrack.Position := Round(Ko*10);
  Str(Ko:4:2, OutStr);
  KoLabel.Caption := 'Ko = ' + OutStr;
  Str(Kv:4:2, OutStr);
  KvLabel.Caption := 'Kv = ' + OutStr;
  Str(Kd:4:2, OutStr);
  KdLabel.Caption := 'Kd = ' + OutStr;
end;

procedure TMainForm.InitControl;
var
  I: Integer;
begin
  for I := 1 to MaxLev1Sys do
  with CSysLev1[i] do
  begin
    r := 0.0; qo := 0.0;
    m := 3.0;
  end;
  SetDefaultGains;
  Slider := ReachTrack.Position;
  HandPitchRef := PitchTrack.Position;
  HandYawRef := YawTrack.Position;
  HandRollRef := RollTrack.Position*pi/4096;
  HandElevRef := ElevationTrack.Position/4;
  ShoulderYaw := LateralTrack.Position*pi/4096;
  ShoulderRoll := ShoulderTrack.Position*pi/4096;
  dt := 1/60.0;
end;

procedure Level2Control;
var
  SinR: double;  // sin hand roll angle
  CosR: double;  // cos hand roll angle
  Temp: double; // hypotenuses;
begin
  Reach := Sqrt(Sqr(FingerTip.x3d - Shoulder.x3d) +
                Sqr(FingerTip.y3d - Shoulder.y3d) +
                Sqr(FingerTip.z3d - Shoulder.z3d));

  ReachRef := 100.0 + Slider/8.9;
  ReachError := ReachRef - Reach;

  HandElev := FingerTip.y3d - Shoulder.y3d;
  HandElevError := HandElevRef - HandElev;

  HandLateral := FingerTip.z3d - Shoulder.z3d;
  HandLateralError := HandLateralRef - HandLateral;

  HandPitch := KnuckleBase.y3d - WristBase.y3d;
  HandPitchError := 16.0*(HandPitchRef - HandPitch);

  HandYaw := KnuckleBase.z3d - WristBase.z3d;
  HandYawError := HandYawRef - HandYaw;

//  HandRoll := ArcTan2(PalmRight.y3d - PalmLeft.y3d,Temp);
  HandRoll := ShldR + ForeR;
  HandRollError := HandRollRef - HandRoll;

  SinR :=  Sin(HandRoll);
  CosR :=  Cos(HandRoll);

  HandYaw := KnuckleBase.z3d - WristBase.z3d;
  HandYawError := HandYawRef - HandYaw;

  with CSysLev1[5] do r := 8.0*HandRollError;
  with CSysLev1[6] do r := -0.02*HandPitchError*SinR - 0.2*HandYawError*CosR;
  with CSysLev1[7] do r := 0.2*HandYawError*SinR - 0.02*HandPitchError*CosR;
  with CSysLev1[2] do r := -ReachError/30.0 - HandElevError/30.0;
  with CSysLev1[4] do r := -ReachError/30.0;
  with CSysLev1[1] do r := -HandLateralError/10.0;;
  with CSysLev1[3] do r := ShoulderRoll;
end;

// Control Systems, 1 to MaxLev1Sys -- all joint angles
procedure Level1Control;
var
  I: Integer;
begin
  MakeDist;
  for I := 1 to MaxLev1Sys do
  begin
    with CSysLev1[I] do
    begin
      Ev := Rv - 5*Kd*v;       // velocity error signal
      f := Kv*Ev{ + dist};     // output force + disturbing force
      v  :=  v + f/m;        // velocity = integral of output forces
      qi := qi + v*dt;      // position := integral of velocity
      p := qi;
      e := r - p;           // position error
      rv  := 20*e*Ko*dt;       // velocity reference signal
    end;
  end;
  // Keep joint positions within physical limits
  for I := 1 to MaxLev1Sys do
  begin
    with QiLimits[I], CSysLev1[I] do
    begin
      if qi < QiMin then qi := QiMin
      else if qi > QiMax then qi := QiMax;
      if r < QiMin then r := QiMin
      else if r > QiMax then r := QiMax;
    end;
  end;
  // Load values for displaying arm
  ShldY  := CSysLev1[01].qi;
  ShldP  := CSysLev1[02].qi;
  ShldR  := CSysLev1[03].qi;
  ElbowP := CSysLev1[04].qi;
  ForeR  := CSysLev1[05].qi;
  WrP    := CSysLev1[06].qi;
  WrY    := CSysLev1[07].qi;
  Th1R   := CSysLev1[08].qi;
  Th1P   := CSysLev1[09].qi;
  Th2P   := CSysLev1[10].qi;
  Th3P   := CSysLev1[11].qi;
  F1P    := CSysLev1[12].qi;
  F2P    := CSysLev1[13].qi;
  F3P    := CSysLev1[14].qi;
end;

procedure TMainForm.Iterate;
begin
  OperateArm(BM);
  ArmPaintBox.Canvas.Draw(0, 0, BM);
  Level2Control;
  Level1Control;
  Level1Control;
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: Boolean);
begin
  Done := False;
  Case Stage of
  0: begin
       WindowState := wsMaximized;
       InitArm;
       InitControl;
       Stage := 1;
     end;
  1: begin
       Iterate;
     end;
  end;
  repeat
    QueryPerformanceCounter(T2);
  until (T2 - T1) >= Freq/60;
  T1:= T2;
end;



procedure TMainForm.FormCreate(Sender: TObject);
begin
  BM := TBitmap.Create;
  Stage := 0;
  ChangeScale(Screen.Width, 1280);
  Application.OnIdle := IdleLoop;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(T1);
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  BM.Width := ArmPaintBox.Width;
  BM.Height := ArmPaintBox.Height;
  RotateVertTrack.Top := DisplayRtPanel.Height div 2
    - RotateVertTrack.Height div 2;
  RotateHorizTrack.Left := (DisplayBotPanel.Width -DisplayRtPanel.Width) div 2
    - RotateHorizTrack.Width div 2;
  RotateViewHorizLabel.Left := RotateHorizTrack.Left
    + RotateHorizTrack.Width div 2 - RotateViewHorizLabel.Width div 2;
end;

procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  BM.Free;
end;

//--------------- Track Bars ----------------

procedure TMainForm.RotateHorizTrackChange(Sender: TObject);
begin
  viewx := RotateHorizTrack.Position * pi/2000.0;
end;

procedure TMainForm.RotateVertTrackChange(Sender: TObject);
begin
  viewy := RotateVertTrack.Position * pi/2000.0;
end;

procedure TMainForm.PitchTrackChange(Sender: TObject);
begin
  HandPitchRef := PitchTrack.Position;
end;

procedure TMainForm.YawTrackChange(Sender: TObject);
begin
  HandYawRef := YawTrack.Position
end;

procedure TMainForm.RollTrackChange(Sender: TObject);
begin
  HandRollRef := RollTrack.Position*pi/4096;
end;

procedure TMainForm.ElevationTrackChange(Sender: TObject);
begin
  HandElevRef := ElevationTrack.Position/4;
end;

procedure TMainForm.LateralTrackChange(Sender: TObject);
begin
 HandLateralRef := LateralTrack.Position/4;
end;

procedure TMainForm.ReachTrackChange(Sender: TObject);
begin
 Slider := ReachTrack.Position;
end;

procedure TMainForm.ShoulderTrackChange(Sender: TObject);
begin
 ShoulderRoll := ShoulderTrack.Position*pi/4096;
end;

procedure TMainForm.KoTrackChange(Sender: TObject);
begin
 Ko := 0.1*KoTrack.Position;
 Str(Ko:4:2, OutStr);
 KoLabel.Caption := 'Ko = ' + OutStr;
end;

procedure TMainForm.KvTrackChange(Sender: TObject);
begin
 Kv :=0.1*KvTrack.Position;
 Str(Kv:4:2, OutStr);
 KvLabel.Caption := 'Kv = ' + OutStr;
end;

procedure TMainForm.KdTrackChange(Sender: TObject);
begin
 Kd := 0.01*KdTrack.Position;
 Str(Kd:4:2, OutStr);
 KdLabel.Caption := 'Kd = ' + OutStr;
end;

procedure TMainForm.FormPaint(Sender: TObject);
begin
  ArmPaintBox.Canvas.Draw(0, 0, BM);
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

procedure TMainForm.GainResetBtnClick(Sender: TObject);
begin
  SetDefaultGains;
end;

procedure TMainForm.RtDisplayPaintPaint(Sender: TObject);
var
  X, Y: Integer;
  S: String;
begin
  with RtDisplayPaint do
  begin
    S := 'Rotate View About Horizontal Axis';
    X := RotateVertTrack.Left + RotateVertTrack.Width + 5;
    Y := RotateVertTrack.Top + RotateVertTrack.Height div 2
         + Canvas.TextWidth(S) div 2;
    TextOutA(Canvas, X, Y, 900, S);
  end;

end;

end.
