unit FloatArmTrk;

interface

uses Windows, Messages, SysUtils, Variants, Classes, Graphics, Forms;

var
  ArmSegXYZ: array[0..18] of
    record
      X, Y, Z, Next: Integer;
    end = ((x:  0; y:25; z:22; Next:1), (x: 60; y:23; z:20; Next:1),
           (x:100; y:19; z:17; Next:1), (x:120; y:22; z:15; Next:1),
           (x:140; y:23; z:16; Next:1), (x:150; y:21; z:15; Next:1),
           (x:170; y:19; z:15; Next:1), (x:200; y:15; z:12; Next:1),
           (x:220; y:14; z:10; Next:1), (x:240; y:16; z:10; Next:5),
           (x:240; y:08; z:10; Next:1), (x:255; y: 8; z:10; Next:1),
           (x:270; y:09; z:08; Next:1), (x:285; y: 5; z:06; Next:0),
           (x:275; y:25; z:06; Next:1), (x:300; y:22; z:07; Next:1),
           (x:315; y:20; z:05; Next:1), (x:325; y:15; z:05; Next:0),
           (x:275; y:0; z:0; Next:0));


  ShldR, ShldP, ShldY,    // Shoulder roll, pitch, yaw
  ElbowP, ForeR,          // elbow pitch, forearm roll
  WrP, WrY,               // wrist pitch, yaw
  Th1P, Th1R, Th2P, Th3P, // thumb base pitch, roll; thumb 2, 3 pitch
  F1P, F2P, F3P: Double;  // fingers 1, 2, 3 pitch
  SurfaceColor: TColor;
  ViewY, ViewX: Double;

  MouseX, MouseY, MouseZ: Integer;

procedure InitArm(aScale: Double);
procedure OperateArm(Bitmap: TBitmap);

implementation

Uses
  FloatGraphics3D;

var
  ArmShape: FaceArrayType;

procedure InitArm(aScale: Double);
var
  Phi, Theta: Double;
begin
  Phi := pi/4;
  Theta := -pi/4;
  Scale := aScale;
  ScreenDistance := 2400;
  NumW := 8;
  NumH := 20;
  Light.Y := Sin(Phi);  // direction cosines
  Light.X := Cos(Phi)*Cos(Theta);
  Light.Z := -Cos(Phi)*Sin(Theta);
  SurfaceColor := RGB(250, 200, 150);
end;

procedure GetCriticalPoints(var Source:FaceArrayType);
begin
  WristBase.X3D :=
    (Source[9*NumW].F[0].X3D + Source[9*NumW + 4].F[0].X3D) / 2;
  WristBase.Y3D :=
    (Source[9*NumW].F[0].Y3D + Source[9*NumW + 4].F[0].Y3D) / 2;
  WristBase.Z3D :=
    (Source[9*NumW].F[0].Z3D + Source[9*NumW + 4].F[0].Z3D) / 2;
  WristBase.X2D :=
    (Source[9*NumW].F[0].X2D + Source[9*NumW + 4].F[0].X2D) / 2;
  WristBase.Y2D :=
    (Source[9*NumW].F[0].Y2D + Source[9*NumW + 4].F[0].Y2D) / 2;
  KnuckleBase.X3D :=
    (Source[9*NumW].F[1].X3D + Source[9*NumW + 4].F[1].X3D) / 2;
  KnuckleBase.Y3D :=
    (Source[9*NumW].F[1].Y3D + Source[9*NumW + 4].F[1].Y3D) / 2;
  KnuckleBase.Z3D :=
    (Source[9*NumW].F[1].Z3D + Source[9*NumW + 4].F[1].Z3D) / 2;
  KnuckleBase.X2D :=
    (Source[9*NumW].F[1].X2D + Source[9*NumW + 4].F[1].X2D) / 2;
  KnuckleBase.Y2D :=
    (Source[9*NumW].F[1].Y2D + Source[9*NumW + 4].F[1].Y2D) / 2;
  PalmRight.X3D :=
    (Source[9*NumW].F[0].X3D + Source[9*NumW + 2].F[0].X3D) / 2;
  PalmRight.Y3D :=
    (Source[9*NumW].F[0].Y3D + Source[9*NumW + 2].F[0].Y3D) / 2;
  PalmRight.Z3D :=
    (Source[9*NumW].F[0].Z3D + Source[9*NumW + 2].F[0].Z3D) / 2;
  PalmRight.X2D :=
    (Source[9*NumW].F[0].X2D + Source[9*NumW + 2].F[0].X2D) / 2;
  PalmRight.Y2D :=
    (Source[9*NumW].F[0].Y2D + Source[9*NumW + 2].F[0].Y2D) / 2;
  PalmLeft.X3D :=
    (Source[9*NumW].F[0].X3D + Source[9*NumW + 6].F[0].X3D) / 2;
  PalmLeft.Y3D :=
    (Source[9*NumW].F[0].Y3D + Source[9*NumW + 6].F[0].Y3D) / 2;
  PalmLeft.Z3D :=
    (Source[9*NumW].F[0].Z3D + Source[9*NumW + 6].F[0].Z3D) / 2;
  PalmLeft.X2D :=
    (Source[9*NumW].F[0].X2D + Source[9*NumW + 6].F[0].X2D) / 2;
  PalmLeft.Y2D :=
    (Source[9*NumW].F[0].Y2D + Source[9*NumW + 6].F[0].Y2D) / 2;

  FingerTip.X3D := (Source[16*NumW].F[0].X3D + Source[16*NumW+4].F[2].X3D)/2.0;
  FingerTip.Y3D := (Source[16*NumW].F[0].Y3D + Source[16*NumW+4].F[2].Y3D)/2.0;
  FingerTip.Z3D := (Source[16*NumW].F[0].Z3D + Source[16*NumW+4].F[2].Z3D)/2.0;

  Shoulder.X3D := (Source[0].F[0].X3D + Source[4].F[0].X3D) / 2;
  Shoulder.Y3D := (Source[0].F[0].Y3D + Source[4].F[0].Y3D) / 2;
  Shoulder.Z3D := (Source[0].F[0].Z3D + Source[4].F[0].Z3D) / 2;
  Shoulder.X2D := (Source[0].F[0].X2D + Source[4].F[0].X2D) / 2;
  Shoulder.Y2D := (Source[0].F[0].Y2D + Source[4].F[0].Y2D) / 2;
end;

procedure MakeTarget(var Source: FaceArrayType);
var
  I: Integer;
  R, T: Double;
begin
  R := 40;
  for I := 0 to 2 do
  with Source[18*NumW + I] do
  begin
    C := clLime;
    T :=  2.0*pi/3.0*I + pi/6;
    F[0].Y3D := R*Sin(T) + MouseY;
    F[0].Z3D := R*Cos(T) + MouseX;
    F[3].Y3D := R*Sin(T+2*pi/3) + MouseY;
    F[3].Z3D := R*Cos(T+2*pi/3) + MouseX;
    F[1].Y3D := F[0].Y3D;
    F[1].Z3D := F[0].Z3D;
    F[2].Y3D := F[3].Y3D;
    F[2].Z3D := F[3].Z3D;
    F[0].X3D := ArmsegXYZ[18].X;
    F[1].X3D := ArmsegXYZ[18].X + 20;
    F[2].X3D := ArmsegXYZ[18].X + 20;
    F[3].X3D := ArmsegXYZ[18].X;
  end;
end;

procedure MakeArm(var Source:FaceArrayType);
var
  I, J, K, M: Integer;
  T: Double;
  N: Word;
begin
  NFace := NumH * NumW;
  NVert := 4;
  SetLength(Source, Nface + 3);  // Room for target
  for I := 0 to Nface - 1 do
    SetLength(Source[I].F, NVert);
  for J := 0 to NumH - 1 do
  begin
    for I := 0 to NumW - 1 do
    begin
      N := J * NumW + I;
      if N < NumW*18 then
      with Source[N] do
      begin
        C := SurfaceColor;
        for K :=  0 to NVert - 1 do
          case K of
            0, 3: with ArmSegXYZ[J] do
                    if Next <> 0 then
                    begin
                      T := Round(a360*(I + K div 2) / (NumW - 1));
                      F[K].X3D := X;
                      // offset thumb in y
                      if J in [10, 11, 12] then M := ArmSegXYZ[9].Y
                        else M := 0;
                      F[K].Y3D := M + y * Sin(T);
                      F[K].Z3D := Z * Cos(T);
                    end;
            1, 2: if ArmSegXYZ[J].Next <> 0 then
                    with ArmSegXYZ[J + ArmSegXYZ[J].Next] do
                    begin
                      T := Round(a360*(I + K div 2) / (NumW - 1));
                      F[k].X3D := X;
                      // offset thumb in y
                      if J in [10, 11, 12, 13] then M := ArmSegXYZ[9].Y
                        else M := 0;
                      F[K].Y3D := M + Y * Sin(T);
                      F[K].Z3D := Z * Cos(T);
                    end;
          end;
      end;
    end;
  end;
end;

procedure OperateArm(Bitmap: TBitmap);
begin
  MakeArm(ArmShape);
  ArmSegXYZ[18].X := MouseZ;
  ArmSegXYZ[18].Y := MouseY;
  ArmSegXYZ[18].Z := MouseX;
  MakeTarget(ArmShape);
  Rotate3D(ArmShape,'y',          F3P,   16, 17, ArmSegXYZ[16].X,  0, 0); // f3p
  Rotate3D(ArmShape,'y',          F2P,   15, 17, ArmSegXYZ[15].X,  0, 0);
  Rotate3D(ArmShape,'y',          F1P,   14, 17, ArmSegXYZ[14].X,  0, 0);
  Rotate3D(ArmShape,'y',    0.1 - Th3P,  12, 13, ArmSegXYZ[12].X,  0, 0);
  Rotate3D(ArmShape,'y',    0.2 - Th2P,  11, 13, ArmSegXYZ[11].X,  0, 0);
  Rotate3D(ArmShape,'y',   -0.5 - Th1P,  10, 13, ArmSegXYZ[ 9].X,  0, 0);
  Rotate3D(ArmShape,'x',    0.2 - Th1R,  10, 13, ArmSegXYZ[ 9].X, 15, 0);
  Rotate3D(ArmShape,'z',          0.2,   10, 13, ArmSegXYZ[ 9].X,  0, 0);
  Rotate3D(ArmShape,'z',          WrY,    9, 17, ArmSegXYZ[ 9].X,  0, 0);
  Rotate3D(ArmShape,'y',          WrP,    9, 17, ArmSegXYZ[ 9].X,  0, 0);
  Rotate3D(ArmShape,'x',   pi/2 + ForeR,  9, 17, ArmSegXYZ[ 9].X,  0, 0);
  Rotate3D(ArmShape,'y',          ElbowP, 3, 17, ArmSegXYZ[ 3].X,  0, 0);
  Rotate3D(ArmShape,'x',  -pi/2 + ShldR,  0, 17, ArmSegXYZ[ 0].X,  0, 0);
  Rotate3D(ArmShape,'z',          ShldP,  0, 17, ArmSegXYZ[ 0].X,  0, 0);
  Rotate3D(ArmShape,'y',          ShldY,  0, 17, ArmSegXYZ[ 0].X,  0, 0);
  GetCriticalPoints(ArmShape);
  Rotate3D(ArmShape,'y',          ViewX,  0, 20, ArmSegXYZ[ 0].X,  0, 0);
  Rotate3D(ArmShape,'x',          ViewY,  0, 20, ArmSegXYZ[ 0].X,  0, 0);
  ClearBitmap(clGray, Bitmap);
  with Bitmap, Canvas do
  begin
    Pen.Color := clRed;
    MoveTo(Width div 2 + Round(WristBase.X2D),
           Height div 2 - Round(WristBase.Y2D));
    LineTo(Width div 2 + Round(KnuckleBase.X2D),
           Height div 2 - Round(KnuckleBase.Y2D));
    MoveTo(Width div 2 + Round(PalmRight.X2D),
           Height div 2 - Round(PalmRight.Y2D));
    LineTo(Width div 2 + Round(PalmLeft.X2D),
           Height div 2 - Round(PalmLeft.Y2D));
  end;
  ProjectFaces(ArmShape, -250, 0, 800);
  ShowFaces(ArmShape, Bitmap);
end;

end.
