unit FloatGraphics3D;

interface

uses  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls;

const
  MaxH = 100;
  MaxW = 100;
  MaxVert = 10;
  a360 = 2*pi;
  a180 = pi;

  Col: array[0..15] of TColor = (clRed,     clBlue,   clGreen,   clAqua,
                                 clLtGray, clSilver,  clWhite,   clTeal,
                                 clMaroon,  clYellow, clFuchsia, clPurple,
                                 clDkGray,  clNavy,   clLime,  clOlive);
type
  PointType = record
                X3D, Y3D, Z3D, x2d, y2d: Double;
              end;
  AngleType = record
                Theta, Theta0, Phi, Phi0, Rho, Rho0: Double;
              end;
  FaceType = array of PointType;
  FaceArrayType = array of record
                    F: FaceType;
                    C, Cr: TColor;  // color and reflected color
                    A: AngleType;
                  end;
  TPoint2D = array of TPoint; // DYNAMIC ARRAY FOR "POLYGON" PROCEDURE
  ListType = array of SmallInt;
  LightType = record
                X, Y, Z: double;
              end;
var
  NumW, NumH, Nface, Nvert, Radius: Word;
  SortKeys: array of double;
  SortLinks: ListType;
  P: Tpoint2d;
  Light: LightType;
  Scale: Double;
  WireFrame: Boolean;
  ScreenDistance: SmallInt;
  // variables for  arm
  WristBase, KnuckleBase, PalmRight, PalmLeft, Shoulder, FingerTip,
  Target: pointtype;

procedure ClearBitmap(BkColor: TColor; Bitmap: TBitmap);
procedure LimitAngle(var A: Double);
procedure Rotate3D(var Source: FaceArrayType;  //rotate 3D points in place
                   Axis: Char;            // rotate about 'x', 'y', or 'z'
                   Angle: Double;        // a360 = 2pi
                   StartAt, EndAt: Integer;
                   x0, y0, z0: Double); // center of rotation
procedure ProjectFaces(var Source: FaceArrayType; X, Y, Z: Integer);
procedure ShowFaces(var Source: FaceArrayType; Bm: TBitmap);
procedure ShowArray(var Source: FaceArrayType; N: Byte; Bm: TBitmap);
procedure CalcLighting(var S: FaceArrayType);

implementation

procedure ClearBitmap(BkColor: TColor; Bitmap: TBitmap);
begin
  with Bitmap, Canvas do
  begin
    Brush.Color := BkColor;
    FillRect(Rect(0, 0, Width, Height));
  end;
end;

procedure LimitAngle(var A: double);
begin
  while A < 0 do A := A + a360;
  while A > a360 do A := A - a360;
end;

{ QUICKSORT sorts elements in the array A with indices between  }
{ LO and HI (both inclusive). It executes itself recursively.}

procedure QuickSort(L, R: word);
var
  I, J, Yy: Word;  //XX deleted
  X: Double;       //Y deleted

  function GetValue(i: Integer): Double; // sort only the pointers to the keys.
  var
    N: Integer;
  begin
    N := I;
    if N < 0 then N := 0;
    if N > Length(SortKeys) then N := Length(SortKeys);
    Result := SortKeys[SortLinks[N]]
  end;

begin
  I := L; J := R; X := (GetValue(L)+ GetValue(R)) / 2;
  repeat
    while GetValue(I) > X do I := I + 1;
    while X > GetValue(J) do J := J - 1;
    if I <= J then
    begin
      Yy := SortLinks[I];
      SortLinks[I] := SortLinks[J];
      SortLinks[J] := Yy;
      I := I + 1; J := J - 1;
    end;
  until I > J;
  if L < J then QuickSort(L, J);
  if I < R then QuickSort(I, R);
end;

// Sort indexes to the faces by the average z distance to the faces.
procedure SortByDistance(var Source: FaceArrayType);
var
  I, J: SmallInt;
  AvgZ: Double;
begin
  SetLength(SortKeys, Nface);  // allocate dynamic arrays
  SetLength(SortLinks, Nface);
  for I := 0 to Nface - 1 do  // scan through all faces
  begin
    AvgZ := 0;
    for J := 0 to Nvert - 1 do AvgZ := AvgZ + Source[I].F[J].Z3D;
    AvgZ := AvgZ/Nvert;   // avg z-distance of vertex
    SortLinks[I] := I;        // load index array in order of indices
    SortKeys[I] := AvgZ;
  end;
  QuickSort(0, Nface-1);      // rearrange the links in place
end;

procedure ShowArray(var Source: FaceArrayType;
                    N: Byte; Bm: TBitmap);
var
  I, J: Word;
  X, Y: SmallInt;
begin
  for I := 0 to Nface-1 do
    for J := 0 to Nvert do
      with Source[I].F[J mod Nvert] do   // wrap around to starting vertex
         with Bm, Canvas do
         begin
           Pen.Color := clRed;
           if N = 3 then
           begin
             X := Width div 2 + Round(X3D); Y := Height div 2 - Round(Y3D);
           end
           else
           begin
             X := Width div 2 + Round(x2d); Y := Height div 2 - Round(y2d);
           end;
           if J  = 0 then MoveTo(X, Y) else LineTo(X, Y);
         end;
end;

procedure ShowFaces(var Source: FaceArrayType; Bm: TBitmap);
var
  I, J, K: word;
begin
  SortByDistance(Source);
  SetLength(P, Nvert);  // allocate dynamic array P
  for K := 0 to Nface - 1 do
  with Bm, Canvas do
  begin
    I := SortLinks[K];
    for J := 0 to NVert - 1 do
      with Source[I] do
      begin
        with F[J] do
        begin
          P[J].X := Width div 2 + Round(x2d) ;
          P[J].Y := Height div 2 - Round(y2d);
        end;
        Brush.Color := Cr;
      end;
      Pen.Color := Brush.Color;   // and pen color
      Polygon(P);                 // draw face to bitmap
  end;
end;

procedure Calclighting(var S: FaceArrayType);
var
  L1, M1, N1, L2, M2, N2: Double;
  Lnorm, Mnorm, Nnorm: Double;
  Temp: Double;
  Denom: Double;
  Brt: Double;
  R, G, B: Byte;
  I,J: Integer;
  lineOK: Boolean;
begin
  for I := 0 to Nface - 1 do
  with S[I] do
  begin
    j := -1;
    lineOK := true;
    repeat
     inc(j);
     if (
          (F[j+1].X3D = F[j].X3D) and
          (F[j+1].Y3D = F[j].Y3D) and
          (F[j+1].Z3D = F[j].Z3D)
         ) then lineOK := false;
    until lineOK or (j = NVert - 2);

    if lineOK then
    begin                     // direction numbers of successive lines

     L1 := F[(j+1) mod NVert].X3D - F[j].X3D;
     M1 := F[(j+1) mod NVert].Y3D - F[j].Y3D;
     N1 := F[(j+1) mod NVert].Z3D - F[j].Z3D;

     L2 := F[(j+2) mod NVert].X3D - F[(j+1) mod NVert].X3D;
     M2 := F[(j+2) mod NVert].Y3D - F[(j+1) mod NVert].Y3D;
     N2 := F[(j+2) mod NVert].Z3D - F[(j+1) mod NVert].Z3D;

    Lnorm := (M1 * N2) - (M2 * N1); // direction cosines of normal to plane
    Mnorm := (N1 * L2) - (N2 * L1);
    Nnorm := (L1 * M2) - (L2 * M1);
    Temp := ((Light.X * Lnorm) + (Light.Y * Mnorm) + (Light.Z * Nnorm));
    Denom := Sqrt(Sqr(Lnorm) + Sqr(Mnorm) + Sqr(Nnorm));
    if Denom = 0 then Brt := 1.0 else Brt := 0.4 - (0.6 * Temp / Denom);
    if Brt < 0.25 then Brt := 0.25;
   end
   else
    BRT := 0.25;
  B := Round(((C and $00ff0000) shr 16) * Brt);
  G := Round(((C and $0000ff00) shr  8) * Brt);
  R := Round((C and $000000ff)          * Brt);
  Cr := RGB(R, G, B);
 end;
end;

// ROTATE ALL POINTS ABOUT ONE AXIS, X, Y, OR Z. REPLACE PREVIOUS COORDINATES
procedure Rotate3D(var Source: FaceArrayType; //rotate 3D points in place
                   Axis: Char;                // rotate about 'x', 'y', or 'z'
                   Angle: Double;            // a360 = 2pi
                   StartAt, EndAt: Integer;
                   x0, y0, z0: double);     // center of rotation
var
  I, J: SmallInt;
  Ca, Sa: Double;
  X, Y, Z, Xp, Yp, Zp: Double;
begin
  LimitAngle(Angle);
  for I := StartAt * NumW to EndAt * NumW - 1 do
  begin
    for J := 0 to Nvert - 1 do
      with Source[I] do
      begin
        with F[J] do
        begin
          X := X3D - x0;     // REMOVE CENTER OF ROTATION
          Y := Y3D - y0;
          Z := Z3D - z0;
        end;
        with A do           // ADD ROTATIONS TO ANGLES (DIRECTION NUMBERS)
          case Axis of
          'x': Theta := Theta0 + Angle;
          'y': Phi := Phi0 + Angle;
          'z': Rho := Rho0 + Angle;
          end;
        Ca := Cos(Angle); Sa := Sin(Angle);
        case Axis of
          'x': begin
                 Xp := X;
                 Yp := Ca * Y + Sa * Z;
                 Zp := -Sa * Y + Ca * Z;
               end;
          'y': begin
                 Xp := Ca * X + Sa * Z;
                 Yp := Y;
                 Zp := -Sa * X + Ca * Z;
               end;
          'z': begin
                 Xp := Ca * X + Sa * Y;
                 Yp := -Sa * X + Ca * Y;
                 Zp := Z;
               end;
          else begin
                 Xp := X;  // if illegal axis, don't rotate
                 Yp := Y;
                 Zp := Z;
                end;
        end;
        with F[J] do
        begin
          X3D := Xp + x0;   // RESTORE OFFSET FROM AXIS OF ROTATION
          Y3D := Yp + y0;
          Z3D := Zp + z0;
        end;
      end; // of <with source[i]>
      if not WireFrame then CalcLighting(Source);
  end;
end;

// PROJECT ALL POINTS FOR DRAWING
procedure ProjectFaces(var Source: FaceArrayType; X, Y, Z: Integer);
var
  I, J, D: SmallInt;
begin
  for I := 0 to Nface - 1 do
    for J := 0 to Nvert - 1 do
      with Source[I].F[J] do
      begin
        D := Round(Scale * Z3D + Z); // avoid division by zero, but don't change Z3D
        if D = -ScreenDistance then Inc(D);
        x2d := round(Scale * X3D + X * ScreenDistance / (ScreenDistance + D));
        y2d := round(Scale * Y3D + Y * ScreenDistance / (ScreenDistance + D));
      end;
end;

end.
