unit Main;

//Program: ArmTrackTarget.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ExtCtrls, Math;

type
  TMainForm = class(TForm)
    ControlPanel: TGroupBox;
    DisplayPanel: TGroupBox;
    ArmControls: TGroupBox;
    DisplayMainPanel: TPanel;
    DisplayBotPanel: TPanel;
    RotateHorizTrack: TTrackBar;
    DisplayRtPanel: TPanel;
    RotateVertTrack: TTrackBar;
    HandPanel: TGroupBox;
    ReachPanel: TGroupBox;
    PitchTrack: TTrackBar;
    YawTrack: TTrackBar;
    RollTrack: TTrackBar;
    ShoulderTrack: TTrackBar;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label8: TLabel;
    GroupBox1: TGroupBox;
    AboutBtn: TButton;
    ExitBtn: TButton;
    ArmPaintBox: TPaintBox;
    GainBox: TGroupBox;
    KoTrack: TTrackBar;
    KdTrack: TTrackBar;
    KoLabel: TLabel;
    KdLabel: TLabel;
    GainResetBtn: TButton;
    RotateViewHorizLabel: TLabel;
    RtDisplayPaint: TPaintBox;
    TargetBox: TGroupBox;
    TargetPanel: TPanel;
    Panel2: TPanel;
    TargetZTrack: TTrackBar;
    Label1: TLabel;
    KvTrack: TTrackBar;
    KvLabel: TLabel;
    DebugPanel: TPanel;
    HandPitchRLabel: TLabel;
    HandPitchPLabel: TLabel;
    WristPitchRLabel: TLabel;
    WristPitchPLabel: TLabel;
    ElbowPitchRLabel: TLabel;
    ElbowPitchPLabel: TLabel;
    ShoulderPitchRLabel: TLabel;
    ShoulderPitchPLabel: TLabel;
    HandRollRLabel: TLabel;
    HandRollPLabel: TLabel;
    HandYawRLabel: TLabel;
    HandYawPLabel: TLabel;
    procedure ExitBtnClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RotateHorizTrackChange(Sender: TObject);
    procedure RotateVertTrackChange(Sender: TObject);
    procedure PitchTrackChange(Sender: TObject);
    procedure YawTrackChange(Sender: TObject);
    procedure RollTrackChange(Sender: TObject);
    procedure KoTrackChange(Sender: TObject);
    procedure KdTrackChange(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
    procedure GainResetBtnClick(Sender: TObject);
    procedure RtDisplayPaintPaint(Sender: TObject);
    procedure TargetPanelMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure TargetZTrackChange(Sender: TObject);
    procedure ShoulderTrackChange(Sender: TObject);
    procedure KvTrackChange(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure IdleLoop(Sender: TObject; var Done: Boolean);
    procedure SetDefaultGains;
    procedure InitControl;
    procedure Level3Control;
    procedure Iterate;
  end;

var
  MainForm: TMainForm;

implementation

uses FloatArmTrk, FloatGraphics3D, About, AngleFont;

{$R *.dfm}

const
  MaxLev1Sys = 14;
  MaxLev2Sys = 4;
  MilliPi = Pi/1000.0;

  QiLimits: array[1..MaxLev1Sys] of
             record
                QiMin, QiMax: Double;
  end = ((QiMin: -pi/2; QiMax: pi/2),    // ShldY.
         (QiMin: -0.5*pi; QiMax: 0.7*pi),// ShldP.
         (QiMin: -pi/2; QiMax: pi/2),    // ShldR.
         (QiMin:     0; QiMax: 0.9*pi),  // ElbowP.
         (QiMin: -pi; QiMax: pi/2),      // ForeR
         (QiMin: -pi/2; QiMax: pi/2),    // WrPitch
         (QiMin: -pi/8; QiMax: pi/4),    // WrYaw
         (QiMin: -pi/2; QiMax: pi/2),    // Th1Yaw
         (QiMin: -pi/8; QiMax: 0.4*pi),  // Th1Pitch
         (QiMin: -pi/8; QiMax:  pi/8),   // Th2Pitch
         (QiMin:  0;    QiMax:  pi/4),   // Th3Pitch
         (QiMin:  0;    QiMax: pi/2),    // F1Pitch
         (QiMin:  0;    QiMax: pi/2),    // F2Pitch
         (QiMin:  0;    QiMax: pi/4));   // F3Pitch

type
  CSystype = record
               V, M, P, E, Ev, R, O, Rv, Qi, Qo, F, Dist: Double;
               D, D1, D2: Double;
             end;

var
  Stage: Integer;
  T1, T2, Freq: Int64;
  BM: TBitmap;
  ArmScale: Double;

  ShoulderRoll, ShoulderRollRef, ShoulderRollError,
  HandElev, HandElevRef, HandElevError,
  HandLateral, HandLateralRef, HandLateralError,
  HandPitch, HandPitchRef, HandPitchError,
  HandYaw, HandYawRef, HandYawError,
  HandRoll, HandRollRef, HandRollError,
  Reach, ReachRef, ReachError,

  TargetXError, TargetYError, TargetZError: Double;
  TargRadius: Double;

  CSysLev1: array[1..MaxLev1Sys] of CSysType;
  CSysLev2: array[1..MaxLev2Sys] of CSysType;
  dt: Double;
  Ko, Kv, Kd: Double;
  OutStr: string;
  OneDouble: Double = 1.0;


procedure TMainForm.SetDefaultGains;
begin
  Ko := 0.3;
  Kv := 130.0;
  Kd := 0.0;
  KoTrack.Position := Round(Ko*100);
  KvTrack.Position := Round(Kv*10);
  KdTrack.Position := Round(Kd*20);
  Str(Ko:4:2, OutStr);
  KoLabel.Caption := 'Ko = ' + OutStr;
  Str(Kv:4:2, OutStr);
  KvLabel.Caption := 'Kv = ' + OutStr;
  Str(Kd:4:2, OutStr);
  KdLabel.Caption := 'Kd = ' + OutStr;
end;

procedure TMainForm.InitControl;
var
  I: Integer;
begin
  MouseZ := TargetZTrack.Position;
  ViewX := RotateHorizTrack.Position * MilliPi;
  ViewY := RotateVertTrack.Position * MilliPi;
  HandPitchRef := PitchTrack.Position * MilliPi;
  HandYawRef := YawTrack.Position * MilliPi;
  HandRollRef := RollTrack.Position * MilliPi;
  for I := 1 to MaxLev1Sys do
  with CSysLev1[I] do
  begin
    R := 0.0;
    Qo := 0.0;
    M := 3.0; //Arm mass
   end;
  SetDefaultGains;
  dt := 1/60.0;
end;

procedure ErrLim(var E: Double; Min, Max: Double);
begin
  if E > Max then E := Max
  else
    if E < Min then E := Min;
end;

var
  Count: Integer;

//For testing only; generates signal to vary references
function SignalGen(Sig: Double): Double;
begin
  Inc(Count);
  if Abs(Count) > 180 then
  begin
    Count := 0;
    OneDouble := -OneDouble;
  end;
  Result := Sig*OneDouble;
end;

procedure TMainForm.Level3Control;
begin
  //Get the target position
  Target.X3D := ArmSegXYZ[18].X;
  Target.Y3D := ArmSegXYZ[18].Y;
  Target.Z3D := ArmSegXYZ[18].Z;

  //Compute reach reference based on target's X position
  TargetXError := FingerTip.X3D - Target.X3D;
  ReachRef := 250 - 15.0*TargetXError; //Slowing
  ErrLim(ReachRef, -1000, 1000);

  //Compute hand elevation reference based on target's Y position
  TargetYError := FingerTip.Y3D - Target.Y3D;
  HandElevRef :=  - 15.0*TargetYError;
  ErrLim(HandElevRef, -900, 900);

  //Compute hand lateral position based on target's Z position
  TargetZError := FingerTip.Z3D - Target.Z3D;
  HandLateralRef := - 15.0*TargetZError;
  ErrLim(HandLateralRef, -900, 900);
end;

procedure Level2Control;
var
  SinR: Double;  // sin hand roll angle, sin hand pitch angle
  CosR: Double;  // cos hand roll angle, cos hand pitch angle
  Temp: Double; // hypotenuses;
begin
//Hand reach error
  //Reach is 3-D distance between shoulder and finger-tip
  Reach := Sqrt(Sqr(FingerTip.X3D - Shoulder.X3D) +
                Sqr(FingerTip.Y3D - Shoulder.Y3D) +
                Sqr(FingerTip.Z3D - Shoulder.Z3D));
  ReachError := 0.5*(ReachRef - Reach);

  //Hand elevation error
  HandElev := FingerTip.Y3D - Shoulder.Y3D;
  HandElevError := 1.0*(HandElevRef - HandElev);

  //Hand lateral error
  HandLateral := FingerTip.Z3D - Shoulder.Z3D;
  HandLateralError := 1.0*(HandLateralRef - HandLateral);

//Hand pitch error
  Temp := Hypot((KnuckleBase.X3D-WristBase.X3D),
                (KnuckleBase.Z3D-WristBase.Z3D));
  if Temp <> 0 then
  HandPitch := ArcTan2(KnuckleBase.Y3D - WristBase.Y3D, Temp);
  HandPitchError := 3000*(HandPitchRef - HandPitch);
  ErrLim(HandPitchError, -500, 500);

//Hand yaw error
  Temp := Hypot((KnuckleBase.X3D-WristBase.X3D),
                (KnuckleBase.Y3D-WristBase.Y3D));
  HandYaw := ArcTan2(KnuckleBase.z3d - WristBase.z3d,Temp);
  HandYawError := 3000*(HandYawRef - HandYaw);
  ErrLim(HandYawError, -500, 500);

  ShoulderRoll := ShldR;
  ShoulderRollError := 1000*(ShoulderRollRef - ShoulderRoll);

//Hand roll error
  HandRoll := ShldR + ForeR;
  HandRollError := 3000*(HandRollRef - HandRoll);

  SinR :=  Sin(HandRoll);
  CosR :=  Cos(HandRoll);
  CSysLev1[1].R := -HandLateralError;                         //Shoulder yaw
  CSysLev1[2].R := (-ReachError - HandElevError);             //Shoulder pitch
  CSysLev1[3].R := ShoulderRollError;                          //Shoulder roll
  CSysLev1[4].R := -2.0*ReachError;                           //Elbow pitch
  CsysLev1[5].R := HandRollError;                             //Forearm roll
  CSysLev1[6].R := -HandPitchError*SinR - HandYawError*CosR;  //Wrist pitch
  CSysLev1[7].R := HandYawError*SinR - HandPitchError*CosR;   //Wrist yaw
  end;

procedure Level1Control; //Joint Angles
var
  I: Integer;
begin
// Control Systems, 1 to MaxLev1Sys -- all joint angles
  for I := 1 to MaxLev1Sys do
  begin
    with CSysLev1[I] do
    begin
//      if i = 3 then R := signalgen(50) else R := 0;
//    Uncomment above line to generate test signal
      Ev := Rv - V;           // Velocity error signal
      F  := Kv*Ev;
      V  := V + (F/M)*dt;    // Velocity = integral of output forces
      if V > 1000 then V := 1000
      else if V < -1000 then V := -1000;

      Qi := Qi + V*dt;       // Position = integral of velocity; v*dt = posit.
      if Qi > 3 then Qi := 3
      else if Qi < -3 then Qi := -3;
      P := Qi;
      E := R - P;            // Position Error
      Rv  := E*Ko*dt;        // Velocity reference signal
    end;
  end;
  // Keep joint Positions within physical limits
  for I := 1 to MaxLev1Sys do
  begin
    with QiLimits[I], CSysLev1[I] do
    begin
      if Qi < QiMin then Qi := QiMin
      else if Qi > QiMax then Qi := QiMax;
      if R < QiMin then R := QiMin
      else if R > QiMax then R := QiMax;
    end;
  end;
  // Load values for displaying arm
  ShldY  := CSysLev1[01].Qi;
  ShldP  := CSysLev1[02].Qi;
  ShldR  := CSysLev1[03].Qi;
  ElbowP := CSysLev1[04].Qi;
  ForeR  := CSysLev1[05].Qi;
  WrP    := CSysLev1[06].Qi;
  WrY    := CSysLev1[07].Qi;
  Th1R   := CSysLev1[08].Qi;
  Th1P   := CSysLev1[09].Qi;
  Th2P   := CSysLev1[10].Qi;
  Th3P   := CSysLev1[11].Qi;
  F1P    := -CSysLev1[12].Qi;
  F2P    := -CSysLev1[13].Qi;
  F3P    := -CSysLev1[14].Qi;
end;

procedure TMainForm.Iterate;
begin
  OperateArm(BM);
  ArmPaintBox.Canvas.Draw(0, 0, BM);
  Level3Control;
  Level2Control;
  Level1Control;

//Information for Debug Panel
//Set DebugPanel's Visibility property to True to see panel at runtime
  Str(RadToDeg(HandPitchRef):4:2, OutStr);
  HandPitchRLabel.Caption     := 'Hand Pitch R    = ' + OutStr;
  Str(RadToDeg(HandPitch):4:2, OutStr);
  HandPitchPLabel.Caption     := 'Hand Pitch       = ' + OutStr;

  Str(RadToDeg(HandRollRef):4:2, OutStr);
  HandRollRLabel.Caption      := 'Hand Yaw R      = ' + OutStr;
  Str(RadToDeg(HandRoll):4:2, OutStr);
  HandRollPLabel.Caption      := 'Hand Yaw         = ' + OutStr;

  Str(RadToDeg(HandRollRef):4:2, OutStr);
  HandYawRLabel.Caption       := 'Hand Roll R     = ' + OutStr;
  Str(RadToDeg(HandRoll):4:2, OutStr);
  HandYawPLabel.Caption       := 'Hand Roll        = ' + OutStr;

  Str(RadToDeg(CSysLev1[6].R):4:2, OutStr);
  WristPitchRLabel.Caption    := 'Wrist Pitch R   = ' + OutStr;
  Str(RadToDeg(CSysLev1[6].Qi):4:2, OutStr);
    WristPitchPLabel.Caption  := 'Wrist Pitch      = ' + OutStr;

  Str(RadToDeg(CSysLev1[4].R):4:2, OutStr);
  ElbowPitchRLabel.Caption    := 'Elbow Pitch R   = ' + OutStr;
  Str(RadToDeg(CSysLev1[4].Qi):4:2, OutStr);
  ElbowPitchPLabel.Caption    := 'Elbow Pitch      = ' + OutStr;

  Str(RadToDeg(CSysLev1[2].R):4:2, OutStr);
  ShoulderPitchRLabel.Caption := 'Shoulder Pitch R = ' + OutStr;
  Str(RadToDeg(CSysLev1[2].Qi):4:2, OutStr);
  ShoulderPitchPLabel.Caption := 'Shoulder Pitch    = ' + OutStr;

end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: Boolean);
begin
  Done := False;
  Case Stage of
  0: begin //Initialize
       WindowState := wsMaximized;
       InitControl;
       ArmScale := 1.5 * ArmPaintBox.Width/713;
       InitArm(ArmScale);
       Stage := 1;
     end;
  1: Iterate; //Main program loop
  end;
  repeat
    QueryPerformanceCounter(T2);
    Application.ProcessMessages;
  until (T2 - T1) >= Freq/60;
  T1:= T2;
end;

//---------- Hand Pitch, Yaw, and Roll Controls----------
procedure TMainForm.PitchTrackChange(Sender: TObject);
begin
  HandPitchRef := PitchTrack.Position * MilliPi;
  Str(RadToDeg(HandPitchRef):4:2, OutStr);
end;

procedure TMainForm.YawTrackChange(Sender: TObject);
begin
  HandYawRef := YawTrack.Position * MilliPi;
end;

procedure TMainForm.RollTrackChange(Sender: TObject);
begin
  HandRollRef := 0.5*RollTrack.Position * MilliPi;
end;

//---------- Shoulder Roll Control ----------
procedure TMainForm.ShoulderTrackChange(Sender: TObject);
begin
 ShoulderRollRef := 0.5*ShoulderTrack.Position * MilliPi;
end;

//---------- Target Positioning Controls ----------
procedure TMainForm.TargetPanelMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if ssLeft in Shift then
  begin
    if (X >=0) and (X < TargetPanel.Width) and
      (Y >=0) and (Y < TargetPanel.Height) then
    begin
      MouseX := TargetPanel.Width div 2 - X;
      MouseY := TargetPanel.Height div 2 - Y;
    end;
  end;
end;

procedure TMainForm.TargetZTrackChange(Sender: TObject);
begin
  MouseZ := TargetZTrack.Position;
end;

//---------- Viewing Angle Controls ----------
procedure TMainForm.RotateHorizTrackChange(Sender: TObject);
begin
  ViewX := RotateHorizTrack.Position * MilliPi;
end;

procedure TMainForm.RotateVertTrackChange(Sender: TObject);
begin
  ViewY := RotateVertTrack.Position * MilliPi;
end;

//------------ Gain Sliders ------------
// Set GainBox's Visible property to true to see these at runtime

procedure TMainForm.KoTrackChange(Sender: TObject);
begin
 Ko := 0.01*KoTrack.Position;
 Str(Ko:4:2, OutStr);
 KoLabel.Caption := 'Ko = ' + OutStr;
end;

procedure TMainForm.KvTrackChange(Sender: TObject);
begin
 Kv := 0.1*KvTrack.Position;
 Str(Kv:4:2, OutStr);
 KvLabel.Caption := 'Kv = ' + OutStr;
end;

procedure TMainForm.KdTrackChange(Sender: TObject);
begin
 Kd := 0.05*KdTrack.Position;
 Str(Kd:4:2, OutStr);
 KdLabel.Caption := 'Kd = ' + OutStr;
end;

//---------- Gain Reset Buttion ----------
procedure TMainForm.GainResetBtnClick(Sender: TObject);
begin
  SetDefaultGains;
end;

//---------- Program Exit Button ----------
procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Close;
end;

//---------- About Box Button ----------
procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

//---------- Initialize On Form-Creation ----------
procedure TMainForm.FormCreate(Sender: TObject);
begin
  BM := TBitmap.Create;
  Stage := 0;
  ChangeScale(Screen.Width, 1280);
  Application.OnIdle := IdleLoop;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(T1);
end;

//---------- Form Resizing ----------
procedure TMainForm.FormResize(Sender: TObject);
begin
  BM.Width := ArmPaintBox.Width;
  BM.Height := ArmPaintBox.Height;
  RotateVertTrack.Top := DisplayRtPanel.Height div 2
    - RotateVertTrack.Height div 2;
  RotateHorizTrack.Left := (DisplayBotPanel.Width -DisplayRtPanel.Width) div 2
    - RotateHorizTrack.Width div 2;
  RotateViewHorizLabel.Left := RotateHorizTrack.Left
    + RotateHorizTrack.Width div 2 - RotateViewHorizLabel.Width div 2;
  ArmScale := 1.5 * ArmPaintBox.Width/713;
  InitArm(ArmScale);
end;

//---------- Painting Events ----------
procedure TMainForm.FormPaint(Sender: TObject);
begin
  ArmPaintBox.Canvas.Draw(0, 0, BM);
end;

procedure TMainForm.RtDisplayPaintPaint(Sender: TObject);
var
  X, Y: Integer;
  S: String;
begin
  with RtDisplayPaint do
  begin
    S := 'Rotate View About Horizontal Axis';
    X := RotateVertTrack.Left + RotateVertTrack.Width + 5;
    Y := RotateVertTrack.Top + RotateVertTrack.Height div 2
         + Canvas.TextWidth(S) div 2;
    TextOutA(Canvas, X, Y, 900, S);
  end;

end;

//---------- Form OnDestroy Event ----------
procedure TMainForm.FormDestroy(Sender: TObject);
begin
  BM.Free; //Free the off-screen arm-display bitmap;
end;

end.


