unit Main;

//Program: SquareCircle.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Math;

type
  TMainForm = class(TForm)
    ControlPanel: TGroupBox;
    DisplayPanel: TGroupBox;
    DisplayPaint: TPaintBox;
    ProgramBox: TGroupBox;
    AboutBtn: TButton;
    ExitBtn: TButton;
    KeyboardBox: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    SpaceBarStaticText: TStaticText;
    procedure FormCreate(Sender: TObject);
    procedure ExitBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure DisplayPaintMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure AboutBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure DrawTriangle(R: Double; C: TColor);
    procedure DrawCircle(R: Double; C: TColor);
    procedure TraceImage;
    procedure ScreenInit;
    procedure Iterate;
    procedure IdleLoop(Sender: TObject; var Done: boolean);
    procedure CheckKey(C: char);
    procedure ReStartClick(Sender: TObject);
  end;

var
  MainForm: TMainForm;
  MouseBitmap, SquareBitmap: TBitmap;

implementation

uses About;

{$R *.dfm}

const
  Kr = pi/180.0;

var
  Theta, D, R, Radius, MouseRadius, SquareRadius: Double;
  Err, ErrInt, A, dt: Double;
  MouseX, MouseY, HCenter, VCenter, Stage: Integer;
  SquareOn, Both, CircleDraw, Direct: Boolean;
  t1, t2, Freq: Int64;

function RadiusToTriangle(Theta: Double): Double;
begin
  if (Theta >= 330) and (Theta < 360)
    or (Theta > 0) and (Theta < 90) then
    D := R*Sin(Kr*30)/Sin(Kr*(60 + Theta))
    else
      if (Theta >= 90) and (Theta < 210) then
      D := R*Sin(Kr*30)/Sin(Kr*(Theta - 60))
      else
        if (Theta >= 210) and (Theta < 330) then
        D := -R*Sin(Kr*30)/Sin(Kr*(Theta));
  Result := D;
end;

function RadiusToSquare(Theta: Double): Double;
begin
  if ((Theta >= 315) and (Theta < 360))
    or ((Theta >= 0) and (Theta < 45)) then
    D := R/Cos(Kr*Theta)
    else
      if (Theta >= 45) and (Theta < 135) then
        D := R/Sin(Kr*Theta)
        else
          if (Theta >= 135) and (Theta < 225) then
            D := -R/Cos(Kr*Theta)
            else
              if (Theta >= 225) and (Theta < 315) then
              D := -R/Sin(Kr*Theta);
  Result := D;
end;

procedure TMainForm.DrawTriangle(R: Double; C: TColor);
begin
  with MouseBitmap, Canvas do
  begin
    Pen.Color := C;
    Pen.Width := 3;
    MoveTo(HCenter, VCenter - Round(R));
    LineTo(HCenter - Round(A*R*Sin(Kr*60)), VCenter + Round(R*Cos(Kr*60)));
    LineTo(HCenter + Round(A*R*Sin(Kr*60)), VCenter + Round(R*Cos(Kr*60)));
    LineTo(HCenter, VCenter - Round(R));
    if Direct then TextOut(HCenter - TextWidth('DIRECT') div 2,
      VCenter,'DIRECT');
  end;
end;

procedure TMainForm.DrawCircle(R: Double; C: TColor);
begin
  with MouseBitmap, Canvas do
  begin
    Pen.Color := C;
    Pen.Width := 3;
    Ellipse(HCenter - Round(A*R), VCenter - Round(R), HCenter + Round(A*R),
      VCenter + Round(R));
    if Direct then TextOut(HCenter - TextWidth('DIRECT') div 2,
      VCenter,'DIRECT');
  end;
end;

procedure TMainForm.TraceImage;
var
  X, Y: Integer;
  Rad2Square: Double;
begin
  with MouseBitmap, Canvas do
  begin

    X := MouseX; Y := MouseY;
    Pen.Color := clAqua;
    Ellipse(X-2, Y-2, X+2, Y+2);
  end;
  with SquareBitmap, Canvas do
  begin
    X := MouseX - HCenter;
    Y := VCenter - MouseY;
    Theta := 180/pi*ArcTan2(Y, X); // -pi < theta < pi
    if Theta < 0 then Theta := Theta + 360;
    if CircleDraw then Radius := R
      else Radius := RadiusToTriangle(Theta{,R});
    Rad2Square := RadiusToSquare(Theta);
    MouseRadius := Sqrt(Sqr(X) + Sqr(Y));
    Err := MouseRadius - Radius;
    ErrInt := ErrInt + Err*dt;
    if ErrInt > 0.5*R then ErrInt := 0.5*R
    else
      if ErrInt < 0.1*R then ErrInt := 0.1*R;
    if not Direct then SquareRadius := 3.0*ErrInt + 0.7*Err
      else SquareRadius := Rad2Square + Err;
    if SquareRadius < 0 then SquareRadius := 0;
    X := HCenter + Round(SquareRadius*Cos(Kr*Theta));
    Y := VCenter - Round(SquareRadius*Sin(Kr*Theta));
    Pen.Color := clWhite;
    Pen.Width := 4;
    Ellipse(X-1, Y-1, X, Y);
  end;
end;

procedure TMainForm.ScreenInit;
begin
  R := 200;
  with DisplayPaint, Canvas do
  begin
    CopyMode := cmSrcCopy;
    Pen.Color := clWhite;
    Brush.Color := clBlack;
    FillRect(Rect(0, 0, ClientWidth, ClientHeight));
    HCenter := ClientWidth div 2;
    VCenter := ClientHeight div 2;
  end;
  with MouseBitmap, Canvas do   // clear bitmap
  begin
    Font.Color := clWhite;
    Brush.Color := clBlack;
    FillRect(Rect(0, 0, Width-1, Height-1));
  end;
  with SquareBitmap, Canvas do // clear screen and draw square
  begin
    Font.Color := clWhite;
    Pen.Width := 4;
    Pen.Color := clRed;
    Brush.Color := clBlack;
    FillRect(Rect(0, 0, Width-1, Height-1));
    Rectangle(HCenter - Round(A*R), VCenter - Round(R),
            HCenter + Round(A*R), VCenter + Round(R));
    if Direct then TextOut(HCenter - TextWidth('DIRECT') div 2,
      VCenter,'DIRECT');
  end;
  SquareRadius := 0;
  SpaceBarStaticText.SetFocus; // Prevents space bar from activating OnKeyPress
  // event handler of About Box by changing focus to this StaticText control.
end;

procedure TMainForm.Iterate;
begin
  with DisplayPaint do
    case Stage of
      0: begin
           WindowState := wsMaximized;
           A := 1;
           SquareOn := True;
           CircleDraw := True;
           Direct := False;
           with MouseBitmap do
           begin
             Width := ClientWidth;
             Height := ClientHeight;
           end;
           with SquareBitmap do
           begin
             Width := ClientWidth;
             Height := ClientHeight;
           end;
           Stage := 1;
       end;
    1: begin
         ScreenInit;
         ErrInt := 100.0;
         Cursor := crNone;
         if CircleDraw then DrawCircle(R, clYellow)
           else DrawTriangle(R, clYellow);
         Stage := 2;
       end;
    2: begin
         TraceImage;
         if Both then Canvas.CopyMode := cmSrcPaint
           else Canvas.CopyMode := cmSrcCopy;
         if not SquareOn or Both then Canvas.Draw(0, 0, MouseBitmap);
         if SquareOn or Both then Canvas.Draw(0, 0, SquareBitmap);
       end;
    end;
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: boolean);
begin
  Done := False;
  QueryPerformanceCounter(t2);
  if (t2 - t1) < Freq/60 then Exit;
  Iterate;
  t1 := t2;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  MouseBitmap := TBitmap.Create;
  SquareBitmap := TBitmap.Create;
  ChangeScale(Screen.Width, 1280);
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(t1);
  dt := 0.01;
  Stage := 0;
  CircleDraw := True;
  Application.OnIdle := IdleLoop;
end;

procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  MouseBitmap.Free;
  SquareBitmap.Free;
end;

procedure TMainForm.CheckKey(C: Char);
begin
 case C of
  'v','V': begin
             SquareOn := not SquareOn;
             if SquareOn then
               DisplayPaint.Canvas.Draw(0, 0, MouseBitmap)
               else
                 DisplayPaint.Canvas.Draw(0, 0, SquareBitmap);
           end;
  'b','B': Both := not Both;
  ' ':     begin
             ScreenInit;
             if CircleDraw then DrawCircle(R, clYellow)
             else DrawTriangle(R, clYellow);
             Stage := 2;
           end;
  'c','C': begin
             CircleDraw := True;
             ScreenInit;
             DrawCircle(R, clYellow)
           end;
  'd','D': begin
             Direct := not Direct;
             ScreenInit;
             if CircleDraw then DrawCircle(R, clYellow)
               else DrawTriangle(R, clYellow);
             Stage := 2;
           end;
  't','T': begin
             CircleDraw := False;
             ScreenInit;
             DrawTriangle(R, clYellow);
           end;
  end;
end;

procedure TMainForm.ReStartClick(Sender: TObject);
begin
 SquareOn := True;
 Both := False;
 Stage := 1;
end;


procedure TMainForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  CheckKey(Key);
end;

procedure TMainForm.DisplayPaintMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  MouseX := X;
  MouseY := Y;
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  AboutBox.ShowModal;
  SpaceBarStaticText.SetFocus; //Remove focus from AboutBoxBtn
end;

end.
