unit Main;

//Program: Crowd.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
//Copyright: 2009

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Math;

type
  TMainForm = class(TForm)
    ControlPanel: TGroupBox;
    ProgramBox: TGroupBox;
    AboutBtn: TButton;
    ExitBtn: TButton;
    DisplayPanel: TGroupBox;
    DisplayPaint: TPaintBox;
    SimPanel: TGroupBox;
    PauseBtn: TButton;
    RunBtn: TButton;
    NewFileBtn: TButton;
    TrailsBtn: TButton;
    LoadParamDlg: TOpenDialog;
    SimNameEdit: TEdit;
    EditSimBtn: TButton;
    SaveParamsBtn: TButton;
    SaveParamDlg: TSaveDialog;
    CurSimLabel: TStaticText;
    InstructPanel: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    procedure ExitBtnClick(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RunBtnClick(Sender: TObject);
    procedure PauseBtnClick(Sender: TObject);
    procedure TrailsBtnClick(Sender: TObject);
    procedure NewFileBtnClick(Sender: TObject);
    procedure DisplayPaintPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure EditSimBtnClick(Sender: TObject);
    procedure SaveParamsBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure LoadParams;
    procedure SaveParams;
    procedure InitRun;
    procedure RunLoop;
    procedure Iterate;
    procedure IdleLoop(Sender: TObject; var Done: boolean);
  end;

const
  MaxObjects = 256;
  BackGrnd = clWhite;
  ParamLegends:
         array[0..17] of string =
                  ( 'InitX..........',
                    'InitY..........',
                    'InitDir........',
                    'AvoidProx.r....',
                    'AvoidProx.g...',
                    'SeekProx.r....',
                    'SeekProx.g...',
                    'SeekDir.r.....',
                    'SeekDir.g....',
                    'DestProx.r....',
                    'DestProx.g...',
                    'DestDir.r.....',
                    'DestDir.g....',
                    'Seek Pers/Dest.',
                    'SeekPerNum.....',
                    'Active/Inact...',
                    'Destination X..',
                    'Destination Y..');

type
  ControlRec = record
                 p, r, e, g, o: Single;
               end;
  PersonRecordType = record
                       InitX, Inity, InitDir: Single;
                       Dummy, AvoidProx, DestProx, DestDir, SeekProx,
                         SeekDir: ControlRec;
                       SeekPD, SeekPerNum, Status: SmallInt;
                       DestX, DestY, PerX, PerY, PerDir: Single;
                     end;

     PersonParamType = array[0..MaxObjects - 1] of PersonRecordType;

var
  MainForm: TMainForm;
  PP: PersonParamType;
  Paramfile: file of PersonParamType;
  Paramfilename: string;
  ButtonHelpOn: Boolean;
  Person: SmallInt;
  Editing, RandomPos, CopyReady, FileClosed: Boolean;
  FirstCopyNum, LastCopyNum: SmallInt;

implementation

uses About, CrowdEdit;

{$R *.dfm}

var
  OffScreen: TBitmap;
  Radius: SmallInt;
  x0, y0: SmallInt;
  Xscale, Yscale: Single;
  OldX: array[0..MaxObjects - 1] of Integer;
  OldY: array[0..MaxObjects - 1] of Integer;
  RelProx, RelDir: Single;
  Vel: Single;
  HCenter, VCenter: Integer;
  LeftP, RightP: Single;
  NumObjects, Myself, Stage: Integer;
  Trails, Running, ParamsLoaded, Pause: Boolean;
  t1, t2, Freq: Int64;

procedure CountPersons;
var
  I: Integer;
begin
  NumObjects := 0;
  for I := 0 to MaxObjects -1 do
    if PP[i].Status <> -1 then Inc(NumObjects);
end;

procedure LoadDefaultParams;
begin
  Person := 0;
  with PP[Person] do
  begin
    InitX := -450.0;
    InitY := 350.0;
    InitDir := 0.0;
    AvoidProx.r := 100.0;
    AvoidProx.g := 60.0;
    DestProx.r := 255.0;
    DestProx.g := 25.0;
    DestDir.r:= 0.0;
    DestDir.g := 100;
    SeekProx.r := 0.0;
    SeekProx.g := 0.0;
    SeekDir.r := 0.0;
    SeekDir.g := 0.0;
    SeekPD := 1;
    SeekPerNum := 0;
    Status := 1;
    DestX := 450.0;
    DestY := -350.0;
  end;
end;

procedure TMainForm.LoadParams;
var
  I, P: Integer;
  GoodFile: Boolean;
begin
  with LoadParamDlg do
  begin
    InitialDir := GetCurrentDir;
    GoodFile := Execute;
    if GoodFile then
    begin
      ParamFilename := Filename;

      P := Pos('.per',ParamFilename);
      if P > 0 then
        Delete(ParamFilename,P,Length(ParamFilename) - P + 1);
      ParamFilename := ParamFilename + '.per';
      Filename := ParamFilename;
      
      AssignFile(ParamFile, ParamFilename);
      {$I-}
      Reset(ParamFile);
      {$I+}
      GoodFile := (IOresult = 0) and (not EOF(ParamFile));
      if GoodFile then
      begin
        {$i-}
        Read(ParamFile, PP);
        {$i+}
        P := IOresult;
        if P <> 0 then CloseFile(ParamFile);
        Delete(ParamFilename,1,Length(InitialDir)+1);
        P := Pos('.',ParamFilename);
        //NOTE!!! EXTENSION STRIPPED OFF PARAMFILENAME HERE
        if P > 0 then
          Delete(ParamFilename,P,Length(ParamFilename) - P + 1);
      end;
    end;
    if not GoodFile then
    begin
      LoadDefaultParams;
      for I := 1 to 255 do PP[I].Status := -1;
    end;
    Person := 0;
  end;
  FileClosed := False;
  Randomize;
  ParamsLoaded := True;
  CountPersons;
end;

procedure TMainForm.SaveParams;
var
  GoodFile: Boolean;
begin
  with SaveParamDlg do
  begin
    InitialDir := GetCurrentDir;
    Filter := 'Person files (*.per)|*.PER';
    DefaultExt := 'per';
    Options := [ofOverwritePrompt];
    if Execute then
    begin
        ParamFilename := Filename;
        AssignFile(ParamFile,ParamFilename);
        {$I-}
        Rewrite(ParamFile);
        {$I+}
        GoodFile := (IOresult = 0);
        Write(Paramfile,PP);
        CloseFile(ParamFile);
        InitRun;
    end;
  end;
end;

procedure TMainForm.InitRun;
var
  I, X, Y: integer;
  TempName: String;
begin
  with OffScreen, Canvas do
  begin
    X0 := 80;
    Y0 := 0;
    Xscale := (Width - X0)/1000.0;
    Yscale := (Height - Y0)/800.0;
    HCenter := X0 + (Width - X0) div 2;
    VCenter := Y0 + (Height - Y0) div 2;
    Brush.Color := clWhite;
    FillRect(rect(0, 0, X0 + Round(1000*Xscale), Y0 + Round(800*Yscale)));
    Pen.Width := 2;
    Radius := 5;
    I := 0;
    while (I < 256) do
      with PP[I] do
      begin
        if (Status in [0, 2]) then // In use but inactive or fixed
        begin
          if Status = 0 then
          begin
            InitX:=  Random(600) - 300;
            InitY := Random(500) - 250;
          end;
          PerX := InitX;
          PerY := InitY;
          X := HCenter + Round(Xscale*PerX);
          Y := VCenter - Round(Yscale*perY);
          Pen.Color := clBlack;
          Ellipse(Rect(X - Radius, Y - Radius,
                       X + Radius, Y + Radius));
        end
        else
          if Status = 1 then // active person
          begin
            PerX := InitX;
            PerY := InitY;
            PerDir := InitDir;
            OldX[I] := Round(Xscale*InitX);
            OldY[I] := Round(YScale*InitY);
          end;
        Inc(I);
      end;
    TempName := ExtractFilename(ParamFilename);
    SimNameEdit.Text := ChangeFileExt(TempName, '');
  end;
  Stage := 0;
  DisplayPaint.Canvas.Draw(0, 0, Offscreen);
end;

{ -------------------------------------------------------------------------}

function ProxFunc(P, ProxFactor: Single): Single;
begin
  ProxFunc := 255.0*ProxFactor/(ProxFactor + Sqr(P));
end;

{ -------------------------------------------------------------------------}

// Returns distance and angle as single
procedure FindRhoTheta(var Distance, Angle: Single;
                       X1, Y1, X2, Y2: Single;
                       Myself: SmallInt);
var
  X, Y: Single;
begin
  X := X2 - X1; Y := Y2 - Y1;
  Angle := ArcTan2(Y, X); // angle in radians, 0 to 2pi
  Distance := Sqrt(Sqr(X) + Sqr(Y));
end;

{ -------------------------------------------------------------------------}
{           THIS LOOP REPEATS RUNS AND ALLOWS PARAMETER CHANGES            }
{ -------------------------------------------------------------------------}

procedure TMainForm.RunLoop;
var
  X, Y, R, I, J: Integer;
  D, Temp: Single;
  ColorCircle: Tcolor;
begin
  J := 0;
  while (J < 255)  do  // scan through all objects
    with PP[J] do
    begin
      with OffScreen.Canvas do
        case SeekPD of
          1: ColorCircle := clRed;
          2: ColorCircle := RGB(0, 200, 100);
          else ColorCircle := clBlack;
        end;
      if Status = 1 then
      begin
        Myself := J;
        FindRhoTheta(DestProx.p, DestDir.p,
                     PerX, PerY, DestX, DestY,
                     Myself);
        DestProx.p := ProxFunc(Destprox.p, 200);  // convert distance to proximity
        Temp := DestDir.P - PerDir;
        while Temp <= -pi do Temp := Temp + 2*pi;
        while Temp > pi do Temp := Temp - 2*pi;
        PP[j].DestDir.P := Temp;   //Dest angle relative to travel angle

        // Routine for avoiding collisions
        LeftP := 0;
        RightP := 0;
        for I := 0 to MaxObjects - 1 do
          if (I <> Myself) and (PP[I].Status <> -1) then
          begin
            FindRhoTheta(RelProx, RelDir,
                         PP[Myself].PerX, PP[Myself].PerY,
                         PP[i].PerX, PP[i].PerY,Myself);
            D := RelProx - Radius; if D < 0.0 then D := 0.0;
            RelDir := RelDir - PerDir;  // dist and direction to obstacle
            while RelDir <= -pi do RelDir := RelDir + 2*pi; // bring within range
            while RelDir > pi do RelDir := RelDir - 2*pi;
            Temp := Cos(RelDir/2);               // cardioid sensitivity pattern
            if Temp < 0 then Temp := 0.0;
            if (RelDir <= 0) then
              RightP := RightP + ProxFunc(D, 100)*Temp  // convert to relative proximity
            else if (RelDir > 0) then                  // and sum, for both sides
              LeftP := LeftP + ProxFunc(D, 100)*Temp;
          end;
        if LeftP < 0.0 then LeftP := 0.0;
        if RightP < 0.0 then RightP := 0.0;

        with AvoidProx do   // travel along contour line
        begin
          p := RightP + LeftP;
          e := r - p;
          if e > 0 then e := 0;
          if LeftP < RightP then e := -e;
          o := 0.02*g*e*(1 + (Random - 0.5)); // 50% dither
        end;

        // Routines for seeking another person's position
        if SeekPD = 2 then
        begin
          FindRhoTheta(Temp, RelDir, PerX, PerY,
                  PP[SeekPerNum].PerX,
                  PP[SeekPerNum].PerY, Myself);
          RelProx := ProxFunc(Temp, 150);
          RelDir := RelDir - PerDir; // convert direction to relative direction
          while RelDir < -pi do RelDir := RelDir + 2*pi;
          while RelDir >= pi do RelDir := RelDir - 2*pi;

          with SeekProx do
          begin
            p := RelProx;
            e := r - p;
            o := 0.05*g*e;
          end;

          with SeekDir do
          begin
            p := RelDir;
            e := r - p;
            while e > pi do e := e - 2*pi;
            while e < -pi do e := e + 2*pi;
            o := -0.2*g*e;
          end;
        end;

        with DestDir do
        begin
          e := r - p;
          while e > pi do e := e - 2*pi;
          while e < -pi do e := e + 2*pi;
          o :=  -0.2*g*e;
        end;

        with DestProx do
        begin
          e := r - p;
          o := 0.002*g*e;
        end;

        PerDir :=  PerDir +  0.01*(DestDir.o + AvoidProx.o + SeekDir.o);
        while PerDir > pi do PerDir := PerDir - 2*pi;
        while PerDir < -pi do PerDir := PerDir + 2*pi;


        Vel := 0.1*(DestProx.o + SeekProx.o)/(1 + 0.05*abs(AvoidProx.e));
        if Vel < 0 then vel := 0.0
        else
          if Vel > 5 then vel := 5;

        // COMPUTE CONTROL OUTPUTS AND PLOT POSITIONS
        with OffScreen, Canvas do
        begin
          Pen.Color := clLime;   // rewrite destination circle
          X := HCenter + Round(Xscale*DestX);
          Y := VCenter - Round(Yscale*DestY);
          if PP[J].SeekPD = 1 then
          Ellipse(Rect(X - 10, Y - 10, X + 10, Y + 10));
          Pen.Color := BackGrnd;
          Ellipse(Rect(Oldx[J] - Radius, Oldy[J] - Radius,
                  Oldx[J] + Radius, Oldy[J] + Radius)); //erase
          PerX := PerX + Vel*Cos(PerDir);
          PerY := PerY + Vel*Sin(PerDir);
          OldX[J] := HCenter + Round(Xscale*PerX);
          OldY[J] := VCenter - Round(Yscale*PerY);
          Pen.Color := ColorCircle;
          Ellipse(Rect(OldX[J] - Radius,OldY[J] - Radius,
                  OldX[J] + Radius,OldY[J] + Radius));
          if Trails then
          begin
            X := HCenter + Round(Xscale*(PerX - 2*Radius*Cos(PerDir)));
            Y := VCenter - Round(Yscale*(PerY - 2*Radius*Sin(PerDir)));
            Ellipse(Rect(X-1, Y-1, X+1, Y+1));
          end;
        end;
    end;
    Inc(J);   // do scan for next object
  end;
  DisplayPaint.Canvas.Draw(0, 0, OffScreen);
end;

procedure TMainForm.Iterate;
begin
  case Stage of
    0: begin
         with OffScreen, Canvas do
         begin
           Width := DisplayPaint.Width;
           Height := DisplayPaint.Height;
           Brush.Color := clWhite;
           FillRect(rect(0, 0, X0 + Round(1000*Xscale), Y0 + Round(800*Yscale)));
         end;
         DisplayPaint.Canvas.Draw(0, 0, OffScreen);
         if not ParamsLoaded then
         begin
           LoadParams;
           InitRun;
           Trails := False;
         end;
       end;
    1: if not Pause then RunLoop;
  end;
end;

procedure TMainForm.IdleLoop(Sender: TObject; var Done: boolean);
begin
  Done := False;
  Iterate;
  repeat
    QueryPerformanceCounter(t2);
  until (t2 - t1) >= Freq/30;
  t1 := t2;
end;

procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  CursimLabel.SetFocus;
  AboutBox.ShowModal;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  OffScreen := TBitmap.Create;
  ChangeScale(Screen.Width, 1280);
  CopyReady := False;
  Running := False;
  Stage := 0;
  ParamsLoaded := False;
  Randomize;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(t1);
  Application.OnIdle := IdleLoop;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  OffScreen.Free;
end;

procedure TMainForm.RunBtnClick(Sender: TObject);
begin
  InitRun;
  Stage := 1;
  if Pause then
  begin
    Pause := False;
    PauseBtn.Caption := 'Pause';
  end;
end;

procedure TMainForm.PauseBtnClick(Sender: TObject);
begin
  Pause := not Pause;
  if Pause then PauseBtn.Caption := 'Continue'
    else PauseBtn.Caption := 'Pause';
end;

procedure TMainForm.TrailsBtnClick(Sender: TObject);
begin
  Trails := not Trails;
  if not Trails then TrailsBtn.Caption := 'Turn Trails On'
  else TrailsBtn.Caption := 'Turn Trails Off';
end;

procedure TMainForm.NewFileBtnClick(Sender: TObject);
begin
  ParamsLoaded := False;
  Stage := 0;
end;

procedure TMainForm.DisplayPaintPaint(Sender: TObject);
begin
  DisplayPaint.Canvas.Draw(0, 0, OffScreen);
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  with OffScreen do
  begin
    Height := DisplayPaint.Height;
    Width := DisplayPaint.Width;
  end;
end;

procedure TMainForm.EditSimBtnClick(Sender: TObject);
begin
  CrowdEditForm.ShowModal;
end;

procedure TMainForm.SaveParamsBtnClick(Sender: TObject);
begin
  SaveParams;
end;

end.
